@@ -44,6 +44,18 @@
  */
 #define DEFAULT_SERVER "localhost irc.efnet.net irc.undernet.org irc.dal.net"
 
+/*
+ * This is an experimental feature to thwart infinite recursion.  It is not
+ * very sophisticated so it's not turned on by default.  In the future I 
+ * will probably do something less lame.
+ *
+ * When this many stack frames are created, epic will refuse to make any
+ * more.  This will stop epic from crashing with a segfault when you do 
+ * infinite recursion, but it's possible that epic may infinitely recurse
+ * if your alias is tricky enough.  Pick your poison
+ */
+#define MAX_STACK_FRAMES 20000
+
 /*
  * The left and right brace characters ('{', '}') are special characters in
  * the IRC character set, becuase they represent scandanavian characters and
@@ -311,12 +323,14 @@
  * remove them.  They are default values for variables and are required for
  * proper compilation.
  */
+#define DEFAULT_ALLOW_C1_CHARS 0
 #define DEFAULT_ALT_CHARSET 1
 #define DEFAULT_ALWAYS_SPLIT_BIGGEST 1
 #define DEFAULT_AUTO_NEW_NICK 1
 #define DEFAULT_AUTO_RECONNECT 1
 #define DEFAULT_AUTO_RECONNECT_DELAY 0
-#define DEFAULT_AUTO_REJOIN 1
+#define DEFAULT_AUTO_REJOIN 0
+#define DEFAULT_AUTO_REJOIN_CONNECT 1
 #define DEFAULT_AUTO_REJOIN_DELAY 0
 #define DEFAULT_AUTO_UNMARK_AWAY 0
 #define DEFAULT_AUTO_WHOWAS 1
@@ -332,6 +346,7 @@
 #define DEFAULT_CLOCK 1
 #define DEFAULT_CLOCK_24HOUR 0
 #define DEFAULT_CLOCK_FORMAT NULL
+#define DEFAULT_CLOCK_INTERVAL 60
 #define DEFAULT_CMDCHARS "/"
 #define DEFAULT_COLOR 1
 #define DEFAULT_COMMAND_MODE 0
@@ -339,9 +354,10 @@
 #define DEFAULT_CONNECT_TIMEOUT 30
 #define DEFAULT_CONTINUED_LINE "+"
 #define DEFAULT_CPU_SAVER_AFTER 0
-#define DEFAULT_CPU_SAVER_EVERY 0
+#define DEFAULT_CPU_SAVER_EVERY 60
 #define DEFAULT_CURRENT_WINDOW_LEVEL NULL
 #define DEFAULT_DCC_AUTO_SEND_REJECTS 1
+#define DEFAULT_DCC_DEQUOTE_FILENAMES 1
 #define DEFAULT_DCC_LONG_PATHNAMES 1
 #define DEFAULT_DCC_SLIDING_WINDOW 1
 #define DEFAULT_DCC_TIMEOUT 3600
@@ -353,8 +369,12 @@
 #define DEFAULT_EIGHT_BIT_CHARACTERS 1
 #define DEFAULT_EXEC_PROTECTION 0
 #define DEFAULT_FLOATING_POINT_MATH 0
+#define DEFAULT_FLOATING_POINT_PRECISION 16
 #define DEFAULT_FLOOD_AFTER 3
+#define DEFAULT_FLOOD_IGNORE 0
+#define DEFAULT_FLOOD_MASKUSER 0
 #define DEFAULT_FLOOD_RATE 3
+#define DEFAULT_FLOOD_RATE_PER 1
 #define DEFAULT_FLOOD_USERS 3
 #define DEFAULT_FLOOD_WARNING 0
 #define DEFAULT_FULL_STATUS_LINE 1
@@ -368,25 +388,28 @@
 #define DEFAULT_HISTORY_CIRCLEQ 1
 #define DEFAULT_HOLD_INTERVAL 10
 #define DEFAULT_HOLD_MODE 0
+#define DEFAULT_HOLD_SLIDER 100
 #define DEFAULT_INDENT 0
 #define DEFAULT_INPUT_ALIASES 0
 #define DEFAULT_INPUT_PROMPT "> "
 #define DEFAULT_INPUT_PROTECTION 1
 #define DEFAULT_INSERT_MODE 1
 #define DEFAULT_INVERSE_VIDEO 1
+#define DEFAULT_KEY_INTERVAL 1000
 #define DEFAULT_LASTLOG 256
 #define DEFAULT_LASTLOG_LEVEL "ALL"
 #define DEFAULT_LOG 0
 #define DEFAULT_LOGFILE "irc.log"
 #define DEFAULT_MAIL 2
+#define DEFAULT_MAIL_INTERVAL 60
 #define DEFAULT_MAX_RECONNECTS 4
-#define DEFAULT_META_STATES 40
+#define DEFAULT_METRIC_TIME 0
 #define DEFAULT_MODE_STRIPPER 0
 #define DEFAULT_ND_SPACE_MAX 160
-#define DEFAULT_NEW_SERVER_LASTLOG_LEVEL "ALL"
+#define DEFAULT_NEW_SERVER_LASTLOG_LEVEL "ALL,-DCC"
 #define DEFAULT_NO_CTCP_FLOOD 1
 #define DEFAULT_NO_FAIL_DISCONNECT 0
-#define DEFAULT_NOTIFY_HANDLER "NOISY"
+#define DEFAULT_NOTIFY 1
 #define DEFAULT_NOTIFY_INTERVAL 60
 #define DEFAULT_NOTIFY_LEVEL "ALL"
 #define DEFAULT_NOTIFY_ON_TERMINATION 1
@@ -413,6 +436,9 @@
 #define DEFAULT_STATUS_AWAY " (Away)"
 #define DEFAULT_STATUS_CHANNEL " %C"
 #define DEFAULT_STATUS_CHANOP "@"
+#define DEFAULT_STATUS_HALFOP "%"
+#define DEFAULT_STATUS_SSL_ON "*SSL*"
+#define DEFAULT_STATUS_SSL_OFF "*RAW*"
 #define DEFAULT_STATUS_CLOCK " %T"
 #define DEFAULT_STATUS_CPU_SAVER " (%L)"
 #define DEFAULT_STATUS_FORMAT "%T [%R] %*%=%@%N%#%S%H%B%Q%A%C%+%I%O%M%F%L %D %U %W"
@@ -425,6 +451,7 @@
 #define DEFAULT_STATUS_MAIL " (Mail: %M)"
 #define DEFAULT_STATUS_NICK "%N"
 #define DEFAULT_STATUS_NO_REPEAT 0
+#define	DEFAULT_STATUS_NOSWAP "(NOSWAP)"
 #define	DEFAULT_STATUS_NOTIFY " (W: %F)"
 #define DEFAULT_STATUS_OPER "*"
 #define DEFAULT_STATUS_OVERWRITE " (Overwrite)"
@@ -478,6 +505,7 @@
 #define DEFAULT_SUPPRESS_FROM_REMOTE_SERVER 0
 #define DEFAULT_SUPPRESS_SERVER_MOTD 0
 #define DEFAULT_SWITCH_CHANNEL_ON_PART 1
+#define DEFAULT_SWITCH_CHANNELS_BETWEEN_WINDOWS 1
 #define DEFAULT_TAB 1
 #define	DEFAULT_TAB_MAX 0
 #define DEFAULT_TERM_DOES_BRIGHT_BLINK 0
@@ -485,6 +513,7 @@
 #define DEFAULT_USERINFO "EPIC4 -- Get your laundry brighter, not just whiter!"
 #define DEFAULT_VERBOSE_CTCP 1
 #define DEFAULT_WORD_BREAK ".,; \t"
+#define DEFAULT_WSERV_TYPE "screen"
 #define DEFAULT_XTERM "xterm"
 #define DEFAULT_XTERM_OPTIONS NULL
 
@@ -557,5 +586,7 @@
 #endif
 /* end of section not to change */
 
+#undef ALLOC_DEBUG
+
 #endif /* _CONFIG_H_ */
 
