@@ -9,7 +9,11 @@
 #ifndef __mail_h__
 #define __mail_h__
 
-	char	*check_mail 		(void);
-	int	check_mail_status 	(void *);
+extern	char	mail_timeref[];
+
+	const char *	check_mail	(void);
+	void	mail_systimer		(void);
+	void	set_mail_interval	(const void *);
+	void	set_mail		(const void *);
 
 #endif /* _MAIL_H_ */
