@@ -1,20 +1,44 @@
+/* $EPIC: ctcp.c,v 1.33 2004/01/27 04:41:39 crazyed Exp $ */
 /*
  * ctcp.c:handles the client-to-client protocol(ctcp). 
  *
- * Written By Michael Sandrof 
- * Copyright 1990, 1995 Michael Sandroff and others
- * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
+ * Copyright (c) 1990 Michael Sandroff.
+ * Copyright (c) 1991, 1992 Troy Rollo.
+ * Copyright (c) 1992-1996 Matthew Green.
+ * Copyright  1993, 2003 EPIC Software Labs
+ * All rights reserved.
  *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notices, the above paragraph (the one permitting redistribution),
+ *    this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ * 3. The names of the author(s) may not be used to endorse or promote
+ *    products derived from this software without specific prior written
+ *    permission.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
+ * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
+ * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
+ * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
+ * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
+ * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+ * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
+ * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
+ * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
+ * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ * SUCH DAMAGE.
+ */
+/*
  * Serious cleanup by jfn (August 1996)
- * Copyright 1996 EPIC Software Labs
  */
 
-#if 0
-static	char	rcsid[] = "@(#)$Id: ctcp.c,v 1.1.1.1 2000/12/05 00:11:56 jnelson Exp $";
-#endif
-
 #include "irc.h"
-#include "crypt.h"
+#include "sedcrypt.h"
 #include "ctcp.h"
 #include "dcc.h"
 #include "commands.h"
@@ -31,6 +55,8 @@ static	char	rcsid[] = "@(#)$Id: ctcp.c,v 1.1.1.1 2000/12/05 00:11:56 jnelson Exp
 #include "window.h"
 #include "if.h"
 #include "flood.h"
+#include "notice.h"
+#include "words.h"
 
 #include <pwd.h>
 #ifdef HAVE_UNAME
@@ -55,31 +81,31 @@ static	int 	split_CTCP (char *, char *, char *);
  */
 
 struct _CtcpEntry;
-typedef char *(*CTCP_Handler) (struct _CtcpEntry *, char *, char *, char *);
+typedef char *(*CTCP_Handler) (struct _CtcpEntry *, const char *, const char *, char *);
 typedef	struct _CtcpEntry
 {
-	char		*name;  /* name of ctcp datatag */
+	const char	*name;  /* name of ctcp datatag */
 	int		id;	/* index of this ctcp command */
 	int		flag;	/* Action modifiers */
-	char		*desc;  /* description returned by ctcp clientinfo */
+	const char	*desc;  /* description returned by ctcp clientinfo */
 	CTCP_Handler 	func;	/* function that does the dirty deed */
 	CTCP_Handler 	repl;	/* Function that is called for reply */
 }	CtcpEntry;
 
 /* forward declarations for the built in CTCP functions */
-static	char	*do_sed 	(CtcpEntry *, char *, char *, char *);
-static	char	*do_version 	(CtcpEntry *, char *, char *, char *);
-static	char	*do_clientinfo 	(CtcpEntry *, char *, char *, char *);
-static	char	*do_ping 	(CtcpEntry *, char *, char *, char *);
-static	char	*do_echo 	(CtcpEntry *, char *, char *, char *);
-static	char	*do_userinfo 	(CtcpEntry *, char *, char *, char *);
-static	char	*do_finger 	(CtcpEntry *, char *, char *, char *);
-static	char	*do_time 	(CtcpEntry *, char *, char *, char *);
-static	char	*do_atmosphere 	(CtcpEntry *, char *, char *, char *);
-static	char	*do_dcc 	(CtcpEntry *, char *, char *, char *);
-static	char	*do_utc 	(CtcpEntry *, char *, char *, char *);
-static	char	*do_dcc_reply 	(CtcpEntry *, char *, char *, char *);
-static	char	*do_ping_reply 	(CtcpEntry *, char *, char *, char *);
+static	char	*do_sed 	(CtcpEntry *, const char *, const char *, char *);
+static	char	*do_version 	(CtcpEntry *, const char *, const char *, char *);
+static	char	*do_clientinfo 	(CtcpEntry *, const char *, const char *, char *);
+static	char	*do_ping 	(CtcpEntry *, const char *, const char *, char *);
+static	char	*do_echo 	(CtcpEntry *, const char *, const char *, char *);
+static	char	*do_userinfo 	(CtcpEntry *, const char *, const char *, char *);
+static	char	*do_finger 	(CtcpEntry *, const char *, const char *, char *);
+static	char	*do_time 	(CtcpEntry *, const char *, const char *, char *);
+static	char	*do_atmosphere 	(CtcpEntry *, const char *, const char *, char *);
+static	char	*do_dcc 	(CtcpEntry *, const char *, const char *, char *);
+static	char	*do_utc 	(CtcpEntry *, const char *, const char *, char *);
+static	char	*do_dcc_reply 	(CtcpEntry *, const char *, const char *, char *);
+static	char	*do_ping_reply 	(CtcpEntry *, const char *, const char *, char *);
 
 
 static CtcpEntry ctcp_cmd[] =
@@ -125,7 +151,7 @@ static CtcpEntry ctcp_cmd[] =
 		NULL, NULL }
 };
 
-static char	*ctcp_type[] =
+static const char	*ctcp_type[] =
 {
 	"PRIVMSG",
 	"NOTICE"
@@ -142,7 +168,7 @@ int     sed = 0;
 int	in_ctcp_flag = 0;
 
 #define CTCP_HANDLER(x) \
-	static char * x (CtcpEntry *ctcp, char *from, char *to, char *cmd)
+	static char * x (CtcpEntry *ctcp, const char *from, const char *to, char *cmd)
 
 
 
@@ -156,8 +182,9 @@ int	in_ctcp_flag = 0;
  */
 CTCP_HANDLER(do_sed)
 {
-	char	*key = NULL,
-		*crypt_who;
+	Crypt	*key = NULL;
+	const char	*crypt_who;
+	char 	*tofrom;
 	char	*ret = NULL, *ret2 = NULL;
 
 	if (*from == '=')		/* DCC CHAT message */
@@ -167,18 +194,30 @@ CTCP_HANDLER(do_sed)
 	else
 		crypt_who = to;
 
-	if ((key = is_crypted(crypt_who)))
+	tofrom = malloc_strdup3(to, ",", from);
+	malloc_strcat2_c(&tofrom, "!", FromUserHost, NULL);
+
+	if ((key = is_crypted(tofrom)) ||
+	    (key = is_crypted(crypt_who)))
 		ret = decrypt_msg(cmd, key);
 
-	if (!key || !ret)
+	new_free(&tofrom);
+
+	if (!key || !ret) {
+		sed = 2;
 		malloc_strcpy(&ret2, "[ENCRYPTED MESSAGE]");
-	else
-	{
+	} else if (!*ret) {
+		sed = 2;
+		malloc_strcpy(&ret2, "[ENCRYPTED MESSAGE - BAD KEY?]");
+	} else {
 		/* 
 		 * There might be a CTCP message in there,
 		 * so we see if we can find it.
 		 */
-		ret2 = m_strdup(do_ctcp(from, to, ret));
+		if (get_server_doing_privmsg(from_server))
+			ret2 = malloc_strdup(do_ctcp(from, to, ret));
+		else if (get_server_doing_notice(from_server))
+			ret2 = malloc_strdup(do_notice_ctcp(from, to, ret));
 		sed = 1;
 	}
 
@@ -189,9 +228,9 @@ CTCP_HANDLER(do_sed)
 CTCP_HANDLER(do_utc)
 {
 	if (!cmd || !*cmd)
-		return m_strdup(empty_string);
+		return malloc_strdup(empty_string);
 
-	return m_strdup(my_ctime(my_atol(cmd)));
+	return malloc_strdup(my_ctime(my_atol(cmd)));
 }
 
 
@@ -215,7 +254,7 @@ CTCP_HANDLER(do_atmosphere)
 		message_from(to, LOG_ACTION);
 		if (do_hook(ACTION_LIST, "%s %s %s", from, to, cmd))
 		{
-			if (is_current_channel(to, 0))
+			if (is_current_channel(to, from_server))
 				put_it("* %s %s", from, cmd);
 			else
 				put_it("* %s:%s %s", from, to, cmd);
@@ -245,11 +284,13 @@ CTCP_HANDLER(do_dcc)
 	char	*size;
 	char	*extra_flags;
 
-	if (!is_me(from_server, to))
+	if (!is_me(from_server, to) && *from != '=')
 		return NULL;
 
 	if     (!(type = next_arg(cmd, &cmd)) ||
-		!(description = next_arg(cmd, &cmd)) ||
+		!(description = (get_int_var(DCC_DEQUOTE_FILENAMES_VAR)
+			 ? new_next_arg(cmd, &cmd)
+			 : next_arg(cmd, &cmd))) ||
 		!(inetaddr = next_arg(cmd, &cmd)) ||
 		!(port = next_arg(cmd, &cmd)))
 			return NULL;
@@ -257,7 +298,7 @@ CTCP_HANDLER(do_dcc)
 	size = next_arg(cmd, &cmd);
 	extra_flags = next_arg(cmd, &cmd);
 
-	register_dcc_offer(from, type, description, inetaddr, port, size, extra_flags);
+	register_dcc_offer(from, type, description, inetaddr, port, size, extra_flags, cmd);
 	return NULL;
 }
 
@@ -298,8 +339,8 @@ CTCP_HANDLER(do_clientinfo)
 
 		for (i = 0; i < NUMBER_OF_CTCPS; i++)
 		{
-			strmcat(buffer, ctcp_cmd[i].name, BIG_BUFFER_SIZE);
-			strmcat(buffer, " ", BIG_BUFFER_SIZE);
+			strlcat(buffer, ctcp_cmd[i].name, sizeof buffer);
+			strlcat(buffer, " ", sizeof buffer);
 		}
 		send_ctcp(CTCP_NOTICE, from, CTCP_CLIENTINFO,
 			"%s :Use CLIENTINFO <COMMAND> to get more specific information", 
@@ -322,8 +363,8 @@ CTCP_HANDLER(do_version)
 	 */
 #if defined(HAVE_UNAME) && !defined(UNAME_HACK)
 	struct utsname un;
-	char	*the_unix,
-		*the_version;
+	const char	*the_unix,
+			*the_version;
 
 	if (uname(&un) < 0)
 	{
@@ -398,6 +439,8 @@ CTCP_HANDLER(do_finger)
 	char	*ctcpuser,
 		*ctcpfinger;
 	const char	*my_host;
+	char	userbuff[NAME_LEN + 1];
+	char	gecosbuff[NAME_LEN + 1];
 
 	if ((my_host = get_server_userhost(from_server)) &&
 			strchr(my_host, '@'))
@@ -405,7 +448,7 @@ CTCP_HANDLER(do_finger)
 	else
 		my_host = hostname;
 
-	diff = time(NULL) - idle_time;
+	diff = time(NULL) - idle_time.tv_sec;
 
 	if (!(pwd = getpwuid(getuid())))
 		return NULL;
@@ -414,18 +457,33 @@ CTCP_HANDLER(do_finger)
 #define GECOS_DELIMITER ','
 #endif
 
-	if ((tmp = strchr(pwd->pw_gecos, GECOS_DELIMITER)) != NULL)
-		*tmp = '\0';
-
 #if !defined(I_DONT_TRUST_MY_USERS)
 	if ((ctcpuser = getenv("IRCUSER"))) 
-		strmcpy(pwd->pw_name, ctcpuser, NAME_LEN);
-	if ((ctcpfinger = getenv("IRCFINGER"))) 
-		strmcpy(pwd->pw_gecos, ctcpfinger, NAME_LEN);
+		strlcpy(userbuff, ctcpuser, sizeof userbuff);
+	else
 #endif
+	{
+		if (pwd->pw_name)
+			strlcpy(userbuff, pwd->pw_name, sizeof userbuff);
+		else
+			strlcpy(userbuff, "epic-user", sizeof userbuff);
+	}
+
+#if !defined(I_DONT_TRUST_MY_USERS)
+	if ((ctcpfinger = getenv("IRCFINGER"))) 
+		strlcpy(gecosbuff, ctcpfinger, sizeof gecosbuff);
+	else
+#endif
+	      if (pwd->pw_gecos)
+		strlcpy(gecosbuff, pwd->pw_gecos, sizeof gecosbuff);
+	else
+		strlcpy(gecosbuff, "Esteemed EPIC User", sizeof gecosbuff);
+	if ((tmp = strchr(gecosbuff, GECOS_DELIMITER)) != NULL)
+		*tmp = 0;
+
 	send_ctcp(CTCP_NOTICE, from, CTCP_FINGER, 
 		"%s (%s@%s) Idle %ld second%s", 
-		pwd->pw_gecos, pwd->pw_name, my_host, diff, plural(diff));
+		gecosbuff, userbuff, my_host, diff, plural(diff));
 
 	return NULL;
 }
@@ -460,11 +518,11 @@ CTCP_HANDLER(do_dcc_reply)
  */
 CTCP_HANDLER(do_ping_reply)
 {
-	struct timeval 	t;
-	time_t 		tsec = 0, 
-			tusec = 0, 
-			orig;
-	char *		ptr;
+	Timeval t;
+	time_t 	tsec = 0, 
+		tusec = 0, 
+		orig;
+	char *	ptr;
 
 	if (!cmd || !*cmd)
 		return NULL;		/* This is a fake -- cant happen. */
@@ -473,7 +531,7 @@ CTCP_HANDLER(do_ping_reply)
 	get_time(&t);
 
 	/* Reply must be between time we started and right now */
-	if (orig < start_time || orig > t.tv_sec)
+	if (orig < start_time.tv_sec || orig > t.tv_sec)
 	{
 		say("Invalid CTCP PING reply [%s] dropped.", cmd);
 		return NULL;
@@ -481,19 +539,20 @@ CTCP_HANDLER(do_ping_reply)
 
 	tsec = t.tv_sec - orig;
 
-	if ((ptr = strchr(cmd, ' ')))
+	if ((ptr = sindex(cmd, " .")))
 	{
 		*ptr++ = 0;
 		tusec = t.tv_usec - my_atol(ptr);
 	}
 
 	/*
-	 * 'cmd', a variable passed in from do_notice_ctcp()
-	 * points to a buffer which is MUCH larger than the
-	 * string 'cmd' points at.  So this is safe, even
-	 * if it looks "unsafe".
+	 * 'cmd' is a pointer to the inside of do_ctcp's 'the_ctcp' buffer
+	 * which is IRCD_BUFFER_SIZE bytes big; cmd points to (allegedly)
+	 * the sixth position.  But just be paranoid and assume half that, 
+	 * so we will always be safe.
 	 */
-	sprintf(cmd, "%f seconds", (float)(tsec + (tusec / 1000000.0)));
+	snprintf(cmd, IRCD_BUFFER_SIZE / 2, "%f seconds", 
+			(float)(tsec + (tusec / 1000000.0)));
 	return NULL;
 }
 
@@ -520,11 +579,11 @@ CTCP_HANDLER(do_ping_reply)
  * contain any CTCPs.  So we check to see if there are any CTCPs in the
  * message before we bother doing anything.  THIS IS AN INELEGANT HACK!
  * But the call to charcount() is less expensive than even one copy to 
- * strcpy() since they both evaluate *each* character, and charcount()
+ * strlcpy() since they both evaluate *each* character, and charcount()
  * doesnt have to do a write unless the character is present.  So it is 
  * definitely worth the cost to save CPU time for 99% of the PRIVMSGs.
  */
-char *	do_ctcp (char *from, char *to, char *str)
+char *	do_ctcp (const char *from, const char *to, char *str)
 {
 	int 	flag;
 	int	fflag;
@@ -547,14 +606,14 @@ static	time_t	last_ctcp_parsed = 0;
 		allow_ctcp_reply = 0;	/* Historical limit of 4 CTCPs */
 
 	flag = check_ignore_channel(from, FromUserHost, to, IGNORE_CTCPS);
-	fflag = new_check_flooding(from, is_channel(to) ? to : NULL,
+	fflag = new_check_flooding(from, FromUserHost, is_channel(to) ? to : NULL,
 						str, CTCP_FLOOD); 
 
 	in_ctcp_flag++;
-	strmcpy(local_ctcp_buffer, str, IRCD_BUFFER_SIZE - 2);
+	strlcpy(local_ctcp_buffer, str, sizeof(local_ctcp_buffer) - 2);
 
 	lastlog_level = set_lastlog_msg_level(LOG_CTCP);
-	for (;;strmcat(local_ctcp_buffer, last, BIG_BUFFER_SIZE - 2))
+	for (;;strlcat(local_ctcp_buffer, last, sizeof(local_ctcp_buffer) - 2))
 	{
 		if (split_CTCP(local_ctcp_buffer, the_ctcp, last))
 			break;		/* All done! */
@@ -587,7 +646,7 @@ static	time_t	last_ctcp_parsed = 0;
 		 * Check to see if the user is ignoring person.
 		 * Or if we're suppressing a flood.
 		 */
-		if (flag == IGNORED || fflag == 0)
+		if (flag == IGNORED || fflag == 1)
 		{
 			if (x_debug & DEBUG_CTCPS)
 				yell("CTCP from [%s] ignored", from);
@@ -625,7 +684,7 @@ static	time_t	last_ctcp_parsed = 0;
 
 
 		/*
-		 * Now its ok to parse the CTCP.  
+		 * Now its ok to parse the CTCP.
 		 * First we remove the argument.
 		 */
 		ctcp_command = the_ctcp;
@@ -705,7 +764,7 @@ static	time_t	last_ctcp_parsed = 0;
 		 * If its an ``INLINE'' CTCP, we paste it back in.
 		 */
 		if (ctcp_cmd[i].flag & CTCP_INLINE)
-			strmcat(local_ctcp_buffer, ptr, BIG_BUFFER_SIZE);
+			strlcat(local_ctcp_buffer, ptr, sizeof local_ctcp_buffer);
 
 		/* 
 		 * If its ``INTERESTING'', tell the user.
@@ -743,11 +802,10 @@ static	time_t	last_ctcp_parsed = 0;
 	in_ctcp_flag--;
 
 	/* 
-	 * Since 'local_ctcp_buffer' is derived from 'str',
-	 * strlen(local_ctcp_buffer) <= strlen(str) in all circumstances,
-	 * so this strcpy is safe.
+	 * 'str' is required to be BIG_BUFFER_SIZE + 1 or bigger per the API.
 	 */
-	return strcpy(str, local_ctcp_buffer);
+	strlcpy(str, local_ctcp_buffer, BIG_BUFFER_SIZE);
+	return str;
 }
 
 
@@ -756,7 +814,7 @@ static	time_t	last_ctcp_parsed = 0;
  * do_notice_ctcp: a re-entrant form of a CTCP reply parser.
  * See the implementation notes in do_ctcp().
  */
-char *	do_notice_ctcp (char *from, char *to, char *str)
+char *	do_notice_ctcp (const char *from, const char *to, char *str)
 {
 	int 	flag;
 	int	lastlog_level;
@@ -779,9 +837,9 @@ char *	do_notice_ctcp (char *from, char *to, char *str)
 	flag = check_ignore_channel(from, FromUserHost, to, IGNORE_CTCPS);
 	if (!in_ctcp_flag)
 		in_ctcp_flag = -1;
-	strmcpy(local_ctcp_buffer, str, IRCD_BUFFER_SIZE - 2);
+	strlcpy(local_ctcp_buffer, str, IRCD_BUFFER_SIZE - 2);
 
-	for (;;strmcat(local_ctcp_buffer, last, BIG_BUFFER_SIZE))
+	for (;;strlcat(local_ctcp_buffer, last, sizeof local_ctcp_buffer))
 	{
 		if (split_CTCP(local_ctcp_buffer, the_ctcp, last))
 			break;		/* All done! */
@@ -840,24 +898,22 @@ char *	do_notice_ctcp (char *from, char *to, char *str)
 		{
 			if ((ptr = ctcp_cmd[i].repl(ctcp_cmd + i, from, to, ctcp_argument)))
 			{
-				strmcat(local_ctcp_buffer, ptr, BIG_BUFFER_SIZE);
+				strlcat(local_ctcp_buffer, ptr, sizeof local_ctcp_buffer);
 				new_free(&ptr);
 				continue;
 			}
 		}
 
-		/* Toss it at the user.  */
-		if (do_hook(CTCP_REPLY_LIST, "%s %s %s", from, ctcp_command, ctcp_argument))
-		{
-			/* Set up the window level/logging */
-			lastlog_level = set_lastlog_msg_level(LOG_CTCP);
-			message_from(NULL, LOG_CTCP);
+		/* Set up the window level/logging */
+		lastlog_level = set_lastlog_msg_level(LOG_CTCP);
+		message_from(NULL, LOG_CTCP);
 
+		/* Toss it at the user.  */
+		if (do_hook(CTCP_REPLY_LIST, "%s %s %s %s", from, to, ctcp_command, ctcp_argument))
 			say("CTCP %s reply from %s: %s", ctcp_command, from, ctcp_argument);
 
-			/* Reset the window level/logging */
-			set_lastlog_msg_level(lastlog_level);
-		}
+		/* Reset the window level/logging */
+		set_lastlog_msg_level(lastlog_level);
 
 		if (!(ctcp_cmd[i].flag & CTCP_NOLIMIT))
 			allow_ctcp_reply = 0;
@@ -870,7 +926,8 @@ char *	do_notice_ctcp (char *from, char *to, char *str)
 	 * local_ctcp_buffer is derived from 'str', so its always
 	 * smaller or equal in size to 'str', so this copy is safe.
 	 */
-	return strcpy(str, local_ctcp_buffer);
+	strlcpy(str, local_ctcp_buffer, BIG_BUFFER_SIZE);
+	return str;
 }
 
 
@@ -890,14 +947,15 @@ int	in_ctcp (void) { return (in_ctcp_flag); }
  * transparantly.  This greatly reduces the logic, complexity, and
  * possibility for error in this function.
  */
-void	send_ctcp (int type, char *to, int datatag, char *format, ...)
+void	send_ctcp (int type, const char *to, int datatag, const char *format, ...)
 {
 	char 	putbuf [BIG_BUFFER_SIZE + 1],
 		*putbuf2;
 	int	len;
 
 	/* Make sure that the final \001 doesnt get truncated */
-	len = IRCD_BUFFER_SIZE - (12 + strlen(to));
+	if ((len = IRCD_BUFFER_SIZE - (12 + strlen(to))) <= 0)
+		return;				/* Whatever. */
 	putbuf2 = alloca(len);
 
 	if (format)
@@ -934,99 +992,6 @@ void	send_ctcp (int type, char *to, int datatag, char *format, ...)
 }
 
 
-/*
- * quote_it: This quotes the given string making it sendable via irc.  A
- * pointer to the length of the data is required and the data need not be
- * null terminated (it can contain nulls).  Returned is a malloced, null
- * terminated string.   
- */
-char	*ctcp_quote_it (char *str, int len)
-{
-	char	buffer[BIG_BUFFER_SIZE + 1];
-	char	*ptr;
-	int	i;
-
-	ptr = buffer;
-	for (i = 0; i < len; i++)
-	{
-		switch (str[i])
-		{
-			case CTCP_DELIM_CHAR:	*ptr++ = CTCP_QUOTE_CHAR;
-						*ptr++ = 'a';
-						break;
-			case '\n':		*ptr++ = CTCP_QUOTE_CHAR;
-						*ptr++ = 'n';
-						break;
-			case '\r':		*ptr++ = CTCP_QUOTE_CHAR;
-						*ptr++ = 'r';
-						break;
-			case CTCP_QUOTE_CHAR:	*ptr++ = CTCP_QUOTE_CHAR;
-						*ptr++ = CTCP_QUOTE_CHAR;
-						break;
-			case '\0':		*ptr++ = CTCP_QUOTE_CHAR;
-						*ptr++ = '0';
-						break;
-			default:		*ptr++ = str[i];
-						break;
-		}
-	}
-	*ptr = '\0';
-	return m_strdup(buffer);
-}
-
-/*
- * ctcp_unquote_it: This takes a null terminated string that had previously
- * been quoted using ctcp_quote_it and unquotes it.  Returned is a malloced
- * space pointing to the unquoted string.  NOTE: a trailing null is added for
- * convenied, but the returned data may contain nulls!.  The len is modified
- * to contain the size of the data returned. 
- */
-char	*ctcp_unquote_it (char *str, int *len)
-{
-	char	*buffer;
-	char	*ptr;
-	char	c;
-	int	i,
-		new_size = 0;
-
-	buffer = (char *) new_malloc(sizeof(char) * *len + 1);
-	ptr = buffer;
-	i = 0;
-	while (i < *len)
-	{
-		if ((c = str[i++]) == CTCP_QUOTE_CHAR)
-		{
-			switch (c = str[i++])
-			{
-				case CTCP_QUOTE_CHAR:
-					*ptr++ = CTCP_QUOTE_CHAR;
-					break;
-				case 'a':
-					*ptr++ = CTCP_DELIM_CHAR;
-					break;
-				case 'n':
-					*ptr++ = '\n';
-					break;
-				case 'r':
-					*ptr++ = '\r';
-					break;
-				case '0':
-					*ptr++ = '\0';
-					break;
-				default:
-					*ptr++ = c;
-					break;
-			}
-		}
-		else
-			*ptr++ = c;
-		new_size++;
-	}
-	*ptr = '\0';
-	*len = new_size;
-	return (buffer);
-}
-
 int 	get_ctcp_val (char *str)
 {
 	int i;
@@ -1071,7 +1036,7 @@ static int split_CTCP (char *raw_message, char *ctcp_dest, char *after_ctcp)
 	}
 	*ctcp_end++ = 0;
 
-	strmcpy(ctcp_dest, ctcp_start, IRCD_BUFFER_SIZE - 2);
-	strmcpy(after_ctcp, ctcp_end, IRCD_BUFFER_SIZE - 2);
+	strlcpy(ctcp_dest, ctcp_start, IRCD_BUFFER_SIZE - 1);
+	strlcpy(after_ctcp, ctcp_end, IRCD_BUFFER_SIZE - 1);
 	return 0;		/* All done! */
 }
