@@ -1,16 +1,44 @@
+/* $EPIC: expr.c,v 1.23 2004/01/31 22:15:01 jnelson Exp $ */
 /*
  * expr.c -- The expression mode parser and the textual mode parser
  * #included by alias.c -- DO NOT DELETE
  *
- * Copyright 1990 Michael Sandrof
- * Copyright 1997 EPIC Software Labs
- * See the COPYRIGHT file for more info
+ * Copyright (c) 1990 Michael Sandroff.
+ * Copyright (c) 1991, 1992 Troy Rollo.
+ * Copyright (c) 1992-1996 Matthew Green.
+ * Copyright  1993, 2003 EPIC Software Labs
+ * All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notices, the above paragraph (the one permitting redistribution),
+ *    this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ * 3. The names of the author(s) may not be used to endorse or promote
+ *    products derived from this software without specific prior written
+ *    permission.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
+ * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
+ * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
+ * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
+ * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
+ * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+ * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
+ * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
+ * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
+ * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ * SUCH DAMAGE.
  */
 
 #include <math.h>
 
 /* Function decls */
-static	void	TruncateAndQuote(char **, const char *, int, const char *);
+static	void	TruncateAndQuote (char **, const char *, ssize_t, const char *);
 static	char	*alias_special_char(char **, char *, const char *, char *, int *);
 static	void	do_alias_string (char *, char *);
 
@@ -19,7 +47,7 @@ char *alias_string = NULL;
 /************************** EXPRESSION MODE PARSER ***********************/
 /* canon_number: canonicalizes number to something relevant */
 /* If FLOATING_POINT_MATH isnt set, it truncates it to an integer */
-static char *canon_number (char *input)
+char *canon_number (char *input)
 {
 	int end = strlen(input);
 
@@ -43,9 +71,9 @@ static char *canon_number (char *input)
 	}
 	else
 	{
-		char *dot = strchr(input, '.');
-		if (dot)
-			*dot = 0;
+		char *dotloc = strchr(input, '.');
+		if (dotloc)
+			*dotloc = 0;
 	}
 
 	return input;
@@ -111,7 +139,7 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 		*ptr2,			/* used to point matching brackets */
 		*right,			/* used to denote end of bracket-set */
 		*lastc,			/* used to denote end of token-set */
-		*tmp = NULL,		/* place to sprintf() into */
+		*tmp = NULL,		/* place to malloc_sprintf() into */
 		op;			/* the op were working on */
 	int	got_sloshed = 0,	/* If the last char was a slash */
 		display;
@@ -243,7 +271,7 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 
 	/* If there's nothing there, return it */
 	if (!*str)
-		return m_strdup(empty_string);
+		return malloc_strdup(empty_string);
 
 
 	/* find the end of the rest of the expression */
@@ -302,7 +330,7 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 		}
 
 		/*
-		 * Parentheses have two contexts:  
+		 * Parentheses have two contexts:
 		 * 1) (text) is a unary precedence operator.  It is nonassoc,
 		 *	and simply parses the insides immediately.
 		 * 2) text(text) is the function operator.  It calls the
@@ -311,6 +339,7 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 		case '(':
 		{
 			int	savec = 0;
+		        ssize_t	span;
 
 			/*
 			 * If we're not in NU_UNIT, then we have a paren-set
@@ -320,15 +349,15 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 			 */
 			if (stage != NU_UNIT || ptr == str)
 			{
-				/* 
-				 * If there is no matching ), gobble up the
-				 * entire expression.
-				 */
-				if (!(ptr2 = MatchingBracket(ptr+1, '(', ')')))
-					ptr = ptr + strlen(ptr) - 1;
-				else
-					ptr = ptr2;
-				break;
+			    /* 
+			     * If there is no matching ), gobble up the
+			     * entire expression.
+			     */
+			    if ((span = MatchingBracket(ptr + 1, '(', ')')) < 0)
+				ptr += strlen(ptr) - 1;
+			    else
+				ptr += 1 + span;
+			    break;
 			}
 
 			/*
@@ -336,16 +365,18 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 			 * function call.  We gobble up the arguments and
 			 * ship it all off to call_function.
 			 */
-			if ((ptr2 = MatchingBracket(ptr + 1, '(', ')')))
+			if ((span = MatchingBracket(ptr + 1, '(', ')')) >= 0)
 			{
-				ptr2++;
+				ptr2 = ptr + 1 + span + 1;
 				savec = *ptr2;
 				*ptr2 = 0;
 			}
+			else
+				ptr2 = NULL;
 
 			result1 = call_function(str, args, arg_flag);
 
-			if (savec)
+			if (ptr2 && savec)
 				*ptr2 = savec;
 
 			/*
@@ -385,11 +416,14 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 		 */
 		case '{':
 		{
+			ssize_t span;
+
 			display = window_display;
 
-			ptr2 = MatchingBracket(ptr + 1, LEFT_BRACE, RIGHT_BRACE);
-			if (!ptr2)
-				ptr2 = ptr + strlen(ptr) - 1;
+			if ((span = MatchingBracket(ptr + 1, '{', '}')) >= 0)
+				ptr2 = ptr + 1 + span;
+			else
+				ptr2 = ptr + strlen(ptr);
 
 			if (stage != NU_UNIT)
 			{
@@ -401,20 +435,8 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 			*ptr++ = 0;
 
 			/* Taken more or less from call_user_function */
-			make_local_stack(NULL);
-			window_display = 0;
-			add_local_alias("FUNCTION_RETURN", empty_string, 0);
-			window_display = display;
-
-			will_catch_return_exceptions++;
-			parse_line(NULL, ptr, args, 0, 1);
-			will_catch_return_exceptions--;
-			return_exception = 0;
-
-			result1 = get_variable("FUNCTION_RETURN");
-			destroy_local_stack();
-			if (!result1)
-				result1 = m_strdup(empty_string);
+			if (!(result1 = call_lambda_function(NULL, ptr, args)))
+				result1 = malloc_strdup(empty_string);
 
 			return result1;
 		}
@@ -432,26 +454,43 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 		 */
 		case '[':
 		{
+			ssize_t	span;
+
+			/* If we are not parsing this yet, skip it */
 			if (stage != NU_UNIT)
 			{
-				if (!(ptr2 = MatchingBracket(ptr+1, LEFT_BRACKET, RIGHT_BRACKET)))
-					ptr = ptr+strlen(ptr)-1;
-				else
-					ptr = ptr2;
-				break;
+			    char *p;
+
+			    if ((span = MatchingBracket(ptr+1, '[', ']')) >= 0)
+				p = ptr + 1 + span;
+			    else if ((p = strchr(ptr + 1, ']')) == NULL)
+				p = ptr + strlen(ptr) - 1;
+
+			    ptr = p;
+			    break;
 			}
 
-			/* ptr points right after the [] set */
+			/*
+			 * 'ptr' points to the opening [.
+			 * We want 'right' to point to the first character
+			 * inside the [...] set.
+			 */
 			*ptr++ = 0;
 			right = ptr;
 	
 			/*
-			 * At this point, we check to see if it really is a
-			 * '[', and if it is, we skip over it.
+			 * Now we look for the matching ']'.  If we don't
+			 * find it, then look for the first ']'.  If we don't
+			 * find *that*, then (ptr == NULL) will slurp up the
+			 * whole rest of the expression, so there.
 			 */
-			if ((ptr = MatchingBracket(right, 
-					LEFT_BRACKET, RIGHT_BRACKET)))
-				*ptr++ = '\0';
+			if ((span = MatchingBracket(right, '[', ']')) >= 0)
+				ptr = right + span;
+			else 
+				ptr = strchr(right, ']');
+
+			if (ptr && *ptr)
+				*ptr++ = 0;
 
 #ifndef NO_CHEATING
 			/*
@@ -464,7 +503,7 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 			{
 			    if (right[1] == '*' && right[2] == 0)
 			    {
-				result1 = m_strdup(args);
+				result1 = malloc_strdup(args);
 				*arg_flag = 1;
 			    }
 			    else
@@ -501,7 +540,7 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 			    }
 			}
 			else if (!strchr(right, '$') && !strchr(right, '\\'))
-				result1 = m_strdup(right);
+				result1 = malloc_strdup(right);
 			else
 #endif
 				result1 = expand_alias(right, args, arg_flag, NULL);
@@ -525,11 +564,8 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 						strlen(result1) + 
 						(ptr ? strlen(ptr) : 0) + 2;
 
-				/* Obviously, these are safe. */
 				result2 = alloca(size);
-				strcpy(result2, str);
-				strcat(result2, ".");
-				strcat(result2, result1);
+				snprintf(result2, size, "%s.%s", str, result1);
 				new_free(&result1);
 
 				/*
@@ -541,8 +577,7 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 				 */
 				if (ptr && *ptr)
 				{
-					/* Obviously, this is safe. */
-					strcat(result2, ptr);
+					strlcat(result2, ptr, size);
 					result1 = next_unit(result2, args, arg_flag, stage);
 				}
 				else
@@ -634,7 +669,7 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 
 				new_free(&result1);
 				new_free(&varname);
-				return m_strdup(ltoa(r));
+				return malloc_strdup(ltoa(r));
 			}
 
 			/* Unary op is ignored */
@@ -650,8 +685,12 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 			else
 				dvalue3 = dvalue1 + dvalue2;
 
-			tmp = m_sprintf("%f", dvalue3);
+#if 0
+			tmp = malloc_sprintf(NULL, "%.50g", dvalue3);
 			canon_number(tmp);
+#else
+			tmp = malloc_strdup(ftoa(dvalue3));
+#endif
 #else
 			SETUP_INTEGER_OPERATION(NU_ADD)
 
@@ -660,7 +699,7 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 			else
 				value3 = value1 + value2;
 
-			tmp = m_strdup(ltoa(value3));
+			tmp = malloc_strdup(ltoa(value3));
 #endif
 			CLEANUP_IMPLIED()
 			return tmp;
@@ -691,7 +730,11 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 			{
 				*ptr++ = '\0';
 				SETUP_BINARY(dvalue1, dvalue2, atof)
-				return m_sprintf("%f", pow(dvalue1, dvalue2));
+#if 0
+				return malloc_sprintf(NULL, "%.50g", pow(dvalue1, dvalue2));
+#else
+				return malloc_strdup(ftoa(pow(dvalue1, dvalue2)));
+#endif
 			}
 
 			SETUP_FLOAT_OPERATION(NU_MULT)
@@ -709,8 +752,12 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 					dvalue3 = (int)dvalue1 % (int)dvalue2;
 			}
 
-			tmp = m_sprintf("%f", dvalue3);
+#if 0
+			tmp = malloc_sprintf(NULL, "%.50g", dvalue3);
 			canon_number(tmp);
+#else
+			tmp = malloc_strdup(ftoa(dvalue3));
+#endif
 			CLEANUP_IMPLIED()
 			return tmp;
 		}
@@ -739,7 +786,7 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 			{
 				char *sval1, *sval2;
 
-				SETUP_IMPLIED(sval1, sval2, m_strdup)
+				SETUP_IMPLIED(sval1, sval2, malloc_strdup)
 				malloc_strcat(&sval1, sval2);
 				new_free(&sval2);
 				tmp = sval1;
@@ -752,7 +799,7 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 				char *sval1, *sval2;
 
 				ptr[1] = '=';	/* XXXX */
-				SETUP_IMPLIED(sval1, sval2, m_strdup)
+				SETUP_IMPLIED(sval1, sval2, malloc_strdup)
 				malloc_strcat(&sval2, sval1);
 				new_free(&sval1);
 				tmp = sval2;
@@ -795,7 +842,7 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 
 				new_free(&result1);
 				new_free(&result2);
-				return m_strdup(value3 ? one : zero);
+				return malloc_strdup(value3 ? one : zero);
 			}
 
 			/* &= is implied binary bitwise and */
@@ -803,7 +850,7 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 			{
 				SETUP_IMPLIED(value1, value2, my_atol)
 				value1 &= value2;
-				tmp = m_strdup(ltoa(value1));
+				tmp = malloc_strdup(ltoa(value1));
 				CLEANUP_IMPLIED();
 				return tmp;
 			}
@@ -812,7 +859,7 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 			else if (ptr[1] != ptr[0] && ptr[1] != '=' && stage == NU_BITW)
 			{
 				SETUP_BINARY(value1, value2, my_atol)
-				return m_strdup(ltoa(value1 & value2));
+				return malloc_strdup(ltoa(value1 & value2));
 			}
 
 			else
@@ -842,7 +889,7 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 
 				new_free(&result1);
 				new_free(&result2);
-				return m_strdup(value3 ? one : zero);
+				return malloc_strdup(value3 ? one : zero);
 			}
 
 			/* |= is implied binary bitwise or */
@@ -850,7 +897,7 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 			{
 				SETUP_IMPLIED(value1, value2, my_atol)
 				value1 |= value2;
-				tmp = m_strdup(ltoa(value1));
+				tmp = malloc_strdup(ltoa(value1));
 				CLEANUP_IMPLIED();
 				return tmp;
 			}
@@ -859,7 +906,7 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 			else if (ptr[1] != ptr[0] && ptr[1] != '=' && stage != NU_BITW)
 			{
 				SETUP_BINARY(value1, value2, my_atol)
-				return m_strdup(ltoa(value1 | value2));
+				return malloc_strdup(ltoa(value1 | value2));
 			}
 
 			else
@@ -882,7 +929,7 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 				new_free(&result1);
 				new_free(&result2);
 
-				return m_strdup(value1 ^ value2 ? one : zero);
+				return malloc_strdup(value1 ^ value2 ? one : zero);
 			}
 
 			/* ^= is implied binary bitwise xor */
@@ -891,7 +938,7 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 
 				SETUP_IMPLIED(value1, value2, my_atol)
 				value1 ^= value2;
-				tmp = m_strdup(ltoa(value1));
+				tmp = malloc_strdup(ltoa(value1));
 				CLEANUP_IMPLIED();
 				return tmp;
 			}
@@ -900,7 +947,7 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 			else if (ptr[1] != ptr[0] && ptr[1] != '=' && stage == NU_BITW)
 			{
 				SETUP_BINARY(value1, value2, my_atol)
-				return m_strdup(ltoa(value1 ^ value2));
+				return malloc_strdup(ltoa(value1 ^ value2));
 			}
 
 			else
@@ -915,35 +962,40 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 		 */
 		case '?':
 		{
-			if (stage == NU_TERT)
-			{
-				*ptr++ = '\0';
-				result1 = next_unit(str, args, arg_flag, stage);
-				ptr2 = MatchingBracket(ptr, '?', ':');
+		    if (stage == NU_TERT)
+		    {
+			ssize_t span;
 
-				/* Unbalanced :, or possibly missing */
-				if (!ptr2)  /* ? but no :, ignore */
-				{
-					ptr = lastop(ptr);
-					break;
-				}
-				*ptr2++ = '\0';
-				if ( check_val(result1) )
-					result2 = next_unit(ptr, args, arg_flag, stage);
-				else
-					result2 = next_unit(ptr2, args, arg_flag, stage);
+			*ptr++ = 0;
+			result1 = next_unit(str, args, arg_flag, stage);
+			span = MatchingBracket(ptr, '?', ':');
 
-				/* XXXX - needed? */
-				ptr2[-1] = ':';
-				new_free(&result1);
-				return result2;
-			}
-
-			else
+			/* Unbalanced :, or possibly missing */
+			if (span < 0)	/* ? but no :, ignore */
 			{
 				ptr = lastop(ptr);
 				break;
 			}
+
+			ptr2 = ptr + span;
+			*ptr2++ = 0;
+
+			if ( check_val(result1) )
+				result2 = next_unit(ptr, args, arg_flag, stage);
+			else
+				result2 = next_unit(ptr2, args, arg_flag, stage);
+
+			/* XXXX - needed? */
+			ptr2[-1] = ':';
+			new_free(&result1);
+			return result2;
+		    }
+
+		    else
+		    {
+			ptr = lastop(ptr);
+			break;
+		    }
 		}
 
 		/*
@@ -1036,7 +1088,7 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 
 				new_free(&result1);
 				new_free(&result2);
-				return m_strdup(ltoa(value3));
+				return malloc_strdup(ltoa(value3));
 			}
 
 			else if (ptr[1] != ptr[0] && stage == NU_COMP)
@@ -1071,7 +1123,7 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 
 				new_free(&result1);
 				new_free(&result2);
-				return m_strdup(ltoa(value2));
+				return malloc_strdup(ltoa(value2));
 			}
 
 			else
@@ -1094,7 +1146,7 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 				else
 					value1 = 0;
 
-				return m_strdup(ltoa(value1));
+				return malloc_strdup(ltoa(value1));
 			}
 
 			else
@@ -1123,7 +1175,7 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 					value2 = ((*result1)?0:1);
 
 				new_free(&result1);
-				return m_strdup(ltoa(value2));
+				return malloc_strdup(ltoa(value2));
 			}
 
 			else if (ptr != str && ptr[1] == '~' && stage == NU_COMP)
@@ -1202,7 +1254,7 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 	 * If the result is a number, return it.
 	 */
 	if (my_isdigit(str))
-		return m_strdup(str);
+		return malloc_strdup(str);
 
 
 	/*
@@ -1217,21 +1269,21 @@ static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 	 * Its not a number, so its a variable, look it up.
 	 */
 	if (!*str)
-		result1 = m_strdup(args);
+		result1 = malloc_strdup(args);
 	else if (!(result1 = get_variable(str)))
-		return m_strdup(empty_string);
+		return malloc_strdup(empty_string);
 
 	/*
-	 * See if we have to take strlen or word_count on the variable.
+	 * See if we have to take strlen or count_words on the variable.
 	 */
 	if (op)
 	{
 		if (op == '#')
-			value1 = word_count(result1);
+			value1 = count_words(result1, DWORD_YES, "\"");
 		else if (op == '@')
 			value1 = strlen(result1);
 		new_free(&result1);
-		return m_strdup(ltoa(value1));
+		return malloc_strdup(ltoa(value1));
 	}
 
 	/*
@@ -1269,7 +1321,7 @@ char	*parse_inline (char *str, const char *args, int *args_flag)
  *		portion is written back into more_text.
  *	Backslash escapes are unescaped.
  */
-char	*expand_alias	(const char *string, const char *args, int *args_flag, char **more_text)
+char	*expand_alias	(const char *string, const char *args, int *args_flag, ssize_t *more_text)
 {
 	char	*buffer = NULL,
 		*ptr,
@@ -1279,14 +1331,15 @@ char	*expand_alias	(const char *string, const char *args, int *args_flag, char *
 	char	quote_temp[2];
 	char	ch;
 	int	is_quote = 0;
-	int	unescape = 1;
+	char *	unescape = empty_string;
+	size_t	buffclue = 0;
 
 	if (!string || !*string)
-		return m_strdup(empty_string);
+		return malloc_strdup(empty_string);
 
 	if (*string == '@' && more_text)
 	{
-		unescape = 0;
+		unescape = NULL;
 		*args_flag = 1; /* Stop the @ command from auto appending */
 	}
 	quote_temp[1] = 0;
@@ -1296,7 +1349,7 @@ char	*expand_alias	(const char *string, const char *args, int *args_flag, char *
 
 	ptr = stuff;
 	if (more_text)
-		*more_text = NULL;
+		*more_text = -1;
 
 	while (ptr && *ptr)
 	{
@@ -1311,6 +1364,7 @@ char	*expand_alias	(const char *string, const char *args, int *args_flag, char *
 		{
 		    case '$':
 		    {
+			char *buffer1 = NULL;
 			/*
 			 * The test here ensures that if we are in the 
 			 * expression evaluation command, we don't expand $. 
@@ -1325,10 +1379,15 @@ char	*expand_alias	(const char *string, const char *args, int *args_flag, char *
 				break;
 			}
 			*ptr++ = 0;
-			if (!*ptr)
-				break;		/* Hrm. */
 
-			m_strcat_ues(&buffer, stuff, unescape);
+			/* Just a lonely old $ at the end of string. */
+			if (!*ptr)
+			{
+				malloc_strcat_c(&buffer, empty_string, &buffclue);
+				break;		/* Hrm. */
+			}
+
+			malloc_strcat_ues_c(&buffer, stuff, unescape, &buffclue);
 
 			for (; *ptr == '^'; ptr++)
 			{
@@ -1338,7 +1397,9 @@ char	*expand_alias	(const char *string, const char *args, int *args_flag, char *
 				quote_temp[0] = *ptr;
 				malloc_strcat(&quote_str, quote_temp);
 			}
-			stuff = alias_special_char(&buffer, ptr, args, quote_str, args_flag);
+			stuff = alias_special_char(&buffer1, ptr, args, quote_str, args_flag);
+			malloc_strcat_c(&buffer, buffer1, &buffclue);
+			new_free(&buffer1);
 			if (quote_str)		/* Why ``stuff''? */
 				new_free(&quote_str);
 			ptr = stuff;
@@ -1352,33 +1413,40 @@ char	*expand_alias	(const char *string, const char *args, int *args_flag, char *
 				ptr++;
 				break;
 			}
-			*more_text = (char *)(string + (ptr - free_stuff) + 1);
-			*ptr = '\0'; /* To terminate the loop */
+			*more_text = (ptr - free_stuff) + 1;
+			*ptr = 0; /* To terminate the loop */
 			break;
 		    }
 
 		    case LEFT_PAREN:
 		    case LEFT_BRACE:
 		    {
+			ssize_t	span;
+
 			ch = *ptr;
-			*ptr = '\0';
-			m_strcat_ues(&buffer, stuff, unescape);
+			*ptr = 0;
+			malloc_strcat_ues_c(&buffer, stuff, unescape, &buffclue);
 			stuff = ptr;
+
 			*args_flag = 1;
-			if (!(ptr = MatchingBracket(stuff + 1, ch,
+			if ((span = MatchingBracket(stuff + 1, ch, 
 					(ch == LEFT_PAREN) ?
-					RIGHT_PAREN : RIGHT_BRACE)))
+					RIGHT_PAREN : RIGHT_BRACE)) < 0)
 			{
 				put_it("Unmatched %c", ch);
-				ptr = stuff + strlen(stuff+1)+1;
+				/* 
+				 * DO NOT ``OPTIMIZE'' THIS BECAUSE
+				 * *STUFF IS NUL SO STRLEN(STUFF) IS 0!
+				 */
+				ptr = stuff + 1 + strlen(stuff + 1);
 			}
 			else
-				ptr++;
+				ptr = stuff + 1 + span + 1;
 
 			*stuff = ch;
 			ch = *ptr;
-			*ptr = '\0';
-			malloc_strcat(&buffer, stuff);
+			*ptr = 0;
+			malloc_strcat_c(&buffer, stuff, &buffclue);
 			stuff = ptr;
 			*ptr = ch;
 			break;
@@ -1398,10 +1466,15 @@ char	*expand_alias	(const char *string, const char *args, int *args_flag, char *
 	}
 
 	if (stuff)
-		m_strcat_ues(&buffer, stuff, unescape);
+		malloc_strcat_ues_c(&buffer, stuff, unescape, &buffclue);
 
 	if (get_int_var(DEBUG_VAR) & DEBUG_EXPANSIONS)
-		yell("Expanded [%s] to [%s]", string, buffer);
+		privileged_yell("Expanded [%s] to [%s]", string, buffer);
+
+#if 0					/* Maybe a good idea later? */
+	if (!buffer && !more_text)
+		panic("expand_alias [%s] returning NULL!", string);
+#endif
 
 	return buffer;
 }
@@ -1419,21 +1492,22 @@ char	*expand_alias	(const char *string, const char *args, int *args_flag, char *
  * The args_flag is set to 1 if any of the $n, $n-, $n-m, $-m, $*, or $() 
  * is used in the alias.  Otherwise it is left unchanged.
  */
-static	char	*alias_special_char(char **buffer, char *ptr, const char *args, char *quote_em, int *args_flag)
+static	char	*alias_special_char (char **buffer, char *ptr, const char *args, char *quote_em, int *args_flag)
 {
 	char	*tmp,
-		*tmp2,
 		c;
-	int	upper,
-		lower,
-		length;
+	int	my_upper,
+		my_lower;
+	ssize_t	length;
+	ssize_t	span;
 
 	length = 0;
 	if ((c = *ptr) == LEFT_BRACKET)
 	{
 		ptr++;
-		if ((tmp = MatchingBracket(ptr, '[', ']')))
+		if ((span = MatchingBracket(ptr, '[', ']')) >= 0)
 		{
+			tmp = ptr + span;
 			*tmp++ = 0;
 			if (*ptr == '$')
 			{
@@ -1468,7 +1542,7 @@ static	char	*alias_special_char(char **buffer, char *ptr, const char *args, char
 		 */
 		case 0:
 		{
-			tmp = m_strdup(empty_string);
+			tmp = malloc_strdup(empty_string);
 			TruncateAndQuote(buffer, tmp, length, quote_em);
 			new_free(&tmp);
 			return ptr;
@@ -1499,8 +1573,12 @@ static	char	*alias_special_char(char **buffer, char *ptr, const char *args, char
 				*tmpsav = NULL,
 				*ph = ptr + 1;
 
-			if ((ptr = MatchingBracket(ph, '(', ')')) || 
-	/* ( */		    (ptr = strchr(ph, ')')))
+			if ((span = MatchingBracket(ph, '(', ')')) >= 0)
+				ptr = ph + span;
+			else
+				ptr = strchr(ph, ')');
+
+			if (ptr)
 				*ptr++ = 0;
 			else
 				yell("Unmatched ( (continuing anyways)");
@@ -1515,14 +1593,14 @@ static	char	*alias_special_char(char **buffer, char *ptr, const char *args, char
 					new_free(&tmpsav);
 				tmpsav = tmp = tmp2;
 			}
-			while (*tmp == '$');
+			while (tmp && *tmp == '$');
 
 			alias_special_char(&sub_buffer, tmp, args, 
 						quote_em, args_flag);
 
 			/* Some kind of bogus expando */
 			if (sub_buffer == NULL)
-				sub_buffer = m_strdup(empty_string);
+				sub_buffer = malloc_strdup(empty_string);
 
 			if (!(x_debug & DEBUG_SLASH_HACK))
 				TruncateAndQuote(buffer, sub_buffer, 
@@ -1562,11 +1640,15 @@ static	char	*alias_special_char(char **buffer, char *ptr, const char *args, char
 			char	*ph = ptr + 1;
 
 			/* 
-			 * BLAH. This didnt allow for nesting before.  
+			 * BLAH. This didnt allow for nesting before.
 			 * How lame.
 			 */
-			if ((ptr = MatchingBracket(ph, '{', '}')) ||
-			    (ptr = strchr(ph, '}')))
+			if ((span = MatchingBracket(ph, '{', '}')) >= 0)
+				ptr = ph + span;
+			else
+				ptr = strchr(ph, '}');
+
+			if (ptr)
 				*ptr++ = 0;
 			else
 				yell("Unmatched { (continuing anyways)");
@@ -1640,12 +1722,12 @@ static	char	*alias_special_char(char **buffer, char *ptr, const char *args, char
 			char 	c2 = 0;
 			char 	*sub_buffer = NULL;
 			char 	*rest, *val;
-			int	dummy;
+			int	my_dummy;
 
-			rest = after_expando(ptr + 1, 0, &dummy);
+			rest = (char *)after_expando(ptr + 1, 0, &my_dummy);
 			if (rest == ptr + 1)
 			{
-			    sub_buffer = m_strdup(args);
+			    sub_buffer = malloc_strdup(args);
 			    *args_flag = 1;
 			}
 			else
@@ -1658,9 +1740,9 @@ static	char	*alias_special_char(char **buffer, char *ptr, const char *args, char
 			}
 
 			if (c == '#')
-			    val = m_strdup(ltoa(word_count(sub_buffer)));
+			    val = malloc_strdup(ltoa(count_words(sub_buffer, DWORD_YES, "\"")));
 			else
-			    val = m_strdup(ltoa(strlen(sub_buffer)));
+			    val = malloc_strdup(ltoa(strlen(sub_buffer)));
 
 #if 0	/* Don't need to do this here */
 			if (!(x_debug & DEBUG_SLASH_HACK))
@@ -1696,6 +1778,8 @@ static	char	*alias_special_char(char **buffer, char *ptr, const char *args, char
 			 */
 			if (isdigit(c) || (c == '-') || c == '~')
 			{
+			    char *tmp2;
+
 			    *args_flag = 1;
 
 			    /*
@@ -1704,7 +1788,7 @@ static	char	*alias_special_char(char **buffer, char *ptr, const char *args, char
 			     */
 			    if (c == '~')
 			    {
-				lower = upper = EOS;
+				my_lower = my_upper = EOS;
 				ptr++;
 			    }
 
@@ -1721,10 +1805,10 @@ static	char	*alias_special_char(char **buffer, char *ptr, const char *args, char
 			     */
 			    else if (c == '-')
 			    {
-				lower = SOS;
+				my_lower = SOS;
 				ptr++;
-				upper = parse_number(&ptr);
-				if (upper == -1)
+				my_upper = parse_number(&ptr);
+				if (my_upper == -1)
 				    return empty_string; /* error */
 			    }
 
@@ -1734,16 +1818,16 @@ static	char	*alias_special_char(char **buffer, char *ptr, const char *args, char
 			     */
 			    else
 			    {
-				lower = parse_number(&ptr);
+				my_lower = parse_number(&ptr);
 				if (*ptr == '-')
 				{
 				    ptr++;
-				    upper = parse_number(&ptr);
-				    if (upper == -1)
-					upper = EOS;
+				    my_upper = parse_number(&ptr);
+				    if (my_upper == -1)
+					my_upper = EOS;
 				}
 				else
-				    upper = lower;
+				    my_upper = my_lower;
 			    }
 
 			    /*
@@ -1756,9 +1840,9 @@ static	char	*alias_special_char(char **buffer, char *ptr, const char *args, char
 			     * chewing the expando.
 			     */
 			    if (!args)
-				tmp2 = m_strdup(empty_string);
+				tmp2 = malloc_strdup(empty_string);
 			    else
-				tmp2 = extractw2(args, lower, upper);
+				tmp2 = extractw2(args, my_lower, my_upper);
 
 			    TruncateAndQuote(buffer, tmp2, length, quote_em);
 			    new_free(&tmp2);
@@ -1774,7 +1858,7 @@ static	char	*alias_special_char(char **buffer, char *ptr, const char *args, char
 			    char  *rest, d = 0;
 			    int	  function_call = 0;
 
-			    rest = after_expando(ptr, 0, &function_call);
+			    rest = (char *)after_expando(ptr, 0, &function_call);
 			    if (*rest)
 			    {
 				d = *rest;
@@ -1787,7 +1871,7 @@ static	char	*alias_special_char(char **buffer, char *ptr, const char *args, char
 				tmp = get_variable_with_args(ptr, args, args_flag);
 
 			    if (!tmp)
-				tmp = m_strdup(empty_string);
+				tmp = malloc_strdup(empty_string);
 
 			    TruncateAndQuote(buffer, tmp, length, quote_em);
 
@@ -1811,10 +1895,13 @@ static	char	*alias_special_char(char **buffer, char *ptr, const char *args, char
 
 /*
  * TruncateAndQuote: This handles string width formatting and quoting for irc 
- * variables when [] or ^x is specified.  
+ * variables when [] or ^x is specified.
  */
-static	void	TruncateAndQuote(char **buff, const char *add, int length, const char *quote_em)
+static	void	TruncateAndQuote (char **buff, const char *add, ssize_t length, const char *quote_em)
 {
+	size_t	real_size;
+	char *	buffer;
+
 	if (!add)
 		panic("add is NULL");
 
@@ -1836,14 +1923,15 @@ static	void	TruncateAndQuote(char **buff, const char *add, int length, const cha
 		 * of the function, so yes, we can reference "add" after
 		 * the end of this block.
 		 */
-		char *buffer = alloca(abs(length) + 1);
+		real_size = abs(length) + 1;
+		buffer = alloca(real_size);
 		add = strformat(buffer, add, length, get_int_var(PAD_CHAR_VAR));
 	}
 	if (quote_em && add)	/* add shouldnt ever be NULL, though! */
 	{
 		/* Don't pass retval from "alloca" as func arg, use local. */
-		char *ptr = alloca(strlen(add) * 2 + 2);
-		add = double_quote(add, quote_em, ptr);
+		buffer = alloca(strlen(add) * 2 + 2);
+		add = double_quote(add, quote_em, buffer);
 	}
 
 	if (buff)
