@@ -1,13 +1,37 @@
+/* $EPIC: help.c,v 1.12 2004/02/20 23:40:23 jnelson Exp $ */
 /*
  * help.c: handles the help stuff for irc 
  *
- * Written by Michael Sandrof
- * Extensively modified by Troy Rollo
- * Re-modified by Matthew Green
+ * Copyright (c) 1990 Michael Sandroff.
+ * Copyright (c) 1991, 1992 Troy Rollo.
+ * Copyright (c) 1992-1996 Matthew Green.
+ * Copyright  1995, 2003 EPIC Software Labs.
+ * All rights reserved.
  *
- * Copyright(c) 1992 
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notices, the above paragraph (the one permitting redistribution),
+ *    this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ * 3. The names of the author(s) may not be used to endorse or promote
+ *    products derived from this software without specific prior written
+ *    permission.
  *
- * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
+ * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
+ * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
+ * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
+ * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
+ * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
+ * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+ * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
+ * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
+ * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
+ * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ * SUCH DAMAGE.
  */
 
 /*
@@ -22,11 +46,6 @@
  * It is invalid to call help from more than one screen, at the moment,
  * because there is to much to keep track of - phone, jan 1993.
  */
-
-#if 0
-static char rcsid[] = "@(#)$Id: help.c,v 1.1.1.1 2000/12/05 00:11:57 jnelson Exp $";
-#endif
-
 #include "irc.h"
 #include "help.h"
 #include "input.h"
@@ -38,7 +57,6 @@ static char rcsid[] = "@(#)$Id: help.c,v 1.1.1.1 2000/12/05 00:11:57 jnelson Exp
 #include "term.h"
 #include "vars.h"
 #include "window.h"
-#include <sys/stat.h>
 #include "glob.h"
 
 /* Forward declarations */
@@ -67,7 +85,7 @@ static	char	help_topic_list[BIG_BUFFER_SIZE + 1];
 static	Window *help_window = (Window *) 0;
 static	char	no_help[] = "NOHELP";
 static	char	paused_topic[128];
-static	char *	this_arg;
+static	const char *	this_arg;
 static	int	use_help_window = 0;
 
 
@@ -76,21 +94,25 @@ static	int	use_help_window = 0;
  * thing, depending on the value of HELP_PAGER_VAR.  If it gets to the end,
  * (in either case it will eventally), it closes the file, and returns 0
  * to indicate this.
+ *
+ * -- This doesn't need to fiddle with 'to_window' -- help_put_it does
+ *    all that for us. (hop, 09/23/2003)
  */ 
 static	int	show_help (Window *window, char *name)
 {
-	Window	*old_to_window;
-	int	rows = 0;
+	int	rows = -1;
 	char	line[256];
 
 	if (!help_fp)
 		return 0;
 
-	old_to_window = to_window;
-	to_window = window ? window : current_window;
-
 	if (get_int_var(HELP_PAGER_VAR))
-		rows = to_window->display_size;
+	{
+	    if (help_window)
+		rows = help_window->display_size;
+	    else
+		rows = current_window->display_size;
+	}
 
 	while (rows)
 	{
@@ -98,7 +120,6 @@ static	int	show_help (Window *window, char *name)
 		{
 			fclose(help_fp);
 			help_fp = NULL;
-			to_window = old_to_window;
 			return 0;
 		}
 
@@ -115,7 +136,6 @@ static	int	show_help (Window *window, char *name)
 		rows--;
 	}
 
-	to_window = old_to_window;
 	return (1);
 }
 
@@ -188,7 +208,7 @@ static	void	help_prompt (char *name, char *line)
 		{
 			char	tmp[BIG_BUFFER_SIZE + 1];
 
-			sprintf(tmp, "%s%sHelp? ", help_topic_list,
+			snprintf(tmp, sizeof tmp, "%s%sHelp? ", help_topic_list,
 				*help_topic_list ? " " : empty_string);
 			if (!dumb_mode)
 				add_wait_prompt(tmp, help_me, help_topic_list,
@@ -219,7 +239,7 @@ static	void	help_topic (char *path, char *name)
 		return;
 
 	/* what is the base name? */
-	filename = m_sprintf("%s/%s", path, name);
+	malloc_sprintf(&filename, "%s/%s", path, name);
 	if (filename[strlen(filename)-1] == '/')
 		chop(filename, 1);
 
@@ -244,7 +264,7 @@ static	void	help_topic (char *path, char *name)
  * not show them, until we've seen the whole file, so we called
  * help_show_paused_topic() when we've seen the file, if it is needed.
  */
-static 	void 	help_pause_add_line (char *format, ...)
+static 	void 	help_pause_add_line (const char *format, ...)
 {
 	char	buf[BIG_BUFFER_SIZE];
 	va_list args;
@@ -274,7 +294,11 @@ static	void	help_show_paused_topic (char *name, char *line)
 	if (toupper(*line) == 'Q')
 		i = help_paused_lines + 1;	/* just big enough */
 
-	rows = help_window->display_size;
+	if (get_int_var(HELP_PAGER_VAR))
+		rows = help_window->display_size;
+	else
+		rows = help_paused_lines;	/* Suitably big enough */
+
 	if (i < help_paused_lines)
 	{
 		for (j = 0; j < rows; j++)
@@ -305,7 +329,8 @@ static	void	help_show_paused_topic (char *name, char *line)
 		{
 			char	buf[BIG_BUFFER_SIZE];
 
-			sprintf(buf, "%s%sHelp? ", name, (name && *name) ? " " : empty_string);
+			snprintf(buf, sizeof buf, "%s%sHelp? ", 
+				name, (name && *name) ? " " : empty_string);
 			if (!dumb_mode)
 				add_wait_prompt(buf, help_me, name, WAIT_PROMPT_LINE, 1);
 		}
@@ -331,7 +356,7 @@ static	void	help_me (char *topics, char *args)
 		cnt,
 		i,
 		cols;
-	struct	stat	stat_buf;
+	Stat	stat_buf;
 	char	path[BIG_BUFFER_SIZE+1];
 	int	help_paused_first_call = 0;
 	char *	help_paused_path = (char *) 0;
@@ -341,10 +366,10 @@ static	void	help_me (char *topics, char *args)
 	char	buffer[BIG_BUFFER_SIZE+1];
 	char *	pattern = NULL;
 
-	strcpy(help_topic_list, topics);
+	strlcpy(help_topic_list, topics, sizeof help_topic_list);
 	ptr = get_string_var(HELP_PATH_VAR);
 
-	sprintf(path, "%s/%s", ptr, topics);
+	snprintf(path, sizeof path, "%s/%s", ptr, topics);
 	for (ptr = path; (ptr = strchr(ptr, ' '));)
 		*ptr = '/';
 
@@ -375,7 +400,7 @@ static	void	help_me (char *topics, char *args)
 		else
 			*help_topic_list = '\0';
 
-		sprintf(tmp, "%s%sHelp? ", help_topic_list, *help_topic_list ? " " : empty_string);
+		snprintf(tmp, sizeof tmp, "%s%sHelp? ", help_topic_list, *help_topic_list ? " " : empty_string);
 
 		if (!dumb_mode)
 			add_wait_prompt(tmp, help_me, help_topic_list, WAIT_PROMPT_LINE, 1);
@@ -403,7 +428,6 @@ static	void	help_me (char *topics, char *args)
 	while (this_arg)
 	{
 		entries = 0;
-		message_from((char *) 0, LOG_CURRENT);
 
 		if (!*this_arg)
 			help_topic(path, NULL);
@@ -425,6 +449,7 @@ static	void	help_me (char *topics, char *args)
 		 * Gather up the names of the files in the help directory.
 		 */
 		{
+			size_t size;
 #ifndef HAVE_FCHDIR
 			char 	opath[MAXPATHLEN + 1];
 			getcwd(opath, MAXPATHLEN);
@@ -432,11 +457,11 @@ static	void	help_me (char *topics, char *args)
 			int 	cwd = open(".", O_RDONLY);
 #endif
 
+			size = strlen(path) + 2 + strlen(this_arg) + 3;
+
 			chdir(path);
-			pattern = alloca(strlen(path) + 2 + 
-					 strlen(this_arg) + 3);
-			strcpy(pattern, this_arg);
-			strcat(pattern, "*");
+			pattern = alloca(size);
+			snprintf(pattern, size, "%s*", this_arg);
 #ifdef GLOB_INSENSITIVE
 			bsd_glob(pattern, GLOB_INSENSITIVE, NULL, &g);
 #else
@@ -453,10 +478,10 @@ static	void	help_me (char *topics, char *args)
 
 		for (i = 0; i < g.gl_matchc; i++)
 		{
-			char	*tmp = g.gl_pathv[i];
-			int 	len = strlen(tmp);
+			char	*tmp2 = g.gl_pathv[i];
+			int 	len = strlen(tmp2);
 
-			if (!end_strcmp(tmp, ".gz", 3))
+			if (!end_strcmp(tmp2, ".gz", 3))
 				len -= 3;
 			entry_size = (len > entry_size) ? len : entry_size;
 		}
@@ -471,7 +496,7 @@ static	void	help_me (char *topics, char *args)
 		if (g.gl_matchc > 1)
 		{
 			char *str1 = g.gl_pathv[0];
-			char *str2 = this_arg;
+			const char *str2 = this_arg;
 			int len1 = strlen(str1);
 			int len2 = strlen(str2);
 
@@ -512,6 +537,7 @@ switch (entries)
 			new_free(&help_paused_path);
 			new_free(&help_paused_name);
 		}
+		message_from(NULL, LOG_CRAP);
 		return;
 	}
 	case 0:
@@ -522,7 +548,7 @@ switch (entries)
 			set_help_screen(NULL);
 			break;
 		}
-		sprintf(tmp, "%s%sHelp? ", help_topic_list, *help_topic_list ? " " : empty_string);
+		snprintf(tmp, sizeof tmp, "%s%sHelp? ", help_topic_list, *help_topic_list ? " " : empty_string);
 		if (!dumb_mode)
 			add_wait_prompt(tmp, help_me, help_topic_list, WAIT_PROMPT_LINE, 1);
 
@@ -538,15 +564,15 @@ switch (entries)
 	}
 	case 1:
 	{
-		sprintf(tmp, "%s/%s", path, g.gl_pathv[0]);
+		snprintf(tmp, sizeof tmp, "%s/%s", path, g.gl_pathv[0]);
 		stat(tmp, &stat_buf);
 		if (stat_buf.st_mode & S_IFDIR)
 		{
-			strcpy(path, tmp);
+			strlcpy(path, tmp, sizeof path);
 			if (*help_topic_list)
-				strcat(help_topic_list, " ");
+				strlcat(help_topic_list, " ", sizeof help_topic_list);
 
-			strcat(help_topic_list, g.gl_pathv[0]);
+			strlcat(help_topic_list, g.gl_pathv[0], sizeof help_topic_list);
 
 			if (!(this_arg = next_arg(args, &args)))
 			{
@@ -569,19 +595,19 @@ switch (entries)
 	default:
 	{
 		help_show_directory = 1;
-		strcpy(paused_topic, help_topic_list);
+		strlcpy(paused_topic, help_topic_list, sizeof paused_topic);
 		help_pause_add_line("*** %s choices:", help_topic_list);
 		entry_size += 2;
 		cols = (current_term->TI_cols - 10) / entry_size;
 
-		strcpy(buffer, empty_string);
+		strlcpy(buffer, empty_string, sizeof(buffer));
 		cnt = 0;
 
 		for (i = 0; i < entries; i++)
 		{
 			if (!end_strcmp(g.gl_pathv[i], ".gz", 3))
 				chop(g.gl_pathv[i], 3);
-			strcat(buffer, g.gl_pathv[i]);
+			strlcat(buffer, g.gl_pathv[i], sizeof buffer);
 
 			/*
 			 * Since we already know how many columns each
@@ -592,7 +618,7 @@ switch (entries)
 			if (++cnt == cols)
 			{
 				help_pause_add_line("%s", buffer);
-				strcpy(buffer, empty_string);
+				*buffer = 0;
 				cnt = 0;
 			}
 
@@ -636,6 +662,8 @@ switch (entries)
 	 */
 	if (!*help_topic_list && finished_help_paging)
 		set_help_screen((Screen *) 0);
+
+	message_from(NULL, LOG_CRAP);
 }
 
 /*
@@ -645,6 +673,7 @@ switch (entries)
 BUILT_IN_COMMAND(help)
 {
 	char	*help_path;
+	char	default_topic[10];
 
 	finished_help_paging = 0;
 	help_show_directory = 0;
@@ -661,7 +690,7 @@ BUILT_IN_COMMAND(help)
 
 	if (!help_path || !*help_path || access(help_path, R_OK | X_OK))
 	{
-		help_put_it(no_help, "*** HELP_PATH variable not set or set to an invalid path");
+		help_put_it(no_help, "*** HELP_PATH variable not set or set to an invalid directory");
 		return;
 	}
 
@@ -681,7 +710,8 @@ BUILT_IN_COMMAND(help)
 
 	help_screen = current_window->screen;
 	help_window = (Window *) 0;
-	help_me(empty_string, (args && *args) ? args : "?");
+	strlcpy(default_topic, "?", sizeof default_topic);
+	help_me(empty_string, (args && *args) ? args : default_topic);
 }
 
 
@@ -702,7 +732,8 @@ static	void create_help_window (void)
 		}
 
 		use_help_window = 1;
-		help_window->hold_mode = OFF;
+		help_window->holding_top_of_display = NULL;
+		recalculate_window_cursor_and_display_ip(help_window);
 		help_window->window_level = LOG_HELP;
 		update_all_windows();
 	}
@@ -743,20 +774,19 @@ static	void	help_put_it	(const char *topic, const char *format, ...)
 
 		if (do_hook(HELP_LIST, "%s %s", topic, putbuf))
 		{
-			int old_level = who_level;
 			Window *old_to_window = to_window;
 
 			/*
-			 * LOG_HELP is a completely bogus mode.  We use
-			 * it only to make sure that the current level is
-			 * not LOG_CURRENT, so that the to_window will stick.
+			 * 'to_window' is of higher priority than who_level
+			 * so we don't have to mangle who_level to send the
+			 * output to the help window.
 			 */
-			who_level = LOG_HELP;
 			if (help_window)
 				to_window = help_window;
+			else
+				to_window = current_window;
 			add_to_screen(putbuf);
 			to_window = old_to_window;
-			who_level = old_level;
 		}
 	}
 }
