@@ -1,10 +1,37 @@
+/* $EPIC: hook.c,v 1.25 2003/12/09 04:37:52 jnelson Exp $ */
 /*
  * hook.c: Does those naughty hook functions. 
  *
- * Copyright 1990 Michael Sandrof
- * Copyright 1994 Matthew Green
- * Copyright 1997 EPIC Software Labs
- * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
+ * Copyright (c) 1990 Michael Sandroff.
+ * Copyright (c) 1991, 1992 Troy Rollo.
+ * Copyright (c) 1992-1996 Matthew Green.
+ * Copyright  1993, 2003 EPIC Software Labs.
+ * All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notices, the above paragraph (the one permitting redistribution),
+ *    this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ * 3. The names of the author(s) may not be used to endorse or promote
+ *    products derived from this software without specific prior written
+ *    permission.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
+ * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
+ * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
+ * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
+ * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
+ * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+ * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
+ * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
+ * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
+ * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ * SUCH DAMAGE.
  */
 
 #include "irc.h"
@@ -33,7 +60,7 @@ typedef enum NoiseEnum {
 	NOISY
 } Noise;
 
-static	char	*noise_level[] = { "UNKNOWN", "SILENT", "QUIET", "NORMAL", "NOISY" };
+static	const char	*noise_level[] = { "UNKNOWN", "SILENT", "QUIET", "NORMAL", "NOISY" };
 
 #define HF_NORECURSE	0x0001
 
@@ -79,7 +106,7 @@ NumericList *numeric_list 	= NULL;
 /* A list of all the hook functions available */
 struct
 {
-	char	*name;			/* The name of the hook type */
+	const char *name;		/* The name of the hook type */
 	Hook	*list;			/* The list of events for type */
 	int	params;			/* Number of parameters expected */
 	int	mark;			/* Hook type is currently active */
@@ -93,8 +120,9 @@ hook_functions[] =
 	{ "CHANNEL_SYNC",	(Hook *) 0,	3,	0,	0 },
 	{ "CONNECT",		(Hook *) 0,	2,	0,	0 },
 	{ "CTCP",		(Hook *) 0,	4,	0,	0 },
-	{ "CTCP_REPLY",		(Hook *) 0,	3,	0,	0 },
+	{ "CTCP_REPLY",		(Hook *) 0,	4,	0,	0 },
 	{ "CTCP_REQUEST",	(Hook *) 0,	4,	0,	0 },
+	{ "DCC_ACTIVITY",	(Hook *) 0,	1,	0,	0 },
 	{ "DCC_CHAT",		(Hook *) 0,	2,	0,	0 },
         { "DCC_CONNECT",        (Hook *) 0,     2,      0,      0 },
 	{ "DCC_LIST",		(Hook *) 0,	8,	0,	0 },
@@ -112,11 +140,13 @@ hook_functions[] =
 	{ "EXEC_PROMPT",	(Hook *) 0,	2,	0,	0 },
         { "EXIT",               (Hook *) 0,     1,      0,      0 },
 	{ "FLOOD",		(Hook *) 0,	3,	0,	0 },
+	{ "GENERAL_NOTICE",	(Hook *) 0,	3,	0,	0 },
+	{ "GENERAL_PRIVMSG",	(Hook *) 0,	3,	0,	0 },
 	{ "HELP",		(Hook *) 0,	2,	0,	0 },
 	{ "HOOK",		(Hook *) 0,	1,	0,	0 },
 	{ "IDLE",		(Hook *) 0,	1,	0,	0 },
 	{ "INPUT",		(Hook *) 0,	1,	0,	HF_NORECURSE },
-	{ "INVITE",		(Hook *) 0,	2,	0,	0 },
+	{ "INVITE",		(Hook *) 0,	3,	0,	0 },
 	{ "JOIN",		(Hook *) 0,	4,	0,	0 },
 	{ "KICK",		(Hook *) 0,	3,	0,	0 },
 	{ "KILL",		(Hook *) 0,	5,	0,	0 },
@@ -149,11 +179,13 @@ hook_functions[] =
 	{ "SEND_NOTICE",	(Hook *) 0,	2,	0,	HF_NORECURSE },
 	{ "SEND_PUBLIC",	(Hook *) 0,	2,	0,	HF_NORECURSE },
 	{ "SEND_TO_SERVER",	(Hook *) 0,	3,	0,	0 },
+	{ "SERVER_ESTABLISHED",	(Hook *) 0,	2,	0,	0 },
 	{ "SERVER_LOST",	(Hook *) 0,	2,	0,	0 },
 	{ "SERVER_NOTICE",	(Hook *) 0,	1,	0,	0 },
 	{ "SET",		(Hook *) 0,	2,	0,	0 },
 	{ "SIGNOFF",		(Hook *) 0,	1,	0,	0 },
 	{ "SILENCE",		(Hook *) 0,	2,	0,	0 },
+	{ "SSL_SERVER_CERT",	(Hook *) 0,	3,	0,	0 },
 	{ "STATUS_UPDATE",	(Hook *) 0,	3,	0,	0 },
 	{ "SWITCH_CHANNELS",	(Hook *) 0,	3,	0,	0 },
 	{ "SWITCH_WINDOWS",	(Hook *) 0,	4,	0,	0 },
@@ -165,8 +197,10 @@ hook_functions[] =
 	{ "WHO",		(Hook *) 0,	6,	0,	0 },
 	{ "WIDELIST",		(Hook *) 0,	1,	0,	0 },
 	{ "WINDOW",		(Hook *) 0,	2,	0,	HF_NORECURSE },
+	{ "WINDOW_COMMAND",	(Hook *) 0,	1, 	0,	0 },
 	{ "WINDOW_CREATE",	(Hook *) 0,	1, 	0,	0 },
-	{ "WINDOW_KILL",	(Hook *) 0,	1,	0,	0 },
+	{ "WINDOW_BEFOREKILL",	(Hook *) 0,	1,	0,	0 },
+	{ "WINDOW_KILL",	(Hook *) 0,	2,	0,	0 },
 	{ "YELL",		(Hook *) 0,	1,	0,	0 },
 };
 
@@ -197,26 +231,24 @@ static char *	fill_it_out (char *str, int params)
 	while ((arg = next_arg(ptr, &ptr)) != NULL)
 	{
 		if (*buffer)
-			strmcat(buffer, " ", BIG_BUFFER_SIZE);
-		strmcat(buffer, arg, BIG_BUFFER_SIZE);
+			strlcat(buffer, " ", sizeof buffer);
+		strlcat(buffer, arg, sizeof buffer);
 		if (++i == params)
 			break;
 	}
 
 	for (; i < params; i++)
-		strmcat(buffer, (i < params-1) ? " %" : " *", BIG_BUFFER_SIZE);
+		strlcat(buffer, (i < params-1) ? " %" : " *", sizeof buffer);
 
 	if (*ptr)
 	{
-		strmcat(buffer, " ", BIG_BUFFER_SIZE);
-		strmcat(buffer, ptr, BIG_BUFFER_SIZE);
+		strlcat(buffer, " ", sizeof buffer);
+		strlcat(buffer, ptr, sizeof buffer);
 	}
-	return m_strdup(buffer);
+	return malloc_strdup(buffer);
 }
 
 
-#define INVALID_HOOKNUM -1001
-
 /*
  * find_hook: returns the numerical value for a specified hook name
  */
@@ -307,7 +339,7 @@ static void add_numeric_hook (int numeric, char *nick, char *stuff, int noisy, i
 	{
 		entry = (NumericList *) new_malloc(sizeof(NumericList));
 		entry->numeric = numeric;
-		sprintf(entry->name, "%3.3u", numeric);
+		snprintf(entry->name, sizeof(entry->name), "%3.3u", numeric);
 		entry->next = NULL;
 		entry->list = NULL;
 		add_numeric_list(entry);
@@ -318,6 +350,7 @@ static void add_numeric_hook (int numeric, char *nick, char *stuff, int noisy, i
 		new_h = (Hook *)new_malloc(sizeof(Hook));
 		new_h->nick = NULL;
 		new_h->stuff = NULL;
+		new_h->filename = NULL;
 	}
 
 	malloc_strcpy(&new_h->nick, nick);
@@ -327,7 +360,7 @@ static void add_numeric_hook (int numeric, char *nick, char *stuff, int noisy, i
 	new_h->sernum = sernum;
 	new_h->flexible = flexible;
 	new_h->global = loading_global;
-	new_h->filename = m_strdup(current_package());
+	malloc_strcpy(&new_h->filename, current_package());
 	new_h->next = NULL;
 
 	upper(new_h->nick);
@@ -355,6 +388,7 @@ static void add_hook (int which, char *nick, char *stuff, int noisy, int not, in
 		new_h = (Hook *)new_malloc(sizeof(Hook));
 		new_h->nick = NULL;
 		new_h->stuff = NULL;
+		new_h->filename = NULL;
 	}
 
 	malloc_strcpy(&new_h->nick, nick);
@@ -364,7 +398,7 @@ static void add_hook (int which, char *nick, char *stuff, int noisy, int not, in
 	new_h->sernum = sernum;
 	new_h->flexible = flexible;
 	new_h->global = loading_global;
-	new_h->filename = m_strdup(current_package());
+	malloc_strcpy(&new_h->filename, current_package());
 	new_h->next = NULL;
 
 	upper(new_h->nick);
@@ -513,7 +547,7 @@ void    flush_on_hooks (void)
 {
         int x;
         int old_display = window_display;
-        
+
         window_display = 0;
         for (x = 1; x < 999; x++)
 		remove_numeric_hook(x, NULL, x, 1);
@@ -560,7 +594,7 @@ void	unload_on_hooks (char *filename)
 
 /* * * * * * SHOWING A HOOK * * * * * * */
 /* show_hook shows a single hook */
-static void 	show_hook (Hook *list, char *name)
+static void 	show_hook (Hook *list, const char *name)
 {
 	say("[%s] On %s from %c%s%c do %s [%s] <%d>",
 	    list->filename[0] ? list->filename : "*",
@@ -586,7 +620,7 @@ static int show_numeric_list (int numeric)
 
 	if (numeric)
 	{
-		sprintf(buf, "%3.3u", numeric);
+		snprintf(buf, sizeof buf, "%3.3u", numeric);
 		if ((tmp = find_numeric_list(numeric)))
 		{
 			for (list = tmp->list; list; list = list->next, cnt++)
@@ -638,12 +672,12 @@ static int show_list (int which)
  * the value of the noisy field of the found entry, or -1 if not found. 
  */
 /* huh-huh.. this sucks.. im going to re-write it so that it works */
-int 	do_hook (int which, char *format, ...)
+int 	do_hook (int which, const char *format, ...)
 {
 	Hook		*tmp, 
 			**list;
-	char		buffer		[BIG_BUFFER_SIZE * 10 + 1],
-			*name 		= (char *) 0;
+	char		buffer		[BIG_BUFFER_SIZE * 10 + 1];
+	const char	*name 		= (char *) 0;
 	int		retval 		= DONT_SUPPRESS_DEFAULT;
 	unsigned	display		= window_display;
 	int		i;
@@ -816,6 +850,7 @@ int 	do_hook (int which, char *format, ...)
 		char *		name_copy;
 		char *		stuff_copy;
 		char *		result = NULL;
+		int		noise, old;
 
 		/*
 		 * This should never happen.
@@ -839,6 +874,10 @@ int 	do_hook (int which, char *format, ...)
 		if (tmp->not || !tmp->stuff || !*tmp->stuff)
 			continue;
 
+		name_copy = LOCAL_COPY(name);
+		stuff_copy = LOCAL_COPY(tmp->stuff);
+		noise = tmp->noisy;
+
 		/*
 		 * If this is a NORMAL or NOISY hook, then we tell the user
 		 * that we're going to execute the hook.
@@ -848,22 +887,30 @@ int 	do_hook (int which, char *format, ...)
 				name, tmp->flexible ? '\'' : '"',
 				buffer, tmp->flexible ? '\'' : '"');
 
+		/*
+		 * YOU CAN'T TOUCH ``tmp'' AFTER THIS POINT
+		 */
 
 		/*
 		 * Save some information that may be reset in the 
 		 * execution, turn off the display if the user specified.
 		 */
 		save_message_from(&saved_who_from, &saved_who_level);
-		if (tmp->noisy < NOISY)
+		if (noise < NOISY)
 			window_display = 0;
-		
-		name_copy = LOCAL_COPY(name);
-		stuff_copy = LOCAL_COPY(tmp->stuff);
+		old = system_exception;
 
-		if (tmp->noisy == UNKNOWN)
+		if (retval == RESULT_PENDING)
 		{
-			result = parse_line_with_return(name_copy, 
-						stuff_copy, buffer, 0, 0);
+			result = call_lambda_function(name_copy, stuff_copy,
+							buffer);
+
+			if (result && atol(result))
+				retval = SUPPRESS_DEFAULT;
+			else
+				retval = DONT_SUPPRESS_DEFAULT;
+
+			new_free(&result);
 		}
 		else
 		{
@@ -879,18 +926,11 @@ int 	do_hook (int which, char *format, ...)
 			return_exception = 0;
 		}
 
-		if (retval == RESULT_PENDING)
-		{
-			if (result && atol(result))
-				retval = SUPPRESS_DEFAULT;
-			else
-				retval = DONT_SUPPRESS_DEFAULT;
-		}
-
 		/*
 		 * Clean up the stuff that may have been mangled by the
 		 * execution.
 		 */
+		system_exception = old;
 		window_display = display;
 		restore_message_from(saved_who_from, saved_who_level);
 	}
@@ -963,7 +1003,7 @@ BUILT_IN_COMMAND(shookcmd)
  * integer number, and is used for the serial number for this event.
  *
  * The "verbosity" of the event may also be modified by specifying at most
- * one of the following characters:  
+ * one of the following characters:
  *	A caret (^) is the SILENT level, and indicates that the event is to
  *		be executed with no output (window_display is turned off),
  *		and the "default action" (whatever that is) for the event is
@@ -985,11 +1025,11 @@ BUILT_IN_COMMAND(oncmd)
 {
 	char	*func,
 		*nick,
-		*serial;
+		*serial		= NULL;
 	Noise	noisy		= NORMAL;
 	int	not		= 0,
 		sernum		= 0,
-		remove		= 0,
+		rem		= 0,
 		which		= INVALID_HOOKNUM;
 	int	flex		= 0;
 	char	type;
@@ -1067,7 +1107,7 @@ BUILT_IN_COMMAND(oncmd)
 			 * trying to register the hook, then they've already
 			 * gotten the error message, just return;
 			 */
-			if (new_new_next_arg(args, &args, &type))
+			if (new_new_next_arg_count(args, &args, &type, 1))
 				return;
 
 			/*
@@ -1085,6 +1125,17 @@ BUILT_IN_COMMAND(oncmd)
 			return;
 		}
 
+		/*
+		 * If sernum is 0 and serial is "+" or "-" get a serial
+		 * number for the event type in question
+		 */
+		if (sernum == 0 && serial != NULL) {
+		    if (!strcmp(serial, "+"))
+			sernum = hook_find_free_serial(1, 0, which);
+		    else if (!strcmp(serial, "-"))
+			sernum = hook_find_free_serial(-1, 0, which);
+		}
+
 		/*
 		 * Check to see if this is a removal event or if this
 		 * is a negated event.
@@ -1092,7 +1143,7 @@ BUILT_IN_COMMAND(oncmd)
 		switch (*args)
 		{
 			case '-':
-				remove = 1;
+				rem = 1;
 				args++;
 				break;
 			case '^':
@@ -1104,7 +1155,7 @@ BUILT_IN_COMMAND(oncmd)
 		/*
 		 * Grab the "nick"
 		 */
-		if ((nick = new_new_next_arg(args, &args, &type)))
+		if ((nick = new_new_next_arg_count(args, &args, &type, 1)))
 		{
 			char *exp;
 
@@ -1129,7 +1180,7 @@ BUILT_IN_COMMAND(oncmd)
 			/*
 			 * If we're doing a removal, do the deed.
 			 */
-			if (remove)
+			if (rem)
 			{
 				remove_hook(which, nick, sernum, 0);
 				new_free(&nick);
@@ -1210,7 +1261,7 @@ BUILT_IN_COMMAND(oncmd)
 			/*
 			 * if its a removal, do the deed
 			 */
-			if (remove)
+			if (rem)
 			{
 				remove_hook(which, (char *) 0, sernum, 0);
 				return;
@@ -1267,9 +1318,9 @@ BUILT_IN_COMMAND(oncmd)
 
 
 /* * * * * * * * * * SAVING A HOOK * * * * * * * * * * */
-static	void	write_hook (FILE *fp, Hook *hook, char *name)
+static	void	write_hook (FILE *fp, Hook *hook, const char *name)
 {
-	char	*stuff = (char *) 0;
+	const char	*stuff = (char *) 0;
 	char	flexi = '"';
 
 	if (hook->flexible)
@@ -1426,7 +1477,8 @@ void	do_stack_on (int type, char *args)
 					nptr->list = p->list;
 					nptr->next = NULL;
 					nptr->numeric = -which;
-					sprintf(nptr->name, "%3.3u", -which);
+					snprintf(nptr->name, sizeof(nptr->name),
+						"%3.3u", -which);
 					add_numeric_list(nptr);
 				}
 			}
@@ -1569,3 +1621,64 @@ static NumericList *remove_numeric_list (int numeric)
 
 	return tmp;
 }
+
+/* this function traverses all the hooks for all the events in the system
+ * trying to find a free serial number (one that is unused by any hook on
+ * any event) in the direction given (either -1 or +1 for - and +) starting
+ * at the given point. */
+int hook_find_free_serial(int dir, int from, int which) {
+    int ser;
+    Hook	*hp;
+    NumericList *nlp;
+    int	wc;
+
+    if (from == 0)
+	from = dir;
+
+    /* We iterate through the specified (or all) lists looking for a serial
+     * number that isn't in use.  If we make it through all of our loops
+     * without breaking out of them, we have found an unused number */
+    for (ser = from; (dir > 0 ? ser <= 32767 : ser >= -32767); ser += dir) {
+	if (which != INVALID_HOOKNUM) {
+	    /* a list was specified */
+	    if (which < 0) {
+		if ((nlp = find_numeric_list(-which)) != NULL)
+		    hp = nlp->list;
+		else /* the numeric list doesn't exist.  this must be unique */
+		    return ser;
+	    } else
+		hp = hook_functions[which].list;
+	    
+	    while (hp != NULL) {
+		if (hp->sernum == ser)
+		    break;
+		hp = hp->next;
+	    }
+	    if (hp == NULL)
+		break;
+	} else {
+	    /* no list was specified.  start digging */
+	    hp = NULL;
+	    for (wc = 0; wc < NUMBER_OF_LISTS; wc++) {
+		for (hp = hook_functions[wc].list; hp != NULL; hp = hp->next)
+		    if (hp->sernum == ser)
+			break;
+		if (hp != NULL)
+		    break;
+	    }
+	    if (hp != NULL)
+		continue; /* found a used one.. */
+	    hp = NULL;
+	    for (nlp = numeric_list; nlp != NULL; nlp = nlp->next) {
+		for (hp = nlp->list; hp != NULL; hp = hp->next)
+		    if (hp->sernum == ser)
+			continue;
+	    }
+	    if (hp == NULL)
+		break; /* found an unused one */
+	}
+    }
+
+    return ser;
+}
+
