@@ -1,12 +1,36 @@
+/* $EPIC: if.c,v 1.25 2003/12/15 18:36:44 jnelson Exp $ */
 /*
  * if.c: the IF, WHILE, FOREACH, DO, FE, FEC, and FOR commands for IRCII 
  *
- * WHILE, FOREACH commands written by Troy Rollo
- * Copyright 1992, 1994 Troy Rollo
+ * Copyright (c) 1991, 1994 Troy Rollo.
+ * Copyright (c) 1992-1996 Matthew R. Green.
+ * Copyright  1993, 2003 EPIC Software Labs.
+ * All rights reserved.
  *
- * IF, FE, FEC, FOR, DO, SWITCH, REPEAT commands written by Jeremy Nelson
- * Copyright 1995, 1997 Jeremy Nelson
- * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notices, the above paragraph (the one permitting redistribution),
+ *    this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ * 3. The names of the author(s) may not be used to endorse or promote
+ *    products derived from this software without specific prior written
+ *    permission.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
+ * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
+ * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
+ * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
+ * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
+ * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+ * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
+ * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
+ * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
+ * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ * SUCH DAMAGE.
  */
 
 #include "irc.h"
@@ -29,6 +53,7 @@ char *	my_next_expr (char **args, char type, int whine, int wantchar)
 {
 	char	*expr_start,
 		*expr_end;
+	ssize_t	span;
 
 	if (!*args || !**args || **args != type)
 	{
@@ -39,12 +64,15 @@ char *	my_next_expr (char **args, char type, int whine, int wantchar)
 
 	/* Find the end of the expression and terminate it. */
 	expr_start = *args;
-	if (!(expr_end = MatchingBracket(expr_start + 1, type, 
-					(type == '(') ? ')' : '}')))
+	if ((span = MatchingBracket(expr_start + 1, type, 
+					(type == '(') ? ')' : '}')) < 0)
 	{
 		say("Unmatched '%c'", type);
 		return NULL;
 	}
+	else
+		expr_end = expr_start + 1 + span;
+
 	*expr_end = 0;
 
 	/* 
@@ -57,7 +85,7 @@ char *	my_next_expr (char **args, char type, int whine, int wantchar)
 
 	/* Remove any extraneous whitespace in the expression */
 	expr_start = skip_spaces(expr_start + 1);
-	remove_trailing_spaces(expr_start);
+	remove_trailing_spaces(expr_start, 0);
 
 	/*
 	 * It is guaranteed that (ptr2[-1] >= *args) and so it is further
@@ -95,7 +123,9 @@ char *	next_expr (char **args, char type)
  *		if (expr) {......}
  *		if (expr) {......} {......}
  *		if (expr) {......} else {......}
+ *		if (expr) {......} elif (expr2) {......}
  *		if (expr) {......} elsif (expr2) {......}
+ *		if (expr) {......} else if (expr2) {......}
  * etc.
  */
 
@@ -105,7 +135,7 @@ BUILT_IN_COMMAND(ifcmd)
 	char *current_expr;
 	char *current_expr_val;
 	int result;
-	char *current_line = NULL;
+	char *stuff = NULL;
 	int flag = 0;
 
 	unless_cmd = (*command == 'U');
@@ -128,21 +158,32 @@ BUILT_IN_COMMAND(ifcmd)
 		}
 		current_expr_val = parse_inline(current_expr, subargs, &flag);
 		if (get_int_var(DEBUG_VAR) & DEBUG_EXPANSIONS)
-			yell("%s expression expands to: (%s)", command, current_expr_val);
+			privileged_yell("%s expression expands to: (%s)", command, current_expr_val);
+
 		result = check_val(current_expr_val);
 		new_free(&current_expr_val);
 
 		if (*args == '{')
-			current_line = next_expr(&args, '{');
+			stuff = next_expr(&args, '{');
 		else
-			current_line = args, args = NULL;
+			stuff = args, args = NULL;
 
 		/* If the expression was FALSE for IF, and TRUE for UNLESS */
 		if (unless_cmd == result)
 		{
 			if (args)
 			{
-				if (!my_strnicmp(args, "elsif ", 6))
+				if (!my_strnicmp(args, "elif ", 5))
+				{
+					args += 5;
+					continue;
+				}
+				else if (!my_strnicmp(args, "else if ", 8))
+				{
+					args += 8;
+					continue;
+				}
+				else if (!my_strnicmp(args, "elsif ", 6))
 				{
 					args += 6;
 					continue;
@@ -154,17 +195,17 @@ BUILT_IN_COMMAND(ifcmd)
 					args++;
 
 				if (*args == '{')
-					current_line = next_expr(&args, '{');
+					stuff = next_expr(&args, '{');
 				else
-					current_line = args, args = NULL;
+					stuff = args, args = NULL;
 
 			}
 			else
-				current_line = NULL;
+				stuff = NULL;
 		}
 
-		if (current_line)
-			parse_line(NULL, current_line, subargs, 0, 0);
+		if (stuff)
+			parse_line(NULL, stuff, subargs, 0, 0);
 
 		break;
 	}
@@ -173,14 +214,16 @@ BUILT_IN_COMMAND(ifcmd)
 BUILT_IN_COMMAND(docmd)
 {
 	char *body, *expr, *cmd, *ptr;
-	char *newexp = (char *) 0;
+	char *newexp = NULL;
 	int args_used = 0;
 	int result;
 
+	if (!subargs)
+		subargs = empty_string;
+
 	if (*args == '{')
 	{
-		body = next_expr(&args, '{');
-		if (body == (char *) 0)
+		if (!(body = next_expr(&args, '{')))
 		{
 			error("DO: unbalanced {");
 			return;
@@ -198,7 +241,7 @@ BUILT_IN_COMMAND(docmd)
 			will_catch_continue_exceptions++;
 			while (1)
 			{
-				parse_line ((char *) 0, body, subargs ? subargs : empty_string, 0, 0);
+				parse_line(NULL, body, subargs, 0, 0);
 				if (break_exception)
 				{
 					break_exception = 0;
@@ -214,8 +257,7 @@ BUILT_IN_COMMAND(docmd)
 
 				/* Alas, too bad the malloc is neccesary */
 				malloc_strcpy(&newexp, expr);
-				ptr = parse_inline(newexp, subargs ? subargs : empty_string,
-					&args_used);
+				ptr = parse_inline(newexp, subargs, &args_used);
 				result = check_val(ptr);
 				new_free(&ptr);
 				if (!result)
@@ -227,30 +269,33 @@ BUILT_IN_COMMAND(docmd)
 			return;
 		}
 		/* falls through to here if its /do {...} */
-		parse_line ((char *) 0, body, subargs ? subargs : empty_string, 0, 0);
+		parse_line(NULL, body, subargs, 0, 0);
 	}
 	/* falls through to here if it its /do ... */
-	parse_line ((char *) 0, args, subargs ? subargs : empty_string, 0, 0);
+	parse_line(NULL, args, subargs, 0, 0);
 }
 
 BUILT_IN_COMMAND(whilecmd)
 {
-	char	*exp = (char *) 0,
+	char	*exp = NULL,
 		*ptr,
-		*body = (char *) 0,
-		*newexp = (char *) 0;
+		*body = NULL,
+		*newexp = NULL;
 	int	args_used;	/* this isn't used here, but is passed
 				 * to expand_alias() */
 	int 	whileval = !strcmp(command, "WHILE");
 
-	if ((ptr = next_expr(&args, '(')) == (char *) 0)
+	if (!subargs)
+		subargs = empty_string;
+
+	if (!(ptr = next_expr(&args, '(')))
 	{
 		error("WHILE: missing boolean expression");
 		return;
 	}
 	exp = LOCAL_COPY(ptr);
 
-	if ((ptr = next_expr_failok(&args, '{')) == (char *) 0)
+	if (!(ptr = next_expr_failok(&args, '{')))
 		ptr = args;
 	body = LOCAL_COPY(ptr);
 
@@ -259,13 +304,13 @@ BUILT_IN_COMMAND(whilecmd)
 	while (1)
 	{
 		newexp = LOCAL_COPY(exp);
-		ptr = parse_inline(newexp, subargs ? subargs : empty_string, &args_used);
+		ptr = parse_inline(newexp, subargs, &args_used);
 		if (check_val(ptr) != whileval)
 			break;
 
 		new_free(&ptr);
 
-		parse_line((char *) 0, body, subargs ?  subargs : empty_string, 0, 0);
+		parse_line(NULL, body, subargs, 0, 0);
 		if (continue_exception)
 		{
 			continue_exception = 0;
@@ -286,10 +331,10 @@ BUILT_IN_COMMAND(whilecmd)
 
 BUILT_IN_COMMAND(foreach)
 {
-	char	*struc = (char *) 0,
+	char	*struc = NULL,
 		*ptr,
-		*body = (char *) 0,
-		*var = (char *) 0;
+		*body = NULL,
+		*var = NULL;
 	char	**sublist;
 	int	total;
 	int	i;
@@ -298,13 +343,16 @@ BUILT_IN_COMMAND(foreach)
 	int	list = VAR_ALIAS;
 	int	af;
 
+	if (!subargs)
+		subargs = empty_string;
+
 	while (args && my_isspace(*args))
 		args++;
 
 	if (*args == '-')
 		args++, list = COMMAND_ALIAS;
 
-	if ((ptr = new_next_arg(args, &args)) == (char *) 0)
+	if (!(ptr = new_next_arg(args, &args)))
 	{
 		error("FOREACH: missing structure expression");
 		return;
@@ -312,7 +360,7 @@ BUILT_IN_COMMAND(foreach)
 
 	struc = upper(remove_brackets(ptr, subargs, &af));
 
-	if ((var = next_arg(args, &args)) == (char *) 0)
+	if (!(var = next_arg(args, &args)))
 	{
 		new_free(&struc);
 		error("FOREACH: missing variable");
@@ -321,7 +369,7 @@ BUILT_IN_COMMAND(foreach)
 	while (my_isspace(*args))
 		args++;		/* why was this (*args)++? */
 
-	if ((body = next_expr(&args, '{')) == (char *) 0)
+	if (!(body = next_expr(&args, '{')))
 	{
 		new_free(&struc);
 		error("FOREACH: missing statement");
@@ -342,8 +390,9 @@ BUILT_IN_COMMAND(foreach)
 	for (i = 0; i < total; i++)
 	{
 		add_local_alias(var, sublist[i] + slen + 1, 0);
+		new_free(&sublist[i]);
 
-		parse_line(NULL, body, subargs ? subargs : empty_string, 0, 0);
+		parse_line(NULL, body, subargs, 0, 0);
 	
 		if (continue_exception)
 		{
@@ -357,9 +406,9 @@ BUILT_IN_COMMAND(foreach)
 		}
 		if (return_exception)
 			break;
-
-		new_free(&sublist[i]);
 	}
+	while (++i < total)
+		new_free(&sublist[i]);
 	will_catch_break_exceptions--;
 	will_catch_continue_exceptions--;
 
@@ -373,29 +422,40 @@ BUILT_IN_COMMAND(foreach)
  */
 BUILT_IN_COMMAND(fe)
 {
-	char    *list = (char *) 0,
-		*templist = (char *) 0,
+	char    *list = NULL,
+		*templist = NULL,
 		*placeholder,
 		*vars,
 		*var[255],
-		*word = (char *) 0,
-		*todo = (char *) 0,
+		*word = NULL,
+		*todo = NULL,
 		fec_buffer[2];
-	int     ind, x, y, blah,args_flag;
+	unsigned	ind, x, y;
+	int     args_flag;
 	int     old_display;
 	int	doing_fe = !strcmp(command, "FE");
+	char	*mapvar = NULL;
+	char	*mapsep = doing_fe ? space : empty_string;
+	char	*map = NULL;
+	size_t	mapclue = 0;
 
-	for (x = 0; x <= 254; var[x++] = (char *) 0)
+#if 0
+	for (x = 0; x <= (sizeof(var) / sizeof(*var)); var[x++] = NULL)
 		;
-
-	if (!(list = next_expr(&args, '('))) {
-		error("%s: Missing List for /%s", command, command);
-		return;
-	}
+#endif
 
 	if (!subargs)
 		subargs = empty_string;
-	templist = expand_alias(list, subargs, &args_flag, NULL);
+
+	if (*args == ':' || isalnum(*args)) {
+		mapvar = next_arg(args, &args);
+		templist = get_variable(mapvar);
+	} else if ((list = next_expr(&args, '('))) {
+		templist = expand_alias(list, subargs, &args_flag, NULL);
+	} else {
+		error("%s: Missing List for /%s", command, command);
+		return;
+	}    
 
 	if (!templist || !*templist) {
 		new_free(&templist);
@@ -419,7 +479,7 @@ BUILT_IN_COMMAND(fe)
 
 	while ((var[ind++] = next_arg(vars, &vars)))
 	{
-		if (ind == 255)
+		if (ind > (sizeof(var) / sizeof(*var)))
 		{
 			error("%s: Too many variables", command);
 			new_free(&templist);
@@ -440,22 +500,18 @@ BUILT_IN_COMMAND(fe)
 	if (!doing_fe)
 		{ word = fec_buffer; word[1] = 0; }
 
-	blah = ((doing_fe) ? (word_count(templist)) : (strlen(templist)));
 	placeholder = templist;
 
 	will_catch_break_exceptions++;
 	will_catch_continue_exceptions++;
-	for ( x = 0 ; x < blah ; )
+	for ( x = 0 ; templist && *templist; )
 	{
 		for ( y = 0 ; y < ind ; y++ )
 		{
 			if (doing_fe)
-				word = ((x+y) < blah)
-				    ? new_next_arg(templist, &templist)
-				    : empty_string;
+				word = new_next_arg(templist, &templist);
 			else
-				word[0] = ((x+y) < blah)
-				    ? templist[x+y] : 0;
+				word[0] = *templist ? *templist++ : 0;
 
 			/* Something is really hosed if this happens */
 			if (!word)
@@ -464,8 +520,14 @@ BUILT_IN_COMMAND(fe)
 			add_local_alias(var[y], word, 0);
 		}
 		x += ind;
-		parse_line((char *) 0, todo, 
-		    subargs?subargs:empty_string, 0, 0);
+		parse_line(NULL, todo, subargs, 0, 0);
+
+		if (mapvar)
+			for ( y = 0 ; y < ind ; y++ ) {
+				char *foo = get_variable(var[y]);
+				malloc_strcat_wordlist_c(&map, mapsep, foo, &mapclue);
+				new_free(&foo);
+			}
 
 		if (continue_exception)
 		{
@@ -483,18 +545,24 @@ BUILT_IN_COMMAND(fe)
 	will_catch_break_exceptions--;
 	will_catch_continue_exceptions--;
 
+	add_var_alias(mapvar, map, 0);
+	new_free(&map);
+
 	window_display = 0;
 	window_display = old_display;
 	new_free(&placeholder);
 }
 
 
-void	for_next_cmd (int argc, char **argv, const char *subargs)
+static void	for_next_cmd (int argc, char **argv, const char *subargs)
 {
 	char 	*var, *cmds;
 	char	istr[256];
 	int	start, end, step = 1, i;
 
+	if (!subargs)
+		subargs = empty_string;
+
 	if ((my_stricmp(argv[1], "from") && my_stricmp(argv[1], "=")) ||
 		(argc != 6 && argc != 8))
 	{
@@ -520,7 +588,7 @@ void	for_next_cmd (int argc, char **argv, const char *subargs)
 		cmds++;
 	will_catch_break_exceptions++;
 	will_catch_continue_exceptions++;
-	for (i = start; i <= end; i += step)
+	for (i = start; step > 0 ? i <= end : i >= end; i += step)
 	{
 		snprintf(istr, 255, "%d", i);
 		add_local_alias(var, istr, 0);
@@ -540,12 +608,15 @@ void	for_next_cmd (int argc, char **argv, const char *subargs)
 	will_catch_continue_exceptions--;
 }
 
-void	for_fe_cmd (int argc, char **argv, const char *subargs)
+static void	for_fe_cmd (int argc, char **argv, const char *subargs)
 {
 	char 	*var, *list, *cmds;
 	char	*next, *real_list, *x;
 	int	args_flag = 0;
 
+	if (!subargs)
+		subargs = empty_string;
+
 	if ((my_stricmp(argv[1], "in")) || (argc != 4)) {
 		error("Usage: /FOR var IN (list) {commands}");
 		return;
@@ -582,15 +653,15 @@ void	for_fe_cmd (int argc, char **argv, const char *subargs)
 	new_free(&x);
 }
 
-void	for_pattern_cmd (int argc, char **argv, const char *subargs)
+static void	for_pattern_cmd (int argc, char **argv, const char *subargs)
 {
 	error("/FOR var IN <pattern> {commands} reserved for future use");
 }
 
-BUILT_IN_COMMAND(loopcmd)
+static BUILT_IN_COMMAND(loopcmd)
 {
 	int	argc;
-	char	*argv[9];
+	char	*argv[10];
 
 	if (!subargs)
 		subargs = empty_string;
@@ -603,7 +674,9 @@ BUILT_IN_COMMAND(loopcmd)
 		for_next_cmd(argc, argv, subargs);
 	else if (!my_stricmp(argv[1], "in"))
 	{
-		if (*argv[2] == '(')
+		if (argc < 4)
+			error("%s: syntax error", command);
+		else if (*argv[2] == '(')
 			for_fe_cmd(argc, argv, subargs);
 		else
 			for_pattern_cmd(argc, argv, subargs);
@@ -626,15 +699,17 @@ BUILT_IN_COMMAND(loopcmd)
  */
 BUILT_IN_COMMAND(forcmd)
 {
-	char        *working        = (char *) 0;
-	char        *commence       = (char *) 0;
-	char        *evaluation     = (char *) 0;
-	char        *lameeval       = (char *) 0;
-	char        *iteration      = (char *) 0;
-	const char  *sa             = (char *) 0;
+	char        *working        = NULL;
+	char        *commence       = NULL;
+	char        *evaluation     = NULL;
+	char        *lameeval       = NULL;
+	char        *iteration      = NULL;
 	int         argsused        = 0;
-	char        *blah           = (char *) 0;
-	char        *commands       = (char *) 0;
+	char        *blah           = NULL;
+	char        *commands       = NULL;
+
+	if (!subargs)
+		subargs = empty_string;
 
 	/* Get the whole () thing */
 	if ((working = next_expr_failok(&args, '(')) == NULL)	/* ) */
@@ -646,42 +721,37 @@ BUILT_IN_COMMAND(forcmd)
 	commence = LOCAL_COPY(working);
 
 	/* Find the beginning of the second expression */
-	evaluation = strchr(commence, ',');
-	if (!evaluation)
+	if (!(evaluation = strchr(commence, ',')))
 	{
 		error("FOR: no components!");
 		return;
 	}
 	do 
-		*evaluation++ = '\0';
+		*evaluation++ = 0;
 	while (my_isspace(*evaluation));
 
 	/* Find the beginning of the third expression */
-	iteration = strchr(evaluation, ',');
-	if (!iteration)
+	if (!(iteration = strchr(evaluation, ',')))
 	{
 		error("FOR: Only two components!");
 		return;
 	}
 	do 
-	{
-		*iteration++ = '\0';
-	}
+		*iteration++ = 0;
 	while (my_isspace(*iteration));
 
 	working = args;
 	while (my_isspace(*working))
-		*working++ = '\0';
+		*working++ = 0;
 
-	if ((working = next_expr(&working, '{')) == (char *) 0)		/* } */
+	if (!(working = next_expr(&working, '{')))		/* } */
 	{
 		error("FOR: badly formed commands");
 		return;
 	}
 	commands = LOCAL_COPY(working);
 
-	sa = subargs ? subargs : empty_string;
-	parse_line((char *) 0, commence, sa, 0, 0);
+	parse_line(NULL, commence, subargs, 0, 0);
 
 	will_catch_break_exceptions++;
 	will_catch_continue_exceptions++;
@@ -689,7 +759,7 @@ BUILT_IN_COMMAND(forcmd)
 	{
 		lameeval = LOCAL_COPY(evaluation);
 
-		blah = parse_inline(lameeval, sa, &argsused);
+		blah = parse_inline(lameeval, subargs, &argsused);
 		if (!check_val(blah))
 		{
 			new_free(&blah);
@@ -697,7 +767,7 @@ BUILT_IN_COMMAND(forcmd)
 		}
 
 		new_free(&blah);
-		parse_line((char *) 0, commands, sa, 0, 0);
+		parse_line(NULL, commands, subargs, 0, 0);
 		if (break_exception)
 		{
 			break_exception = 0;
@@ -708,7 +778,7 @@ BUILT_IN_COMMAND(forcmd)
 		if (return_exception)
 			break;
 
-		parse_line((char *) 0, iteration, sa, 0, 0);
+		parse_line(NULL, iteration, subargs, 0, 0);
 	}
 	will_catch_break_exceptions--;
 	will_catch_continue_exceptions--;
@@ -760,6 +830,9 @@ BUILT_IN_COMMAND(switchcmd)
 		*commands;
 	int 	af;
 
+	if (!subargs)
+		subargs = empty_string;
+
 	if (!(control = next_expr(&args, '(')))
 	{
 		say("SWITCH: String to be matched not found where expected");
@@ -768,7 +841,7 @@ BUILT_IN_COMMAND(switchcmd)
 
 	control = expand_alias(control, subargs, &af, NULL);
 	if (get_int_var(DEBUG_VAR) & DEBUG_EXPANSIONS)
-		yell("%s expression expands to: (%s)", command, control);
+		privileged_yell("%s expression expands to: (%s)", command, control);
 
 	if (!(body = next_expr(&args, '{')))
 		say("SWITCH: Execution body not found where expected");
@@ -793,7 +866,7 @@ BUILT_IN_COMMAND(switchcmd)
 			}
 			header = expand_alias(header, subargs, &af, NULL);
 			if (get_int_var(DEBUG_VAR) & DEBUG_EXPANSIONS)
-				yell("%s expression expands to: (%s)", command, header);
+				privileged_yell("%s expression expands to: (%s)", command, header);
 			if (wild_match(header, control))
 				hooked = 1;
 			new_free(&header);
@@ -817,7 +890,7 @@ BUILT_IN_COMMAND(switchcmd)
 			break;
 		}
 
-		if (*body == ';')
+		while (*body && (*body == ';' || isspace(*body)))
 			body++;		/* grumble */
 	}
 
@@ -828,41 +901,38 @@ BUILT_IN_COMMAND(repeatcmd)
 {
 	char *	num_expr = NULL;
 	int 	value;
+	char *	tmp_val;
+
+	if (!subargs)
+		subargs = empty_string;
 
 	while (isspace(*args))
 		args++;
 
 	if (*args == '(')
 	{
-		char *		tmp_val;
 		char *		dumb_copy;
 		int 		argsused;
-		const char *	sa = subargs ? subargs : empty_string;
 
 		num_expr = next_expr(&args, '(');
 		dumb_copy = LOCAL_COPY(num_expr);
-
-		tmp_val = parse_inline(dumb_copy,sa,&argsused);
-		value = my_atol(tmp_val);
-		new_free(&tmp_val);
+		tmp_val = parse_inline(dumb_copy, subargs, &argsused);
 	}
 	else
 	{
-		char *		tmp_val;
-		int 		af;
-
 		num_expr = new_next_arg(args, &args);
-		tmp_val = expand_alias(num_expr, subargs, &af, NULL);
-		value = my_atol(tmp_val);
-		new_free(&tmp_val);
+		tmp_val = malloc_strdup(num_expr);
 	}
 
+	value = my_atol(tmp_val);
+	new_free(&tmp_val);
+
 	if (value <= 0)
 		return;
 
 	/* Probably want to catch break and continue here */
 	while (value--)
-		parse_line(NULL, args, subargs ? subargs : empty_string, 0, 0);
+		parse_line(NULL, args, subargs, 0, 0);
 
 	return;
 }
