@@ -1,22 +1,52 @@
+/* $EPIC: input.c,v 1.19 2004/02/03 19:36:52 wd Exp $ */
 /*
  * input.c: does the actual input line stuff... keeps the appropriate stuff
  * on the input line, handles insert/delete of characters/words... the whole
  * ball o wax.
  *
+ * Copyright (c) 1990 Michael Sandroff.
+ * Copyright (c) 1991, 1992 Troy Rollo.
+ * Copyright (c) 1992-1996 Matthew Green.
+ * Copyright  1999, 2003 EPIC Software Labs.
+ * All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notices, the above paragraph (the one permitting redistribution),
+ *    this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ * 3. The names of the author(s) may not be used to endorse or promote
+ *    products derived from this software without specific prior written
+ *    permission.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
+ * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
+ * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
+ * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
+ * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
+ * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+ * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
+ * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
+ * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
+ * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ * SUCH DAMAGE.
+ */
+/*
  * This file has been mostly rewritten and reorganized by now.  One of the
  * things we can do now is have completely independant input lines on each
  * screen, and there are really no global variables any more.  A lot of the
  * original code still lies about, as its still useful, but the macros have
  * made the original code hard to distinguish.
- *
- * Copyright 1990 Michael Sandrof
- * Copyright 1999 EPIC Software Labs
- * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
  */
 
 #define __need_term_flush__
 #include "irc.h"
 #include "alias.h"
+#include "clock.h"
 #include "commands.h"
 #include "exec.h"
 #include "history.h"
@@ -59,7 +89,7 @@
 #define MIN_CHAR 		INPUT_BUFFER[MIN_POS]
 #define PREV_CHAR 		INPUT_BUFFER[THIS_POS-1]
 #define NEXT_CHAR 		INPUT_BUFFER[THIS_POS+1]
-#define ADD_TO_INPUT(x) 	strmcat(INPUT_BUFFER, (x), INPUT_BUFFER_SIZE);
+#define ADD_TO_INPUT(x) 	strlcat(INPUT_BUFFER, (x), sizeof INPUT_BUFFER);
 #define INPUT_ONSCREEN 		current_screen->input_visible
 #define INPUT_VISIBLE 		INPUT_BUFFER[INPUT_ONSCREEN]
 #define ZONE			current_screen->input_zone_len
@@ -71,9 +101,6 @@
 #define CUT_BUFFER		cut_buffer
 #define SET_CUT_BUFFER(x)	malloc_strcpy(&CUT_BUFFER, x);
 
-#define BUILT_IN_KEYBINDING(x) void x (char key, char *string)
-
-
 /* XXXX Only used here anyhow XXXX */
 static int 	safe_puts (char *str, int len, int echo) 
 {
@@ -110,6 +137,36 @@ void 	cursor_to_input (void)
 	output_screen = last_input_screen = oldscreen;
 }
 
+/* do_input_timeouts: walk through the list of screens and potentially do
+ * timeouts for input */
+int 	do_input_timeouts (void *ignored)
+{
+    Screen *oldscreen = last_input_screen;
+    Screen *screen;
+    int server;
+
+    if (!foreground)
+	return 0;		/* Not a lot of input.. :) */
+
+    /* I hope this is sufficient to allow input to be processed.  I took
+     * a lot of the guts from do_screens().  Here's hoping I didn't
+     * break anything. ;) */
+    for (screen = screen_list; screen; screen = screen->next) {
+	if (screen->alive) {
+	    server = from_server;
+	    last_input_screen = screen;
+	    output_screen = screen;
+	    make_window_current(screen->current_window);
+	    from_server = current_window->server;
+	    screen->last_key = timeout_keypress(screen->last_key,
+		    screen->last_press);
+	    from_server = server;
+	}
+    }
+    output_screen = last_input_screen = oldscreen;
+    return 0;
+}
+
 /*
  * update_input: does varying amount of updating on the input line depending
  * upon the position of the cursor and the update flag.  If the cursor has
@@ -135,7 +192,7 @@ void	update_input (int update)
 		free_it = 0,
 		max;
 	char	*prompt;
-	int	echo = 1;
+	int	do_echo = 1;
 	Screen	*os = last_input_screen;
 	Screen	*ns;
 	Window	*saved_current_window = current_window;
@@ -211,8 +268,7 @@ void	update_input (int update)
 		 * Mangle out any ansi chars or so forth.
 		 */
 		ptr_free = ptr;
-		ptr = strip_ansi(ptr);
-		strcat(ptr, ALL_OFF_STR);	/* Yes, we can do this */
+		ptr = normalize_string(ptr, 0);	/* This should be ok */
 		if (free_it)
 			new_free(&ptr_free);
 		free_it = 1;
@@ -269,7 +325,6 @@ void	update_input (int update)
 		ZONE = last_input_screen->co - (WIDTH * 2);
 		if (ZONE < 10)
 			ZONE = 10;		/* Take that! */
-
 		START_ZONE = WIDTH;
 		END_ZONE = last_input_screen->co - WIDTH;
 
@@ -320,9 +375,11 @@ void	update_input (int update)
 	 * left of the zone, after adjusting for the length of the prompt.
 	 */
 	if (START_ZONE == WIDTH)
+	    INPUT_ONSCREEN = 0;
+	else {
+	    if ((INPUT_ONSCREEN = START_ZONE - WIDTH - INPUT_PROMPT_LEN) < 0)
 		INPUT_ONSCREEN = 0;
-	else
-		INPUT_ONSCREEN = START_ZONE - WIDTH - INPUT_PROMPT_LEN;
+	}
 
 	/*
 	 * And the cursor is simply how many characters away THIS_POS is
@@ -354,7 +411,7 @@ void	update_input (int update)
 			/*
 			 * Forcibly turn on echo.
 			 */
-			int	echo = term_echo(1);
+			do_echo = term_echo(1);
 
 			/*
 			 * Crop back the input prompt so it does not extend
@@ -372,15 +429,15 @@ void	update_input (int update)
 			 * Turn the echo back to what it was before,
 			 * and output the rest of the input buffer.
 			 */
-			term_echo(echo);
-			safe_puts(INPUT_BUFFER, last_input_screen->co - INPUT_PROMPT_LEN, echo);
+			term_echo(do_echo);
+			safe_puts(INPUT_BUFFER, last_input_screen->co - INPUT_PROMPT_LEN, do_echo);
 		}
 
 		/*
 		 * Otherwise we just output whatever we have.
 		 */
-		else if (echo)
-			safe_puts(&(INPUT_VISIBLE), last_input_screen->co, echo);
+		else if (do_echo)
+			safe_puts(&(INPUT_VISIBLE), last_input_screen->co, do_echo);
 
 		/*
 		 * Clear the rest of the input line and reset the cursor
@@ -388,6 +445,7 @@ void	update_input (int update)
 		 */
 		term_clear_to_eol();
 		term_move_cursor(INPUT_CURSOR, INPUT_LINE);
+		cursor_not_in_display(last_input_screen);
 	}
 
 	/*
@@ -408,9 +466,10 @@ void	update_input (int update)
 
 		if ((len = strlen(&(THIS_CHAR))) > max)
 			len = max;
-		safe_puts(&(THIS_CHAR), len, echo);
+		safe_puts(&(THIS_CHAR), len, do_echo);
 		term_clear_to_eol();
 		term_move_cursor(INPUT_CURSOR, INPUT_LINE);
+		cursor_not_in_display(last_input_screen);
 	}
 
 	/*
@@ -418,7 +477,10 @@ void	update_input (int update)
 	 * line, then go ahead and do that.
 	 */
 	else if (update == UPDATE_JUST_CURSOR)
+	{
 		term_move_cursor(INPUT_CURSOR, INPUT_LINE);
+		cursor_not_in_display(last_input_screen);
+	}
 
 	/*
 	 * Turn the terminal echo back on, and flush all of the output
@@ -437,7 +499,7 @@ void 	change_input_prompt (int direction)
 	/* XXXXXX THIS is totaly wrong. XXXXXX */
 	if (!last_input_screen->promptlist)
 	{
-		strcpy(INPUT_BUFFER, last_input_screen->saved_input_buffer);
+		strlcpy(INPUT_BUFFER, last_input_screen->saved_input_buffer, sizeof INPUT_BUFFER);
 		THIS_POS = last_input_screen->saved_buffer_pos;
 		MIN_POS = last_input_screen->saved_min_buffer_pos;
 		*last_input_screen->saved_input_buffer = 0;
@@ -450,7 +512,7 @@ void 	change_input_prompt (int direction)
 
 	else if (!last_input_screen->promptlist->next)
 	{
-		strcpy(last_input_screen->saved_input_buffer, INPUT_BUFFER);
+		strlcpy(last_input_screen->saved_input_buffer, INPUT_BUFFER, sizeof INPUT_BUFFER);
 		last_input_screen->saved_buffer_pos = THIS_POS;
 		last_input_screen->saved_min_buffer_pos = MIN_POS;
 		*INPUT_BUFFER = 0;
@@ -489,8 +551,9 @@ void	input_move_cursor (int dir)
  */
 void	set_input (char *str)
 {
-	strmcpy(INPUT_BUFFER + MIN_POS, str, INPUT_BUFFER_SIZE - MIN_POS);
+	strlcpy(INPUT_BUFFER + MIN_POS, str, INPUT_BUFFER_SIZE - MIN_POS);
 	THIS_POS = strlen(INPUT_BUFFER);
+	update_input(UPDATE_ALL);
 }
 
 /*
@@ -521,8 +584,10 @@ char *	get_input_prompt (void)
  * buffer.  This prompt cannot be backspaced over, etc.  It's a prompt.
  * Setting the prompt to null uses no prompt 
  */
-void	set_input_prompt (char *prompt)
+void	set_input_prompt (const void *stuff)
 {
+	const char *prompt = (const char *)stuff;
+
 	if (prompt)
 		malloc_strcpy(&input_prompt, prompt);
 	else if (input_prompt)
@@ -547,7 +612,7 @@ void	set_input_prompt (char *prompt)
  * input_forward_word: move the input cursor forward one word in the input
  * line 
  */
-BUILT_IN_KEYBINDING(input_forward_word)
+BUILT_IN_BINDING(input_forward_word)
 {
 	cursor_to_input();
 
@@ -563,7 +628,7 @@ BUILT_IN_KEYBINDING(input_forward_word)
 }
 
 /* input_backward_word: move the cursor left on word in the input line */
-BUILT_IN_KEYBINDING(input_backward_word)
+BUILT_IN_BINDING(input_backward_word)
 {
 	cursor_to_input();
 
@@ -619,11 +684,12 @@ static void	input_delete_char_from_screen (void)
 		else
 			pos = INPUT_ONSCREEN + last_input_screen->co - 1;
 
-		if (pos < strlen(INPUT_BUFFER))
+		if (pos < (int)strlen(INPUT_BUFFER))
 		{
 			term_move_cursor(last_input_screen->co - 1, INPUT_LINE);
 			term_putchar(INPUT_BUFFER[pos]);
 			term_move_cursor(INPUT_CURSOR, INPUT_LINE);
+			cursor_not_in_display(last_input_screen);
 		}
 
 		/* XXX - Very possibly, this is pointless */
@@ -635,7 +701,7 @@ static void	input_delete_char_from_screen (void)
  * input_delete_character -- Deletes the character currently under the
  * 			     input cursor.
  */
-BUILT_IN_KEYBINDING(input_delete_character)
+BUILT_IN_BINDING(input_delete_character)
 {
 	cursor_to_input();
 
@@ -660,7 +726,7 @@ BUILT_IN_KEYBINDING(input_delete_character)
  *		      delete_character.  No, this is not significantly
  *		      more expensive than the old way.
  */
-BUILT_IN_KEYBINDING(input_backspace)
+BUILT_IN_BINDING(input_backspace)
 {
 	cursor_to_input();
 
@@ -680,7 +746,7 @@ BUILT_IN_KEYBINDING(input_backspace)
  * input_beginning_of_line: moves the input cursor to the first character in
  * the input buffer 
  */
-BUILT_IN_KEYBINDING(input_beginning_of_line)
+BUILT_IN_BINDING(input_beginning_of_line)
 {
 	cursor_to_input();
 	THIS_POS = MIN_POS;
@@ -691,7 +757,7 @@ BUILT_IN_KEYBINDING(input_beginning_of_line)
  * input_end_of_line: moves the input cursor to the last character in the
  * input buffer 
  */
-BUILT_IN_KEYBINDING(input_end_of_line)
+BUILT_IN_BINDING(input_end_of_line)
 {
 	cursor_to_input();
 	THIS_POS = strlen(INPUT_BUFFER);
@@ -703,15 +769,16 @@ BUILT_IN_KEYBINDING(input_end_of_line)
  * position from the input line, and puts it into the cut buffer.  It does
  * the requisite redraw as well.
  */
-void	cut_input (int anchor)
+static void	cut_input (int anchor)
 {
 	char *	buffer;
-	int	size;
+	size_t	size;
 
 	if (anchor < THIS_POS)
 	{
-		buffer = alloca((size = THIS_POS - anchor) + 1);
-		strmcpy(buffer, &INPUT_BUFFER[anchor], size);
+		size = THIS_POS - anchor;
+		buffer = alloca(size + 1);
+		strlcpy(buffer, &INPUT_BUFFER[anchor], size + 1);
 		malloc_strcpy(&cut_buffer, buffer);
 
 		buffer = LOCAL_COPY(&THIS_CHAR);
@@ -721,8 +788,9 @@ void	cut_input (int anchor)
 	}
 	else
 	{
-		buffer = alloca((size = anchor - THIS_POS) + 1);
-		strmcpy(buffer, &THIS_CHAR, size);
+		size = anchor - THIS_POS;
+		buffer = alloca(size + 1);
+		strlcpy(buffer, &THIS_CHAR, size + 1);
 		malloc_strcpy(&cut_buffer, buffer);
 
 		buffer = LOCAL_COPY(&INPUT_BUFFER[anchor]);
@@ -746,7 +814,7 @@ void	cut_input (int anchor)
  *		This is e input buffer
  *			^			(the input cursor)
  */
-BUILT_IN_KEYBINDING(input_delete_to_previous_space)
+BUILT_IN_BINDING(input_delete_to_previous_space)
 {
 	int	anchor;
 
@@ -766,7 +834,7 @@ BUILT_IN_KEYBINDING(input_delete_to_previous_space)
  * space character.  This is probably going to be the same effect as 
  * delete_to_previous_space, but hey -- you know.
  */
-BUILT_IN_KEYBINDING(input_delete_previous_word)
+BUILT_IN_BINDING(input_delete_previous_word)
 {
 	int	anchor;
 
@@ -784,7 +852,7 @@ BUILT_IN_KEYBINDING(input_delete_previous_word)
  * input_delete_next_word: deletes from the cursor to the end of the next
  * word 
  */
-BUILT_IN_KEYBINDING(input_delete_next_word)
+BUILT_IN_BINDING(input_delete_next_word)
 {
 	int	anchor;
 
@@ -802,7 +870,7 @@ BUILT_IN_KEYBINDING(input_delete_next_word)
  * input_add_character: adds the character c to the input buffer, repecting
  * the current overwrite/insert mode status, etc 
  */
-BUILT_IN_KEYBINDING(input_add_character)
+BUILT_IN_BINDING(input_add_character)
 {
 	int	display_flag = NO_UPDATE;
 
@@ -878,7 +946,7 @@ BUILT_IN_KEYBINDING(input_add_character)
 }
 
 /* input_clear_to_eol: erases from the cursor to the end of the input buffer */
-BUILT_IN_KEYBINDING(input_clear_to_eol)
+BUILT_IN_BINDING(input_clear_to_eol)
 {
 	/* This doesnt really speak to the implementation, but it works.  */
 	cursor_to_input();
@@ -892,7 +960,7 @@ BUILT_IN_KEYBINDING(input_clear_to_eol)
  * input_clear_to_bol: clears from the cursor to the beginning of the input
  * buffer 
  */
-BUILT_IN_KEYBINDING(input_clear_to_bol)
+BUILT_IN_BINDING(input_clear_to_bol)
 {
 	char	c = THIS_CHAR;
 
@@ -906,21 +974,25 @@ BUILT_IN_KEYBINDING(input_clear_to_bol)
 	THIS_POS = MIN_POS;
 	term_move_cursor(INPUT_PROMPT_LEN, INPUT_LINE);
 	term_clear_to_eol();
+	cursor_not_in_display(last_input_screen);
 	update_input(UPDATE_FROM_CURSOR);
 }
 
 /*
  * input_clear_line: clears entire input line
  */
-BUILT_IN_KEYBINDING(input_clear_line)
+BUILT_IN_BINDING(input_clear_line)
 {
 	cursor_to_input();
 
-	malloc_strcpy(&cut_buffer, INPUT_BUFFER + MIN_POS);
+	if (MIN_CHAR != '\0')
+		/* only copy if there is input. -wd */
+		malloc_strcpy(&cut_buffer, INPUT_BUFFER + MIN_POS);
 	MIN_CHAR = 0;
 	THIS_POS = MIN_POS;
 	term_move_cursor(INPUT_PROMPT_LEN, INPUT_LINE);
 	term_clear_to_eol();
+	cursor_not_in_display(last_input_screen);
 	update_input(NO_UPDATE);
 	if (get_int_var(HISTORY_CIRCLEQ_VAR))
 		abort_history_browsing(0);
@@ -930,7 +1002,7 @@ BUILT_IN_KEYBINDING(input_clear_line)
  * input_transpose_characters: swaps the positions of the two characters
  * before the cursor position 
  */
-BUILT_IN_KEYBINDING(input_transpose_characters)
+BUILT_IN_BINDING(input_transpose_characters)
 {
 	cursor_to_input();
 	if (last_input_screen->buffer_pos > MIN_POS)
@@ -952,7 +1024,7 @@ BUILT_IN_KEYBINDING(input_transpose_characters)
 		 * then swap the two characters before the cursor
 		 * XXX This strlen() is probably unnecesary.
 		 */
-		else if (strlen(INPUT_BUFFER) > MIN_POS + 2)
+		else if ((int) strlen(INPUT_BUFFER) > MIN_POS + 2)
 		{
 			pos = THIS_POS - 1;
 			end_of_line = 1;
@@ -996,7 +1068,7 @@ BUILT_IN_KEYBINDING(input_transpose_characters)
 }
 
 
-BUILT_IN_KEYBINDING(refresh_inputline)
+BUILT_IN_BINDING(refresh_inputline)
 {
 	update_input(UPDATE_ALL);
 }
@@ -1005,7 +1077,7 @@ BUILT_IN_KEYBINDING(refresh_inputline)
  * input_yank_cut_buffer: takes the contents of the cut buffer and inserts it
  * into the input line 
  */
-BUILT_IN_KEYBINDING(input_yank_cut_buffer)
+BUILT_IN_BINDING(input_yank_cut_buffer)
 {
 	char	*ptr = NULL;
 
@@ -1030,41 +1102,53 @@ BUILT_IN_KEYBINDING(input_yank_cut_buffer)
 #define LEFT 0
 
 /* BIND functions: */
-BUILT_IN_KEYBINDING(forward_character)
+BUILT_IN_BINDING(forward_character)
 {
 	input_move_cursor(RIGHT);
 }
 
-BUILT_IN_KEYBINDING(backward_character)
+BUILT_IN_BINDING(backward_character)
 {
 	input_move_cursor(LEFT);
 }
 
-BUILT_IN_KEYBINDING(backward_history)
+BUILT_IN_BINDING(backward_history)
 {
 	get_history(OLDER);		/* Cursor up -- older -- prev */
 }
 
-BUILT_IN_KEYBINDING(forward_history)
+BUILT_IN_BINDING(forward_history)
 {
 	get_history(NEWER);		/* Cursor down -- newer -- next */
 }
 
-BUILT_IN_KEYBINDING(toggle_insert_mode)
+BUILT_IN_BINDING(toggle_insert_mode)
 {
-	set_var_value(INSERT_MODE_VAR, "TOGGLE");
+	char *whatever;
+	whatever = LOCAL_COPY("TOGGLE");
+	set_var_value(INSERT_MODE_VAR, whatever);
 }
 
-BUILT_IN_KEYBINDING(send_line)
+BUILT_IN_BINDING(send_line)
 {
 	int	server = from_server;
 	char *	line = LOCAL_COPY(get_input());
+	int	do_unhold;
 
 	from_server = get_window_server(0);
-	unhold_a_window(last_input_screen->current_window);
 	MIN_CHAR = 0;
 	THIS_POS = MIN_POS;
 
+	/*
+	 * If the window is not holding, then advance the hold view so a
+	 * full screen of output from the command is seen before it holds.
+	 * This is consistent with ircII behavior.
+	 */
+	do_unhold = window_is_holding(last_input_screen->current_window);
+	if (!do_unhold)
+		unhold_a_window(last_input_screen->current_window);
+
+
 	if (last_input_screen->promptlist && 
 		last_input_screen->promptlist->type == WAIT_PROMPT_LINE)
 	{
@@ -1080,9 +1164,10 @@ BUILT_IN_KEYBINDING(send_line)
 	}
 	else
 	{
+		int old = system_exception;
+
 		/* Clear the input line before dispatching the command */
 		update_input(UPDATE_ALL);
-
 		if (do_hook(INPUT_LIST, "%s", line))
 		{
 			if (get_int_var(INPUT_ALIASES_VAR))
@@ -1090,14 +1175,24 @@ BUILT_IN_KEYBINDING(send_line)
 			else
 				parse_line(NULL, line, NULL, 1, 0);
 		}
+
+		system_exception = old;
 	}
 
+	/*
+	 * If the window is holding, then any output from the above command
+	 * was already held, and so we need to unhold one screenful here.
+	 * This is consistent with ircII behavior.
+	 */
+	if (do_unhold)
+		unhold_a_window(last_input_screen->current_window);
+
 	from_server = server;
 }
 
 
 
-BUILT_IN_KEYBINDING(quote_char)
+BUILT_IN_BINDING(quote_char)
 {
 	last_input_screen->quote_hit = 1;
 }
@@ -1108,39 +1203,39 @@ BUILT_IN_KEYBINDING(quote_char)
  * appropriate characters when you press any key to which you have bound
  * that highlight character. >;-)
  */
-BUILT_IN_KEYBINDING(insert_bold)
+BUILT_IN_BINDING(insert_bold)
 {
 	input_add_character(BOLD_TOG, string);
 }
 
-BUILT_IN_KEYBINDING(insert_reverse)
+BUILT_IN_BINDING(insert_reverse)
 {
 	input_add_character(REV_TOG, string);
 }
 
-BUILT_IN_KEYBINDING(insert_underline)
+BUILT_IN_BINDING(insert_underline)
 {
 	input_add_character(UND_TOG, string);
 }
 
-BUILT_IN_KEYBINDING(highlight_off)
+BUILT_IN_BINDING(highlight_off)
 {
 	input_add_character(ALL_OFF, string);
 }
 
-BUILT_IN_KEYBINDING(insert_blink)
+BUILT_IN_BINDING(insert_blink)
 {
 	input_add_character(BLINK_TOG, string);
 }
 
-BUILT_IN_KEYBINDING(insert_altcharset)
+BUILT_IN_BINDING(insert_altcharset)
 {
 	yell("alt_tog!");
 	input_add_character(ALT_TOG, string);
 }
 
 /* type_text: the BIND function TYPE_TEXT */
-BUILT_IN_KEYBINDING(type_text)
+BUILT_IN_BINDING(type_text)
 {
 	for (; *string; string++)
 		input_add_character(*string, empty_string);
@@ -1150,39 +1245,30 @@ BUILT_IN_KEYBINDING(type_text)
  * clear_screen: the CLEAR_SCREEN function for BIND.  Clears the screen and
  * starts it if it is held 
  */
-BUILT_IN_KEYBINDING(clear_screen)
+BUILT_IN_BINDING(clear_screen)
 {
-	hold_mode(NULL, OFF, 1);
-	clear_window_by_refnum(0);
+	clear_window_by_refnum(0, 1);
 }
 
-BUILT_IN_KEYBINDING(input_unclear_screen)
+BUILT_IN_BINDING(input_unclear_screen)
 {
-	hold_mode(NULL, OFF, 1);
-	unclear_window_by_refnum(0);
+	unclear_window_by_refnum(0, 1);
 }
 
 /* parse_text: the bindable function that executes its string */
-BUILT_IN_KEYBINDING(parse_text)
+BUILT_IN_BINDING(parse_text)
 {
+	int	old = system_exception;
 	parse_line(NULL, string, empty_string, 0, 0);
+	system_exception = old;
 }
 
-BUILT_IN_KEYBINDING(cpu_saver_on)
-{
-	cpu_saver = 1;
-	update_all_status();
-}
-
-
 /*
  * edit_char: handles each character for an input stream.  Not too difficult
  * to work out.
  */
 void	edit_char (u_char key)
 {
-	void		(*func) 	(char, char *) = NULL;
-	char *		ptr = NULL;
 	u_char		extended_key;
 	WaitPrompt *	oldprompt;
 	u_char		dummy[2];
@@ -1254,38 +1340,27 @@ void	edit_char (u_char key)
 		input_add_character(extended_key, empty_string);
 	}
 
-	/* Otherwise find out what its binding is and dispatch it. */
-	else
-	{
-		int	m = last_input_screen->meta_hit;
-		int	i;
-
-		if ((i = get_binding(m, key, &func, &ptr)))
-		{
-			if (m == 4 && i == 4)
-				last_input_screen->meta_hit = 0;
-			else
-				last_input_screen->meta_hit = i;
-		}
-		else if (last_input_screen->meta_hit != 4)
-			last_input_screen->meta_hit = 0;
-		if (func)
-			func(extended_key, SAFE(ptr));
+	/* Otherwise, let the keybinding system take care of the work. */
+	else {
+		last_input_screen->last_key = handle_keypress(
+			last_input_screen->last_key,
+			last_input_screen->last_press, key);
+		get_time(&last_input_screen->last_press);
 	}
 }
 
 /*
  * type: The TYPE command.  This parses the given string and treats each
- * character as though it were typed in by the user.  Thus key bindings 
- * are used for each character parsed.  Special case characters are control 
- * character sequences, specified by a ^ follow by a legal control key.  
- * Thus doing "/TYPE ^B" will be as tho ^B were hit at the keyboard, 
+ * character as though it were typed in by the user.  Thus key bindings
+ * are used for each character parsed.  Special case characters are control
+ * character sequences, specified by a ^ follow by a legal control key.
+ * Thus doing "/TYPE ^B" will be as tho ^B were hit at the keyboard,
  * probably moving the cursor backward one character.
  *
  * This was moved from keys.c, because it certainly does not belong there,
  * and this seemed a reasonable place for it to go for now.
  */
-BUILT_IN_COMMAND(type)
+BUILT_IN_COMMAND(typecmd)
 {
 	int	c;
 	char	key;
