@@ -1,11 +1,38 @@
+/* $EPIC: irc.c,v 1.705 2004/02/21 17:12:18 keerf Exp $ */
 /*
  * ircII: a new irc client.  I like it.  I hope you will too!
  *
- * Written By Michael Sandrof
- * Copyright(c) 1990 
- * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
+ * Copyright (c) 1990 Michael Sandroff.
+ * Copyright (c) 1991, 1992 Troy Rollo.
+ * Copyright (c) 1992-1996 Matthew Green.
+ * Copyright  1994 Jake Khuon.
+ * Copyright  1993, 2003 EPIC Software Labs.
+ * All rights reserved.
  *
- * Reformatted by Matthew Green, November 1992.
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notices, the above paragraph (the one permitting redistribution),
+ *    this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ * 3. The names of the author(s) may not be used to endorse or promote
+ *    products derived from this software without specific prior written
+ *    permission.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
+ * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
+ * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
+ * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
+ * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
+ * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+ * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
+ * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
+ * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
+ * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ * SUCH DAMAGE.
  */
 #include "irc.h"
 
@@ -13,19 +40,19 @@
 /*
  * irc_version is what $J returns, its the common-name for the version.
  */
-const char irc_version[] = "EPIC4-1.0.1";
-const char useful_info[] = "epic4 1 0 1";
+const char irc_version[] = "EPIC4-2.0";
+const char useful_info[] = "epic4 2 0 0";
 
 /*
  * internal_version is what $V returns, its the integer-id for the
  * version, and corresponds to the date of release, YYYYMMDD.
  */ 
-const char internal_version[] = "20010417";
+const char internal_version[] = "20040219";
 
 /*
  * In theory, this number is incremented for every commit.
  */
-const unsigned long	commit_id = 74;
+const unsigned long	commit_id = 710;
 
 /*
  * As a way to poke fun at the current rage of naming releases after
@@ -33,10 +60,11 @@ const unsigned long	commit_id = 74;
  * reality, I have decided to start doing that with EPIC.  These names
  * are intentionally and maliciously silly.  Complaints will be ignored.
  */
-const char ridiculous_version_name[] = "Odyssey";
+const char ridiculous_version_name[] = "Circumlocution";
 
 #define __need_putchar_x__
 #include "status.h"
+#include "clock.h"
 #include "dcc.h"
 #include "names.h"
 #include "vars.h"
@@ -74,6 +102,9 @@ int		irc_port = IRC_PORT;
 /* Set if ircII should usurp flow control, unset if not.  Probably bogus. */
 int		use_flow_control = 1;
 
+/* Set if ircII should turn on IEXTEN, unset to suppress */
+int		use_iexten = -1;
+
 /* 
  * When a numeric is being processed, this holds the negative value
  * of the numeric.  Its negative so if we pass it to do_hook, it can
@@ -109,10 +140,10 @@ int		quick_startup = 0;
 int		dont_connect = 0;
 
 /* Set to the current time, each time you press a key. */
-time_t		idle_time = 0;
+Timeval		idle_time = { 0, 0 };
 
 /* Set to the time the client booted up */
-time_t		start_time;
+Timeval		start_time;
 
 /* The number of child processes still unreaped. */
 int		child_dead = 0;
@@ -120,9 +151,6 @@ int		child_dead = 0;
 /* Set if the current output is from a trusted source */
 int		trusted_output = 1;
 
-/* Whether or not we're in CPU saver mode */
-int		cpu_saver = 0;
-
 /* Set to 0 when you want to suppress all beeps (such as window repaints) */
 int		global_beep_ok = 1;
 
@@ -151,7 +179,13 @@ int		inhibit_logging = 0;
 int		loading_global = 0;
 
 /* This is reset every time io() is called.  Use this to save calls to time */
-time_t		now = 0;
+Timeval		now = {0, 0};
+
+/* When you "highlight ignore", this is what is used to embellish */
+char *		highlight_char;
+
+/* Output which is displayed without modification by the user */
+int		privileged_output = 0;
 
 /*
  * If set, outbound connections will be bind()ed to the address
@@ -164,16 +198,19 @@ time_t		now = 0;
  * set to addresses that do not belong to the current hostname.
  * If that happens, outbound connections will fail, and its not my fault.
  */
-char		*LocalHostName = NULL;
-struct	in_addr	LocalHostAddr;
+char *		LocalHostName = NULL;
+ISA *		LocalIPv4Addr = NULL;
+#ifdef INET6
+ISA6 *		LocalIPv6Addr = NULL;
+#endif
 
 int		inbound_line_mangler = 0,
 		outbound_line_mangler = 0;
 
-char		*invite_channel = (char *) 0,	/* last channel of an INVITE */
-		*ircrc_file = (char *) 0,	/* full path .ircrc file */
+char		*startup_file = NULL,		/* full path .epicrc file */
+		*epicrc_file = NULL,		/* full path .epicrc file */
+		*ircrc_file = NULL,		/* full path .ircrc file */
 		*my_path = (char *) 0,		/* path to users home dir */
-		*irc_path = (char *) 0,		/* paths used by /load */
 		*irc_lib = (char *) 0,		/* path to the ircII library */
 		*default_channel = NULL,	/* Channel to join on connect */
 		nickname[NICKNAME_LEN + 1],	/* users nickname */
@@ -194,7 +231,9 @@ char		*invite_channel = (char *) 0,	/* last channel of an INVITE */
 		comma[] = ",",
 		*cut_buffer = (char *) 0;	/* global cut_buffer */
 
-fd_set		readables, writables;
+fd_set		readables, held_readables;
+fd_set		writables, held_writables;
+int		global_max_fd = -1;
 
 
 static		char	switch_help[] =
@@ -208,8 +247,10 @@ static		char	switch_help[] =
 #endif
 "      -B\tLoads your .ircrc file before you connect to a server.      \n\
       -d\tThe program should run in ``dumb mode'' (no fancy screen)   \n\
-      -f\tThe program wont mess with your flow control                \n\
+      -f\tThe program won't mess with your flow control                \n\
       -F\tThe program will mess with your flow control                \n\
+      -o\tThe program wll turn on IEXTEN terminal setting (^V/^O)     \n\
+      -O\tThe program will turn off IEXTEN (so you can bind ^V/^O)    \n\
       -h\tPrint this help message                                     \n\
       -q\tThe program will not load your .ircrc file                  \n\
       -s\tThe program will not connect to a server upon startup       \n\
@@ -225,17 +266,24 @@ static		char	switch_help[] =
 
 
 
-SIGNAL_HANDLER(sig_irc_exit)
+static SIGNAL_HANDLER(sig_irc_exit)
 {
 	irc_exit (1, NULL);
 }
 
 /* irc_exit: cleans up and leaves */
-void	irc_exit (int really_quit, char *format, ...)
+void	irc_exit (int really_quit, const char *format, ...)
 {
 	char 	buffer[BIG_BUFFER_SIZE];
 	char *	sub_format;
 	int	old_window_display = window_display;
+	int	value;
+#ifdef PERL
+	extern void perlstartstop(int);
+#endif
+#ifdef TCL
+	extern void tclstartstop(int);
+#endif
 
 	/*
 	 * If we get called recursively, something is hosed.
@@ -276,8 +324,16 @@ void	irc_exit (int really_quit, char *format, ...)
 
 	/* Do some clean up */
 	do_hook(EXIT_LIST, "%s", buffer);
+#ifdef TCL
+	tclstartstop(0);
+#endif
+#ifdef PERL
+	perlstartstop(0);  /* In case there's perl code in the exit hook. */
+#endif
 	close_all_servers(buffer);
-	logger(0);
+	value = 0;
+	logger(&value);
+	get_child_exit(-1);  /* In case some children died in the exit hook. */
 	clean_up_processes();
 
 	/* Arrange to have the cursor on the input line after exit */
@@ -292,12 +348,13 @@ void	irc_exit (int really_quit, char *format, ...)
 	/* Try to free as much memory as possible */
 	window_display = 0;
 	dumpcmd(NULL, NULL, NULL);
-	set_lastlog_size(0);
-	set_history_size(0);
+	set_lastlog_size(&value);
+	set_history_size(&value);
 	remove_channel(NULL, 0);
 	destroy_call_stack();
 	remove_bindings();
 	delete_all_windows();
+	destroy_server_list();
 	window_display = old_window_display;
 	fprintf(stdout, "\r");
 	fflush(stdout);
@@ -318,15 +375,15 @@ volatile int	dead_children_processes;
  * This is needed so that the fork()s we do to read compressed files dont
  * sit out there as zombies and chew up our fd's while we read more.
  */
-SIGNAL_HANDLER(child_reap)
+static SIGNAL_HANDLER(child_reap)
 {
-	dead_children_processes++;
+	dead_children_processes = 1;
 }
 
 volatile int	segv_recurse = 0;
 
 /* sigsegv: something to handle segfaults in a nice way */
-SIGNAL_HANDLER(coredump)
+static SIGNAL_HANDLER(coredump)
 {
 	if (segv_recurse++)
 		exit(1);
@@ -351,22 +408,22 @@ to it, please dont worry about filling it out.  You might try talking	\n\
 to the person who is in charge of IRC at your site and see if you can	\n\
 get them to help you.							\n\
 									\n\
-This version of IRC II is  --->[%s]					\n\
+This version of IRC II is  --->[%s (%lu)]				\n\
 The date of release is     --->[%s]					\n\
 									\n\
 * * * * * * * * * * * * * * * * * * * * * * * *				\n\
-The program will now terminate.						\n", irc_version, internal_version);
+The program will now terminate.						\n", irc_version, commit_id, internal_version);
 
 		fflush(stdout);
 		panic_dump_call_stack();
 	}
 
         if (x_debug & DEBUG_CRASH)
-                irc_exit(0, "Hmmm. %s has another bug.  Go figure...",
-			irc_version);
+                irc_exit(0, "Hmmm. %s (%lu) has another bug.  Go figure...",
+			irc_version, commit_id);
         else
-                irc_exit(1, "Hmmm. %s has another bug.  Go figure...",
-			irc_version);
+                irc_exit(1, "Hmmm. %s (%lu) has another bug.  Go figure...",
+			irc_version, commit_id);
 }
 
 /*
@@ -402,7 +459,7 @@ void irc_quit (char unused, char *not_used)
  * the program *is* frozen -- if it doesnt die when you do this, then the
  * program is functioning correctly (ie, something else is wrong)
  */
-SIGNAL_HANDLER(cntl_c)
+static SIGNAL_HANDLER(cntl_c)
 {
 	/* after 5 hits, we stop whatever were doing */
 	if (cntl_c_hit++ >= 4)
@@ -411,12 +468,12 @@ SIGNAL_HANDLER(cntl_c)
 		kill(getpid(), SIGALRM);
 }
 
-SIGNAL_HANDLER(nothing)
+static SIGNAL_HANDLER(nothing)
 {
 	/* nothing to do! */
 }
 
-SIGNAL_HANDLER(sig_user1)
+static SIGNAL_HANDLER(sig_user1)
 {
 	say("Got SIGUSR1, closing DCC connections and EXECed processes");
 	close_all_dcc();
@@ -425,7 +482,7 @@ SIGNAL_HANDLER(sig_user1)
 
 static	void	show_version (void)
 {
-	printf("ircII %s (Date of release: %s)\n\r",irc_version,internal_version);
+	printf("ircII %s (Commit id: %lu) (Date of release: %s)\n\r", irc_version, commit_id, internal_version);
 	exit (0);
 }
 
@@ -451,9 +508,10 @@ static	void	parse_args (int argc, char **argv)
 	int ch;
 	int add_servers = 0;
 	struct passwd *entry;
-	struct hostent *hp;
 	char *ptr = (char *) 0;
 	char *tmp_hostname = NULL;
+	char *the_path = NULL;
+	char *translation_path = NULL;
 
 	extern char *optarg;
 	extern int optind;
@@ -473,11 +531,11 @@ static	void	parse_args (int argc, char **argv)
 		{
 			if ((ptr = strchr(entry->pw_gecos, ',')))
 				*ptr = 0;
-			strmcpy(realname, entry->pw_gecos, REALNAME_LEN);
+			strlcpy(realname, entry->pw_gecos, sizeof realname);
 		}
 
 		if (entry->pw_name && *(entry->pw_name))
-			strmcpy(username, entry->pw_name, NAME_LEN);
+			strlcpy(username, entry->pw_name, sizeof username);
 
 		if (entry->pw_dir && *(entry->pw_dir))
 			malloc_strcpy(&my_path, entry->pw_dir);
@@ -485,7 +543,7 @@ static	void	parse_args (int argc, char **argv)
 
 
 	if ((ptr = getenv("IRCNICK")))
-		strmcpy(nickname, ptr, NICKNAME_LEN);
+		strlcpy(nickname, ptr, sizeof nickname);
 
 	/*
 	 * We now allow users to use IRCUSER or USER if we couldnt get the
@@ -499,26 +557,26 @@ static	void	parse_args (int argc, char **argv)
 	 */
 	if (!*username)
 		if ((ptr = getenv("LOGNAME")) && *ptr)
-			strmcpy(username, ptr, NAME_LEN);
+			strlcpy(username, ptr, sizeof username);
 
 #ifndef ALLOW_USER_SPECIFIED_LOGIN
 	if (!*username)
 #endif
 		if ((ptr = getenv("IRCUSER")) && *ptr) 
-			strmcpy(username, ptr, NAME_LEN);
+			strlcpy(username, ptr, sizeof username);
 #ifdef ALLOW_USER_SPECIFIED_LOGIN
 		else if (*username)
 			;
 #endif
 		else if ((ptr = getenv("USER")) && *ptr) 
-			strmcpy(username, ptr, NAME_LEN);
+			strlcpy(username, ptr, sizeof username);
 		else if ((ptr = getenv("HOME")) && *ptr)
 		{
 			char *ptr2 = strrchr(ptr, '/');
 			if (ptr2)
-				strmcpy(username, ptr2, NAME_LEN);
+				strlcpy(username, ptr2, sizeof username);
 			else
-				strmcpy(username, ptr, NAME_LEN);
+				strlcpy(username, ptr, sizeof username);
 		}
 		else
 		{
@@ -529,11 +587,11 @@ static	void	parse_args (int argc, char **argv)
 		}
 
 	if ((ptr = getenv("IRCNAME")))
-		strmcpy(realname, ptr, REALNAME_LEN);
+		strlcpy(realname, ptr, sizeof realname);
 	else if ((ptr = getenv("NAME")))
-		strmcpy(realname, ptr, REALNAME_LEN);
+		strlcpy(realname, ptr, sizeof realname);
 	else if (!*realname)
-		strmcpy(realname, "*Unknown*", REALNAME_LEN);
+		strlcpy(realname, "*Unknown*", sizeof realname);
 
 	if ((ptr = getenv("HOME")))
 		malloc_strcpy(&my_path, ptr);
@@ -545,36 +603,47 @@ static	void	parse_args (int argc, char **argv)
 	if ((ptr = getenv("IRCPORT")))
 		irc_port = my_atol(ptr);
 
-	if ((ptr = getenv("IRCRC")))
-		ircrc_file = m_strdup(ptr);
+	if ((ptr = getenv("EPICRC")))
+		epicrc_file = malloc_strdup(ptr);
 	else
-		ircrc_file = m_2dup(my_path, IRCRC_NAME);
+		epicrc_file = malloc_strdup2(my_path, EPICRC_NAME);
+
+	if ((ptr = getenv("IRCRC")))
+		ircrc_file = malloc_strdup(ptr);
+	else
+		ircrc_file = malloc_strdup2(my_path, IRCRC_NAME);
 
 	if ((ptr = getenv("IRCLIB")))
-		irc_lib = m_2dup(ptr, "/");
+		irc_lib = malloc_strdup2(ptr, "/");
 	else
-		irc_lib = m_strdup(IRCLIB);
+		irc_lib = malloc_strdup(IRCLIB);
 
 	if ((ptr = getenv("IRCUMODE")))
-		send_umode = m_strdup(ptr);
+		send_umode = malloc_strdup(ptr);
 
 	if ((ptr = getenv("IRCPATH")))
-		irc_path = m_strdup(ptr);
+		the_path = malloc_strdup(ptr);
 	else
-		irc_path = m_sprintf(DEFAULT_IRCPATH, irc_lib);
+		the_path = malloc_sprintf(NULL, DEFAULT_IRCPATH, irc_lib);
 
-	set_string_var(LOAD_PATH_VAR, irc_path);
-	new_free(&irc_path);
+	set_string_var(LOAD_PATH_VAR, the_path);
+	new_free(&the_path);
 
 	if ((ptr = getenv("IRCHOST")) && *ptr)
 		tmp_hostname = ptr;
 
+	if ((ptr = getenv("IRCTRANSLATIONPATH")))
+		translation_path = malloc_strdup(ptr);
+	else
+		translation_path = malloc_strdup2(IRCLIB, "/translation/");
 
+	set_string_var(TRANSLATION_PATH_VAR, translation_path);
+	new_free(&translation_path);
 
 	/*
 	 * Parse the command line arguments.
 	 */
-	while ((ch = getopt(argc, argv, "aBbc:dfFhH:l:L:n:p:qsvxz:")) != EOF)
+	while ((ch = getopt(argc, argv, "aBbc:dfFhH:l:L:n:oOp:qsvxz:")) != EOF)
 	{
 		switch (ch)
 		{
@@ -594,13 +663,21 @@ static	void	parse_args (int argc, char **argv)
 				use_flow_control = 0;
 				break;
 
+			case 'o': /* Use IEXTEN */
+				use_iexten = 1;
+				break;
+
+			case 'O': /* dont use IEXTEN */
+				use_iexten = 0;
+				break;
+
 			case 'd': /* use dumb mode */
 				dumb_mode = 1;
 				break;
 
 			case 'l': /* Load some file instead of ~/.ircrc */
 			case 'L': /* Same as above. Doesnt work like before */
-				malloc_strcpy(&ircrc_file, optarg);
+				malloc_strcpy(&epicrc_file, optarg);
 				break;
 
 			case 'a': /* add server, not replace */
@@ -627,7 +704,7 @@ static	void	parse_args (int argc, char **argv)
 				break;
 
 			case 'n':
-				strmcpy(nickname, optarg, NICKNAME_LEN);
+				strlcpy(nickname, optarg, sizeof nickname);
 				break;
 
 			case 'x': /* x_debug flag */
@@ -636,7 +713,7 @@ static	void	parse_args (int argc, char **argv)
 
 			case 'z':
 #ifdef ALLOW_USER_SPECIFIED_LOGIN
-				strmcpy(username, optarg, NAME_LEN);
+				strlcpy(username, optarg, sizeof username);
 #endif
 				break;
 
@@ -663,20 +740,29 @@ static	void	parse_args (int argc, char **argv)
 	argv += optind;
 
 	if (argc && **argv && !strchr(*argv, '.'))
-		strmcpy(nickname, *argv++, NICKNAME_LEN), argc--;
+		strlcpy(nickname, *argv++, sizeof nickname), argc--;
 
 	/*
  	 * "nickname" needs to be valid before we call build_server_list,
 	 * so do a final check on whatever nickname we're going to use.
 	 */
 	if (!*nickname)
-		strmcpy(nickname, username, NICKNAME_LEN);
+		strlcpy(nickname, username, sizeof nickname);
 
 	for (; *argv; argc--, argv++)
 		if (**argv)
-			build_server_list(*argv);
-
+			build_server_list(*argv, NULL);
 
+	if (!use_input && quick_startup)
+	{
+		fprintf(stderr, "Cannot use -b and -q at the same time\n");
+		exit(1);
+	}
+	if (!use_input && dont_connect)
+	{
+		fprintf(stderr, "Cannot use -b and -s at the same time\n");
+		exit(1);
+	}
 
 	if (!check_nickname(nickname, 1))
 	{
@@ -689,15 +775,15 @@ static	void	parse_args (int argc, char **argv)
 	 * Find and build the server lists...
 	 */
 	if ((ptr = getenv("IRCSERVER")))
-		build_server_list(ptr);
+		build_server_list(ptr, NULL);
 
 	if (!server_list_size() || add_servers)
 	{
 		read_server_file();
 		if (!server_list_size())
 		{
-			ptr = m_strdup(DEFAULT_SERVER);
-			build_server_list(ptr);
+			ptr = malloc_strdup(DEFAULT_SERVER);
+			build_server_list(ptr, NULL);
 			new_free(&ptr);
 		}
 	}
@@ -705,18 +791,12 @@ static	void	parse_args (int argc, char **argv)
 	/*
 	 * Figure out our virtual hostname, if any.
 	 */
+	LocalHostName = NULL;
 	if (tmp_hostname)
 	{
-		fprintf(stderr, "You specified a hostname of [%s]\n", tmp_hostname);
-		memset(&LocalHostAddr, 0, sizeof(LocalHostAddr));
-		if ((hp = gethostbyname(tmp_hostname)))
-		{
-			memmove((void *)&LocalHostAddr, hp->h_addr, sizeof(LocalHostAddr));
-			fprintf(stderr, "Ok.  I'll buy that.\n");
-			LocalHostName = m_strdup(tmp_hostname);
-		}
-		else
-			fprintf(stderr, "I can't configure for that address! (Trying normal hostname...)\n");
+		char *s = switch_hostname(tmp_hostname);
+		fprintf(stderr, "%s\n", s);
+		new_free(&s);
 	}
 
 	/*
@@ -753,21 +833,17 @@ static	void	parse_args (int argc, char **argv)
 void	io (const char *what)
 {
 static	const	char	*caller[51] = { NULL }; /* XXXX */
-	static	int	first_time = 1,
-			level = 0;
-static	struct	timeval	clock_timeout,
-			right_away,
-			timer,
-			*timeptr = NULL;
-		int	hold_over;
-		int	dccs;
-		fd_set	rd;
+static	int		level = 0,
+			old_level = 0,
+			last_warn = 0;
+static 	const Timeval	right_away = { 0, 0 };
+
+	Timeval		timer;
+	fd_set		rd, wd;
 
-	static	int	old_level = 0;
-	static	int	last_warn = 0;
 
 	level++;
-	now = time(NULL);
+	get_time(&now);
 
 	/* Don't let this accumulate behind the user's back. */
 	cntl_c_hit = 0;
@@ -784,7 +860,7 @@ static	struct	timeval	clock_timeout,
 	if (level && (level - last_warn == 5))
 	{
 		last_warn = level;
-		yell("io's nesting level is [%d],  [%s]<-[%s]<-[%s]<-[%s]<-[%s]", level, what, caller[level-1], caller[level-2], caller[level-3], caller[level-4]);
+		yell("io's recursion level is [%d],  [%s]<-[%s]<-[%s]<-[%s]<-[%s]", level, what, caller[level-1], caller[level-2], caller[level-3], caller[level-4]);
 		if (level % 50 == 0)
 			panic("Ahoy there matey!  Abandon ship!");
 	}
@@ -794,65 +870,46 @@ static	struct	timeval	clock_timeout,
 
 	caller[level] = what;
 
-	/* first time we run this function, set up the timeouts */
-	if (first_time)
-	{
-		first_time = 0;
-
-		/*
-		 * time delay for updating of internal clock
-		 *
-		 * Instead of looking every 15 seconds and seeing if
-		 * the clock has changed, we now figure out how much
-		 * time there is to the next clock change and then wait
-		 * until then.  There is a small performance penalty 
-		 * in actually calculating when the next minute will tick, 
-		 * but that will be offset by the fact that we will only
-		 * call select() once a minute instead of 4 times.
-		 * Plus you get the benefit of the clock actually changing
-		 * on time rather than up to 15 seconds later.
-		 */
-		clock_timeout.tv_usec = 0L;
-
-		right_away.tv_usec = 0L;
-		right_away.tv_sec = 0L;
-
-		timer.tv_usec = 0L;
-	}
-
-	/* CHECK FOR CPU SAVER MODE */
-	if (!cpu_saver && get_int_var(CPU_SAVER_AFTER_VAR))
-		if (now - idle_time > get_int_var(CPU_SAVER_AFTER_VAR) * 60)
-			cpu_saver_on(0, NULL);
+#if 1
+	/* 
+	 * XXXXXXX! It pains me greatly to call this here.
+	 * This is an experimental test to see if i can get
+	 * away with not having to call update_all_windows()
+	 * in about a zillion other places by doing it here.
+	 */
+	update_all_windows();
+#endif
 
 	/* SET UP FD SETS */
 	rd = readables;
+	wd = writables;
 
-	clock_timeout.tv_sec = 60 - now % 60;
-	if (cpu_saver && get_int_var(CPU_SAVER_EVERY_VAR))
-		clock_timeout.tv_sec += (get_int_var(CPU_SAVER_EVERY_VAR) - 1) * 60;
+	/* If there is a timer that expires sooner, wait for that */
+	/* There is now a timer at all times, so this is our baseline */
+	timer = TimerTimeout();
 
-	if (!timeptr)
-		timeptr = &clock_timeout;
-	timer.tv_sec = TimerTimeout();
-	if (timer.tv_sec <= timeptr->tv_sec)
-		timeptr = &timer;
-
-	if ((hold_over = unhold_windows()))
-		timeptr = &right_away;
-	if ((dccs = dcc_dead()))		/* XXX HACK! XXX */
-		timeptr = &right_away;
+	/* If for any reason the timeout is negative, do a poll */
+	if (time_diff(right_away, timer) < 0)
+		timer = right_away;
 
 	/* GO AHEAD AND WAIT FOR SOME DATA TO COME IN */
-	switch (new_select(&rd, NULL, timeptr))
+	switch (new_select(&rd, &wd, &timer))
 	{
 		/* Timeout -- nothing interesting. */
 		case 0:
+		{
+			get_time(&now);
+#ifdef HAVE_SSL
+			/* Yes, this is slow, but we have to check for this */
+			do_server(&rd, &wd);
+#endif
 			break;
+		}
 
 		/* Interrupted system call -- check for SIGINT */
 		case -1:
 		{
+			get_time(&now);
 			if (cntl_c_hit)		/* SIGINT is useful */
 			{
 				edit_char('\003');
@@ -866,46 +923,29 @@ static	struct	timeval	clock_timeout,
 		/* Check it out -- something is on one of our descriptors. */
 		default:
 		{
+			get_time(&now);
 			make_window_current(NULL);
-			dcc_check(&rd);
-			do_server(&rd);
-			do_processes(&rd);
-			do_screens(&rd);
+			dcc_check(&rd, &wd);
+			do_server(&rd, &wd);
+			do_processes(&rd, &wd);
+			do_screens(&rd, &wd);
 			break;
 		} 
 	}
 
-	if (dccs)			/* XXX HACK XXX */
-		dcc_check(&rd);		/* XXX HACK XXX */
-
 	ExecuteTimers();
 	get_child_exit(-1);
 	if (level == 1 && need_defered_commands)
 		do_defered_commands();
 
 	cursor_to_input();
-	timeptr = &clock_timeout;
-
-	if (update_clock(RESET_TIME))
-	{
-		/* 
-		 * Do notify first so the server is working on it
-		 * while we do the other crap.
-		 */
-		do_notify();
-
-		if (get_int_var(CLOCK_VAR) || check_mail_status(NULL))
-		{
-			update_all_status();
-			cursor_to_input();
-		}
-	}
 
 	/* (set in term.c) -- we need to redraw the screen */
 	if (need_redraw)
-		refresh_screen('\0', NULL);
+		refresh_a_screen(main_screen);
 
 	window_check_channels();
+	update_all_windows();
 	alloca(0);
 	caller[level] = NULL;
 	level--;
@@ -923,7 +963,7 @@ static void check_password (void)
 #define INPUT_PASSWD_LEN 15
 	char 	input_passwd[INPUT_PASSWD_LEN];
 #ifdef HAVE_GETPASS
-	strlcpy(input_passwd, getpass("Passwd: "), INPUT_PASSWD_LEN);
+	strlcpy(input_passwd, getpass("Passwd: "), sizeof input_password);
 #else
 	fprintf(stderr, "Passwd: ");
 	fgets(input_passwd, INPUT_PASSWD_LEN - 1, stdin);
@@ -1065,119 +1105,23 @@ static void check_invalid_host (void)
 	return;
 }
 
-/*
- * I moved this here, because it didnt really belong in status.c
- */
-/* 
- * XXX This is bogus -- it gets set when you do /set clock_format.
- * The reason we use a global variable is to cut down on the number of
- * calls to get_string_var, which is truly bogus, but neccesary for any
- * semblance of efficiency.
- */
-	char	*time_format = (char *) 0;	/* XXX Bogus XXX */
-static	char	*strftime_24hour = "%R";
-static	char	*strftime_12hour = "%I:%M%p";
-	int	broken_clock = 0;
-
-/* update_clock: figures out the current time and returns it in a nice format */
-char	*update_clock (int flag)
-{
-static		char	time_str[61];
-static		int	min = -1;
-static		int	hour = -1;
-
-	struct	timeval	tv;
-static 	struct 	tm	time_val;
-static		time_t	last_minute = -1;
-		time_t	hideous;
-		int	new_minute = 0;
-
-	if (x_debug & DEBUG_BROKEN_CLOCK)
-	{
-		snprintf(time_str, 60, "%c12%c:%c00AM%c",
-			BLINK_TOG, BLINK_TOG, BLINK_TOG, BLINK_TOG);
-
-		if (flag == GET_TIME)
-			return(time_str);
-		else
-			return NULL;
-	}
-
-	/*
-	 * This is cheating because we only call localtime() once per minute.
-	 * This implicitly assumes that you only want to get the localtime
-	 * info once every minute, which we do.  If you wanted to get it every
-	 * second (which we dont), you DONT WANT TO DO THIS!
-	 */
-	get_time(&tv);
-	hideous = tv.tv_sec;
-#ifndef NO_CHEATING
-	if (hideous / 60 != last_minute)
-	{
-		last_minute = hideous / 60;
-		time_val = *localtime(&hideous);
-	}
-#else
-	time_val = *localtime(&hideous);
-#endif
-
-	if (flag == RESET_TIME || time_val.tm_min != min || time_val.tm_hour != hour)
-	{
-		if (time_format)	/* XXX Bogus XXX */
-			strftime(time_str, 60, time_format, &time_val);
-		else if (get_int_var(CLOCK_24HOUR_VAR))
-			strftime(time_str, 60, strftime_24hour, &time_val);
-		else
-			strftime(time_str, 60, strftime_12hour, &time_val);
-
-		if ((time_val.tm_min != min) || (time_val.tm_hour != hour))
-		{
-			int	old_server = from_server;
-
-			new_minute = 1;
-			hour = time_val.tm_hour;
-			min = time_val.tm_min;
-
-			from_server = primary_server;
-			do_hook(TIMER_LIST, "%02d:%02d", hour, min);
-			do_hook(IDLE_LIST, "%ld", (tv.tv_sec - idle_time) / 60);
-			from_server = old_server;
-		}
-
-		if (flag != RESET_TIME || new_minute)
-			return time_str;
-		else
-			return NULL;
-	}
-
-	if (flag == GET_TIME)
-		return time_str;
-	else
-		return NULL;
-}
-
-void	reset_clock (char *unused)
-{
-	update_clock(RESET_TIME);
-	update_all_status();
-}
-
-
+/*************************************************************************/
 int 	main (int argc, char *argv[])
 {
 #ifdef SOCKS
 	SOCKSinit(argv[0]);
 #endif
-        start_time = time(NULL);
+        get_time(&start_time);
 	check_password();
 	check_valid_user();
 	check_invalid_host();
 	parse_args(argc, argv);
+	init_binds();
 	init_keys();
 
 	fprintf(stderr, "EPIC Version 4 -- %s\n", ridiculous_version_name);
-	fprintf(stderr, "EPIC Software Labs (2000)\n");
-	fprintf(stderr, "Version (%s) -- Date (%s)\n", irc_version, internal_version);
+	fprintf(stderr, "EPIC Software Labs (2004)\n");
+	fprintf(stderr, "Version (%s), Commit Id (%lu) -- Date (%s)\n", irc_version, commit_id, internal_version);
 	fprintf(stderr, "%s\n", compile_info);
 	fprintf(stderr, "Process [%d]", getpid());
 	if (isatty(0))
@@ -1188,6 +1132,8 @@ int 	main (int argc, char *argv[])
 
 	FD_ZERO(&readables);
 	FD_ZERO(&writables);
+	FD_ZERO(&held_readables);
+	FD_ZERO(&held_writables);
 
 	/* If we're a bot, do the bot thing. */
 	if (!use_input && fork())
@@ -1228,37 +1174,27 @@ int 	main (int argc, char *argv[])
 	}
 
 	/* Get the terminal-specific keybindings now */
-	init_keys2();
+	init_termkeys();
 
 	/* The all-collecting stack frame */
 	make_local_stack("TOP");
 
 	/* XXXX Move this somewhere else eventually XXXX */
 	if (load_ircrc_right_away)
-	{
-		char	buffer[7];
-		strcpy(buffer, "global");
+		load_ircrc();
 
-		loading_global = 1;
-		load("LOAD", buffer, empty_string);
-		loading_global = 0;
-
-		/* read the .ircrc file */
-		if (access(ircrc_file, R_OK) == 0 && !quick_startup)
-			load("LOAD", ircrc_file, empty_string);
-
-		ircrc_loaded = 1;
-	}
+	set_input(empty_string);
+	set_input_prompt(get_string_var(INPUT_PROMPT_VAR));
 
 	if (dont_connect)
 		display_server_list();		/* Let user choose server */
 	else
-		reconnect(-1);			/* Connect to default server */
+		reconnect(NOSERV, 0);		/* Connect to default server */
 
-	time(&idle_time);
-	set_input(empty_string);
-	set_input_prompt(get_string_var(INPUT_PROMPT_VAR));
-	for (;;)
+	get_time(&idle_time);
+	reset_system_timers();
+
+	for (;;system_exception = 0)
 		io("main");
 	/* NOTREACHED */
 }
