@@ -1,12 +1,38 @@
+/* $EPIC: status.c,v 1.35 2003/12/07 20:16:52 jnelson Exp $ */
 /*
  * status.c: handles the status line updating, etc for IRCII 
  *
- * Written By Michael Sandrof
- * Extensive modifications by Jake Khuon and others (epic software labs)
+ * Copyright (c) 1990 Michael Sandroff.
+ * Copyright (c) 1991, 1992 Troy Rollo.
+ * Copyright (c) 1992-1996 Matthew Green.
+ * Copyright  1994 Jake Khuon.
+ * Copyright  1995, 2003 EPIC Software Labs.
+ * All rights reserved.
  *
- * Copyright(c) 1990  Michael Sandrof
- * Copyright 1997, 1998 EPIC Software Labs
- * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notices, the above paragraph (the one permitting redistribution),
+ *    this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ * 3. The names of the author(s) may not be used to endorse or promote
+ *    products derived from this software without specific prior written
+ *    permission.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
+ * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
+ * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
+ * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
+ * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
+ * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+ * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
+ * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
+ * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
+ * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ * SUCH DAMAGE.
  */
 
 #define __need_putchar_x__
@@ -26,6 +52,7 @@
 #include "names.h"
 #include "ircaux.h"
 #include "alias.h"
+#include "clock.h"
 
 #ifdef Char
 #undef Char
@@ -34,9 +61,9 @@
 
 /*
  * Maximum number of "%" expressions in a status line format.  If you change
- * this number, you must manually change the sprintf() in make_status 
+ * this number, you must manually change the snprintf() in make_status 
  */
-#define STATUS_FUNCTION(x) static Char * ## x (Window *window, int map, int key)
+#define STATUS_FUNCTION(x) Char * x (Window *window, int map, int key)
 #define MAX_FUNCTIONS 40
 #define MAX_STATUS_USER 39
 
@@ -44,6 +71,7 @@ STATUS_FUNCTION(status_nickname);
 STATUS_FUNCTION(status_query_nick);
 STATUS_FUNCTION(status_right_justify);
 STATUS_FUNCTION(status_chanop);
+STATUS_FUNCTION(status_ssl);
 STATUS_FUNCTION(status_channel);
 STATUS_FUNCTION(status_server);
 STATUS_FUNCTION(status_mode);
@@ -54,6 +82,7 @@ STATUS_FUNCTION(status_away);
 STATUS_FUNCTION(status_oper);
 STATUS_FUNCTION(status_user);
 STATUS_FUNCTION(status_dcc);
+STATUS_FUNCTION(status_dcc_all);
 STATUS_FUNCTION(status_hold);
 STATUS_FUNCTION(status_version);
 STATUS_FUNCTION(status_clock);
@@ -61,6 +90,7 @@ STATUS_FUNCTION(status_hold_lines);
 STATUS_FUNCTION(status_window);
 STATUS_FUNCTION(status_mail);
 STATUS_FUNCTION(status_refnum);
+STATUS_FUNCTION(status_refnum_real);
 STATUS_FUNCTION(status_null_function);
 STATUS_FUNCTION(status_notify_windows);
 STATUS_FUNCTION(status_voice);
@@ -71,6 +101,7 @@ STATUS_FUNCTION(status_scroll_info);
 STATUS_FUNCTION(status_windowspec);
 STATUS_FUNCTION(status_percent);
 STATUS_FUNCTION(status_test);
+STATUS_FUNCTION(status_swappable);
 
 /* These are used as placeholders for some expandos */
 static	char	*mode_format 		= (char *) 0;
@@ -85,8 +116,16 @@ static	char	*nick_format		= (char *) 0;
 static	char	*server_format 		= (char *) 0;
 static	char	*notify_format 		= (char *) 0;
 
-Status	main_status;
-int	main_status_init = 0;
+	Status	main_status;
+	int	main_status_init = 0;
+static void	init_status	(void);
+
+/*
+ * Status updates are not permitted until we are connected to a server.
+ * Got_initial_version_28 will set this to 1 when we first get connected
+ * to a server.
+ */
+int     status_updates_permitted = 0;
 
 /*
  * This is the list of possible expandos.  Note that you should not use
@@ -132,6 +171,7 @@ struct status_formats status_expandos[] = {
 { 0, '=', status_voice,		NULL, 			-1 },
 { 0, '>', status_right_justify,	NULL, 			-1 },
 { 0, '@', status_chanop,	NULL, 			-1 },
+{ 0, '|', status_ssl,		NULL,			-1 },
 { 0, '0', status_user,		NULL, 			-1 },
 { 0, '1', status_user,		NULL, 			-1 },
 { 0, '2', status_user,		NULL, 			-1 },
@@ -152,10 +192,14 @@ struct status_formats status_expandos[] = {
 { 1, '7', status_user,		NULL, 			-1 },
 { 1, '8', status_user,		NULL, 			-1 },
 { 1, '9', status_user,		NULL, 			-1 },
+{ 1, 'D', status_dcc_all,	NULL, 			-1 },
 { 1, 'F', status_notify_windows,&notify_format,		STATUS_NOTIFY_VAR },
 { 1, 'K', status_scroll_info,	NULL,			-1 },
+{ 1, 'R', status_refnum_real,   NULL, 			-1 },
 { 1, 'S', status_server,        &server_format,     	STATUS_SERVER_VAR },
 { 1, 'T', status_test,		NULL,			-1 },
+{ 1, 'W', status_swappable,	NULL,			-1 },
+{ 1, '+', status_mode,		&mode_format,       	STATUS_MODE_VAR },
 { 2, '0', status_user,	 	NULL, 			-1 },
 { 2, '1', status_user,	 	NULL, 			-1 },
 { 2, '2', status_user,	 	NULL, 			-1 },
@@ -185,7 +229,7 @@ struct status_formats status_expandos[] = {
  * sub-portions of the status line to a format statement specially designed
  * for that sub-portions.  convert_sub_format looks for a single occurence of
  * %c (where c is passed to the function). When found, it is replaced by "%s"
- * for use is a sprintf.  All other occurences of % followed by any other
+ * for use in a snprintf.  All other occurences of % followed by any other
  * character are left unchanged.  Only the first occurence of %c is
  * converted, all subsequence occurences are left unchanged.  This routine
  * mallocs the returned string. 
@@ -230,7 +274,7 @@ char	*convert_sub_format (const char *format, char c)
 	}
 
 	buffer[pos] = 0;
-	return m_strdup(buffer);
+	return malloc_strdup(buffer);
 }
 
 
@@ -238,16 +282,16 @@ char	*convert_sub_format (const char *format, char c)
  * This walks a raw format string and parses out any expandos that it finds.
  * An expando is handled by pre-fetching any string variable that is used
  * by the callback function, the callback function is registered, and a
- * %s format is put in the sprintf()-able return value (stored in buffer).
+ * %s format is put in the snprintf()-able return value (stored in buffer).
  * All other characters are copied as-is to the return value.
  */
-void	build_status_format (Status *s, int k)
+static void	build_status_format (Status *s, int k)
 {
 	char	buffer[BIG_BUFFER_SIZE + 1];
 	int	cp;
 	int	map;
 	char	key;
-	int	i;
+	unsigned	i;
 	Char	*raw = s->line[k].raw;
 	char	*format = buffer;
 
@@ -321,10 +365,13 @@ void	build_status_format (Status *s, int k)
 }
 
 /*
- * This function does two things:
- * 1) Rebuilds the status_func[] tables for each of the three possible
- *    status_format's (single, lower double, upper double)
- * 2) Causes the status bars to be redrawn immediately.
+ * This function either rebuilds the status_func[] tables for each of the
+ * three possible status_format's (single, lower double, upper double) for
+ * a particular window (w != NULL) or for the global default status bars
+ * (w == NULL)
+ *
+ * This function should only by the functions that actually change the 
+ * status formats (build_status, new_window, and window_status_format*)
  */
 void	rebuild_a_status (Window *w)
 {
@@ -355,7 +402,7 @@ void	rebuild_a_status (Window *w)
 		 */
 		else if (w)
 		{
-			s->line[k].format = m_strdup(main_status.line[k].format);
+			s->line[k].format = malloc_strdup(main_status.line[k].format);
 			for (i = 0; i < MAX_FUNCTIONS; i++)
 			{
 			    s->line[k].func[i] = main_status.line[k].func[i];
@@ -383,7 +430,33 @@ void	rebuild_a_status (Window *w)
 	}
 }
 
-void	init_status	(void)
+/*
+ * This function causes every status bar, both global and per-window to be
+ * rebuilt, and marks every window as needing its status bar redrawn.
+ * This function can be quite expensive.
+ *
+ * This function should be called whenever you change the global status format
+ * (/SET STATUS_FORMAT*) or one of the sub-components (/SET STATUS_*)
+ * 
+ * This function is a /SET callback, so it must always take a (char *) as an
+ * argument even though we don't care about it.
+ */
+void	build_status	(const void *stuff)
+{
+	Window 	*w = NULL;
+
+	if (!main_status_init)
+		init_status();
+
+	rebuild_a_status(w);
+	while (traverse_all_windows(&w))
+		rebuild_a_status(w);
+
+	update_all_status();
+}
+
+
+static void	init_status	(void)
 {
 	int	i, k;
 
@@ -405,18 +478,14 @@ void	init_status	(void)
 	main_status_init = 1;
 }
 
-void	build_status	(char *unused)
+/*
+ * permit_status_update: sets the status_update_flag to whatever flag is.
+ */
+int     permit_status_update (int flag)
 {
-	Window 	*w = NULL;
-
-	if (!unused && !main_status_init)
-		init_status();
-
-	rebuild_a_status(w);
-	while (traverse_all_windows(&w))
-		rebuild_a_status(w);
-
-	update_all_status();
+        int old_flag = status_updates_permitted;
+        status_updates_permitted = flag;
+        return old_flag;
 }
 
 
@@ -425,7 +494,7 @@ void	build_status	(char *unused)
  * but the crecendo of complaints with regards to this just got to be too 
  * irritating, so i fixed it early.
  */
-void	make_status (Window *window)
+int	make_status (Window *window, int must_redraw)
 {
 	int	status_line;
 	u_char	buffer	    [BIG_BUFFER_SIZE + 1];
@@ -433,6 +502,11 @@ void	make_status (Window *window)
 	u_char	rhs_buffer  [BIG_BUFFER_SIZE + 1];
 	Char	*func_value [MAX_FUNCTIONS];
 	u_char	*ptr;
+	size_t	save_size;
+
+	/* We do NOT redraw status bars for hidden windows */
+	if (!window->screen || !status_updates_permitted)
+		return -1;
 
 	for (status_line = 0; status_line < window->status.double_status + 1; status_line++)
 	{
@@ -492,6 +566,9 @@ void	make_status (Window *window)
 		 */
 		for (i = 0; i < MAX_FUNCTIONS; i++)
 		{
+			if (window->screen == NULL)
+				return -1;
+
 			if (window->status.line[line].func[i] == NULL)
 				panic("status callback null.  Window [%d], line [%d], function [%d]", window->refnum, line, i);
 			func_value[i] = window->status.line[line].func[i]
@@ -537,12 +614,18 @@ void	make_status (Window *window)
 		if (get_int_var(STATUS_DOES_EXPANDOS_VAR))
 		{
 			int  af = 0;
+			int  old_fs = from_server;
 			Window *old = current_window;
+			int	owd = window_display;
 
 			current_window = window;
+			from_server = current_window->server;
+			window_display = 0;
 			str = expand_alias(buffer, empty_string, &af, NULL);
+			window_display = owd;
+			from_server = old_fs;
 			current_window = old;
-			strmcpy(buffer, str, BIG_BUFFER_SIZE);
+			strlcpy(buffer, str, sizeof buffer);
 			new_free(&str);
 		}
 
@@ -553,7 +636,7 @@ void	make_status (Window *window)
 		 * with nothing but logical characters, which are then easy
 		 * to count. :-)
 		 */
-		str = strip_ansi(buffer);
+		str = normalize_string(buffer, 3);
 
 		/*
 		 * Count out the characters.
@@ -582,40 +665,18 @@ void	make_status (Window *window)
 				prc = &pr_rhs;
 			}
 
-			/*
-			 * Skip over color attributes if we're not
-			 * doing color.
-			 */
-			else if (*ptr == '\003')
-			{
-				const u_char *end = skip_ctl_c_seq(ptr, NULL, NULL);
-				while (ptr < end)
-					*cp++ = *ptr++;
-			}
-
-			/*
-			 * If we have a ROM character code here, we need to
-			 * copy it to the output buffer, as well as the fill
-			 * character, and increment the printable counter by
-			 * only 1.
-			 */
-			else if (*ptr == ROM_CHAR)
-			{
-				fillchar[0] = *cp++ = *ptr++;
-				fillchar[1] = *cp++ = *ptr++;
-				fillchar[2] = *cp++ = *ptr++;
-				fillchar[3] = *cp++ = *ptr++;
-				fillchar[4] = 0;
-				*prc += 1;
-			}
-
-			/*
-			 * Is it NOT a printable character?
-			 */
-			else if ((*ptr == REV_TOG) || (*ptr == UND_TOG) ||
-				 (*ptr == ALL_OFF) || (*ptr == BOLD_TOG) ||
-				 (*ptr == BLINK_TOG))
-					*cp++ = *ptr++;
+                        /*
+                         * Skip over attribute changes, not useful.
+                         */
+                        else if (*ptr == '\006')
+                        {
+                                /* Copy the next 5 values */
+                                *cp++ = *ptr++;
+                                *cp++ = *ptr++;
+                                *cp++ = *ptr++;
+                                *cp++ = *ptr++;
+                                *cp++ = *ptr++;
+                        }
 
 			/*
 			 * XXXXX This is a bletcherous hack.
@@ -670,8 +731,8 @@ void	make_status (Window *window)
 
 			numf = window->screen->co - pr_lhs - pr_rhs -1;
 			while (numf-- >= 0)
-				strmcat(lhs_buffer, lhs_fillchar, 
-						BIG_BUFFER_SIZE);
+				strlcat(lhs_buffer, lhs_fillchar, 
+						sizeof lhs_buffer);
 		}
 
 		/*
@@ -682,13 +743,14 @@ void	make_status (Window *window)
 			int chars = window->screen->co - pr_lhs - 1;
 
 			while (chars-- >= 0)
-				strmcat(lhs_buffer, lhs_fillchar, 
-						BIG_BUFFER_SIZE);
+				strlcat(lhs_buffer, lhs_fillchar, 
+						sizeof lhs_buffer);
 		}
 
-		strlcpy(buffer, lhs_buffer, BIG_BUFFER_SIZE);
-		strlcat(buffer, rhs_buffer, BIG_BUFFER_SIZE);
-		strlcat(buffer, ALL_OFF_STR, BIG_BUFFER_SIZE);
+		save_size = strlen(all_off());
+		strlcpy(buffer, lhs_buffer, sizeof buffer - save_size);
+		strlcat(buffer, rhs_buffer, sizeof buffer - save_size);
+		strlcat(buffer, all_off(), sizeof buffer);
 		new_free(&str);
 
 		/*
@@ -701,14 +763,15 @@ void	make_status (Window *window)
 			status_line, 
 			buffer);
 
-		if (dumb_mode)
+		if (dumb_mode || !foreground)
 			continue;
 
 		/*
 		 * Update the status line on the screen.
 		 * First check to see if it has changed
+		 * Remember this is only done in full screen mode.
 		 */
-		if (!window->status.line[status_line].result ||
+		if (must_redraw || !window->status.line[status_line].result ||
 			strcmp(buffer, window->status.line[status_line].result))
 		{
 			/*
@@ -720,18 +783,15 @@ void	make_status (Window *window)
 			/*
 			 * Output the status line to the screen
 			 */
-			if (!dumb_mode)
-			{
-				output_screen = window->screen;
-				term_move_cursor(0, window->bottom + status_line);
-				output_line(buffer);
-				cursor_in_display(window);
-				term_all_off();
-			}
+			output_screen = window->screen;
+			term_move_cursor(0, window->bottom + status_line);
+			output_with_count(buffer, 1, 1);
+			cursor_in_display(window);
 		}
 	}
 
 	cursor_to_input();
+	return 0;
 }
 
 
@@ -800,7 +860,7 @@ static	char	my_buffer[64];
 	/*
 	 * If this window isnt connected to a server, say so.
 	 */
-	if (window->server == -1)
+	if (window->server == NOSERV)
 		return "No Server";
 
 	/*
@@ -924,18 +984,18 @@ static	char	my_buffer[81];
 		if (window->miscflags & WINDOW_NOTIFIED)
 		{
 			if (doneone++)
-				strmcat(buf2, ",", 80);
-			strmcat(buf2, (map == 1 && window->name) ? 
+				strlcat(buf2, ",", sizeof buf2);
+			strlcat(buf2, (map == 1 && window->name) ? 
 					window->name :
-					ltoa(window->refnum), 80);
+					ltoa(window->refnum), sizeof buf2);
 		}
 	}
 
 	/*
-	 * Only do the sprintf if there are windows to process.
+	 * Only do the snprintf if there are windows to process.
 	 */
 	if (doneone && notify_format)
-		snprintf(my_buffer, 80, notify_format, buf2);
+		snprintf(my_buffer, sizeof my_buffer, notify_format, buf2);
 
 	return my_buffer;
 }
@@ -948,7 +1008,7 @@ STATUS_FUNCTION(status_clock)
 	static	char	my_buffer[81];
 
 	if (get_int_var(CLOCK_VAR) && clock_format && DISPLAY_ON_WINDOW)
-		snprintf(my_buffer, 80, clock_format, update_clock(GET_TIME));
+		snprintf(my_buffer, 80, clock_format, get_clock());
 	else
 		*my_buffer = 0;
 
@@ -960,30 +1020,43 @@ STATUS_FUNCTION(status_clock)
  */
 STATUS_FUNCTION(status_mode)
 {
-	char	*mode = (char *) 0;
-static  char    my_buffer[81];
+	const char *	mode = NULL;
+	const char *	chan = NULL;
+static  char    	my_buffer[81];
 
+	/* If anything goes wrong, we return an empty string */
 	*my_buffer = 0;
-	if (window->current_channel && window->server != -1)
-	{
-                mode = get_channel_mode(window->current_channel,window->server);
-		if (mode && *mode && mode_format)
-		{
-			/*
-			 * This gross hack is required to make sure that the 
-			 * channel key doesnt accidentally contain anything 
-			 * dangerous...
-			 */
-			if (get_int_var(STATUS_DOES_EXPANDOS_VAR))
-			{
-				char *mode2 = alloca(strlen(mode) * 2 + 1);
-				double_quote(mode, "$", mode2);
-				mode = mode2;
-			}
 
-			snprintf(my_buffer, 80, mode_format, mode);
-		}
+	/* If the user has no mode format, or we're not connected, punt. */
+        if (!mode_format || window->server == NOSERV)
+		return my_buffer;
+
+	/* If there is a current channel, get it's mode */
+	if ((chan = get_echannel_by_refnum(window->refnum)))
+		mode = get_channel_mode(chan, window->server);
+
+	/* If this is %+, and there is not a mode, punt. */
+	if (mode == 0 && (!mode || !*mode))
+		return my_buffer;
+
+	/* If this is %{1}+, and there is not a mode, use empty string */
+	if (!mode)
+		mode = empty_string;
+
+	/*
+	 * This gross hack is required to make sure that the 
+	 * channel key doesnt accidentally contain anything 
+	 * dangerous...
+	 */
+	if (get_int_var(STATUS_DOES_EXPANDOS_VAR))
+	{
+		char *mode2 = alloca(strlen(mode) * 2 + 1);
+		double_quote(mode, "$", mode2);
+		mode = mode2;
 	}
+
+	/* Press the mode into the status format. */
+	snprintf(my_buffer, 80, mode_format, mode);
 	return my_buffer;
 }
 
@@ -1009,7 +1082,7 @@ static	char	my_buffer[81];
 	if (window->server < 0)
 		return empty_string;
 
-	strmcpy(localbuf, get_umode(window->server), 19);
+	strlcpy(localbuf, get_umode(window->server), sizeof localbuf);
 	if (!*localbuf)
 		return empty_string;
 
@@ -1024,12 +1097,35 @@ static	char	my_buffer[81];
 STATUS_FUNCTION(status_chanop)
 {
 	char	*text;
+	const char *chan;
 
-	if (window->current_channel && window->server != -1 &&
-	    get_channel_oper(window->current_channel, window->server) &&
-	    (text = get_string_var(STATUS_CHANOP_VAR)))
-		return text;
+	if (window->server == NOSERV ||
+           (!(chan = get_echannel_by_refnum(window->refnum))))
+		return empty_string;
+	
+	if (get_channel_oper(chan, window->server) &&
+		(text = get_string_var(STATUS_CHANOP_VAR)))
+			return text;
 
+	if (get_channel_halfop(chan, window->server) &&
+		(text = get_string_var(STATUS_HALFOP_VAR)))
+			return text;
+
+	return empty_string;
+}
+
+/*
+ * are we using SSL conenction?
+ */
+STATUS_FUNCTION(status_ssl)
+{
+	char *text;
+
+	if (window->server != NOSERV && get_server_isssl(window->server) &&
+		(text = get_string_var(STATUS_SSL_ON_VAR)))
+			return text;
+	else if ((text = get_string_var(STATUS_SSL_OFF_VAR)))
+			return text;
 	return empty_string;
 }
 
@@ -1043,11 +1139,20 @@ STATUS_FUNCTION(status_hold_lines)
 	int	num;
 static	char	my_buffer[81];
 	int	interval = window->hold_interval;
+	int	lines_held;
 
 	if (interval == 0)
 		interval = 1;		/* XXX WHAT-ever */
 
-	if ((num = (window->lines_held / interval) * interval))
+	if (window->holding_distance_from_display_ip > window->scrollback_distance_from_display_ip)
+		lines_held = window->holding_distance_from_display_ip - window->display_size;
+	else
+		lines_held = window->scrollback_distance_from_display_ip - window->display_size;
+
+	if (lines_held <= 0)
+		return empty_string;
+
+	if ((num = (lines_held / interval) * interval))
 	{
 		snprintf(my_buffer, 80, hold_lines_format, ltoa(num));
 		return my_buffer;
@@ -1061,21 +1166,25 @@ static	char	my_buffer[81];
  */
 STATUS_FUNCTION(status_channel)
 {
+	const char *chan;
 	char 	channel[IRCD_BUFFER_SIZE + 1];
 static	char	my_buffer[IRCD_BUFFER_SIZE + 1];
 	int	num;
 
-	if (!window->current_channel || window->server == -1)
+	if (window->server == NOSERV || !channel_format)
+		return empty_string;
+
+	if (!(chan = get_echannel_by_refnum(window->refnum)))
 		return empty_string;
 
 	if (get_int_var(HIDE_PRIVATE_CHANNELS_VAR) && 
-	    is_channel_private(window->current_channel, window->server))
-		strmcpy(channel, "*private*", IRCD_BUFFER_SIZE);
+	    is_channel_private(chan, window->server))
+		strlcpy(channel, "*private*", sizeof channel);
 	else
-		strmcpy(channel, window->current_channel, IRCD_BUFFER_SIZE);
+		strlcpy(channel, chan, sizeof channel);
 
 	num = get_int_var(CHANNEL_NAME_WIDTH_VAR);
-	if (num > 0 && strlen(channel) > num)
+	if (num > 0 && (int)strlen(channel) > num)
 		channel[num] = 0;
 
 	snprintf(my_buffer, IRCD_BUFFER_SIZE, 
@@ -1091,10 +1200,14 @@ static	char	my_buffer[IRCD_BUFFER_SIZE + 1];
 STATUS_FUNCTION(status_voice)
 {
 	char *text;
+	const char *chan;
 
-	if (window->current_channel && window->server != -1 &&
-	    get_channel_voice(window->current_channel, window->server) &&
-	    !get_channel_oper(window->current_channel, window->server) &&
+	if (window->server == NOSERV ||
+           (chan = get_echannel_by_refnum(window->refnum)) == NULL)
+		return empty_string;
+
+	if (get_channel_voice(chan, window->server) &&
+	    !get_channel_oper(chan, window->server) &&
 	    (text = get_string_var(STATUS_VOICE_VAR)))
 		return text;
 
@@ -1106,7 +1219,7 @@ STATUS_FUNCTION(status_voice)
  */
 STATUS_FUNCTION(status_mail)
 {
-	char	*number;
+	const char *	number;
 static	char	my_buffer[81];
 
 	/*
@@ -1168,7 +1281,7 @@ STATUS_FUNCTION(status_away)
 	if (connected_to_server == 1 && !DISPLAY_ON_WINDOW)
 		return empty_string;
 
-	if (window->server != -1 && 
+	if (window->server != NOSERV && 
 	    get_server_away(window->server) && 
 	    (text = get_string_var(STATUS_AWAY_VAR)))
 		return text;
@@ -1178,7 +1291,7 @@ STATUS_FUNCTION(status_away)
 
 
 /*
- * This is a generic status_userX variable.  
+ * This is a generic status_userX variable.
  */
 STATUS_FUNCTION(status_user)
 {
@@ -1236,18 +1349,18 @@ STATUS_FUNCTION(status_hold)
 {
 	char *text;
 
-	if (window->holding_something && 
-	    (text = get_string_var(STATUS_HOLD_VAR)))
-		return text;
-	else
-		return empty_string;
+	if (window->holding_distance_from_display_ip > window->display_size  ||
+	    window->scrollback_distance_from_display_ip > window->display_size)
+		if ((text = get_string_var(STATUS_HOLD_VAR)))
+			return text;
+	return empty_string;
 }
 
 STATUS_FUNCTION(status_oper)
 {
 	char *text;
 
-	if (window->server != -1 && get_server_operator(window->server) &&
+	if (window->server != NOSERV && get_server_operator(window->server) &&
 	    (text = get_string_var(STATUS_OPER_VAR)) &&
 	    (connected_to_server != 1 || DISPLAY_ON_WINDOW))
 		return text;
@@ -1271,14 +1384,22 @@ STATUS_FUNCTION(status_refnum)
 	static char my_buffer[81];
 
 	strlcpy(my_buffer, window->name ? window->name 
-					: ltoa(window->refnum), 80);
+					: ltoa(window->refnum), sizeof my_buffer);
+	return my_buffer;
+}
+
+STATUS_FUNCTION(status_refnum_real)
+{
+	static char my_buffer[81];
+
+	strlcpy(my_buffer, ltoa(window->refnum), sizeof my_buffer);
 	return my_buffer;
 }
 
 STATUS_FUNCTION(status_version)
 {
 	if (DISPLAY_ON_WINDOW)
-		return (char *)irc_version; /* XXXX */
+		return irc_version; /* XXXX */
 
 	return empty_string;
 }
@@ -1303,6 +1424,14 @@ STATUS_FUNCTION(status_dcc)
 	return empty_string;
 }
 
+/*
+ * As above, but always return the indicator.
+ */
+STATUS_FUNCTION(status_dcc_all)
+{
+	return DCC_get_current_transfer();
+}
+
 /*
  * This displays something if the client is in 'cpu saver' mode.
  * A discussion of that mode should be found in irc.c, so i wont get
@@ -1330,12 +1459,14 @@ STATUS_FUNCTION(status_position)
 {
 	static char my_buffer[81];
 
-	snprintf(my_buffer, 80, "(%d-%d)", 
+	snprintf(my_buffer, 80, "(%d/%d/%d-%d-%d)", 
+			window->scrolling_distance_from_display_ip,
+			window->holding_distance_from_display_ip,
+			window->scrollback_distance_from_display_ip,
+			window->display_size, window->cursor);
+#if 0
 			window->screen->input_line,
 			window->screen->input_cursor);
-#if 0
-			window->lines_scrolled_back,
-			window->distance_from_display);
 #endif
 	return my_buffer;
 }
@@ -1348,7 +1479,7 @@ STATUS_FUNCTION(status_scrollback)
 {
 	char *stuff;
 
-	if (window->scrollback_point && 
+	if (window->scrollback_top_of_display &&
 	    (stuff = get_string_var(STATUS_SCROLLBACK_VAR)))
 		return stuff;
 	else
@@ -1359,10 +1490,10 @@ STATUS_FUNCTION(status_scroll_info)
 {
 	static char my_buffer[81];
 
-	if (window->scrollback_point)
+	if (window->scrollback_top_of_display)
 	{
 		snprintf(my_buffer, 80, " (Scroll: %d of %d)", 
-				window->distance_from_display,
+				window->scrollback_distance_from_display_ip,
 				window->display_buffer_size - 1);
 	}
 	else
@@ -1377,7 +1508,7 @@ STATUS_FUNCTION(status_windowspec)
 	static char my_buffer[81];
 
 	if (window->status.special)
-		strmcpy(my_buffer, window->status.special, 80);
+		strlcpy(my_buffer, window->status.special, sizeof my_buffer);
 	else
 		*my_buffer = 0;
 
@@ -1396,3 +1527,19 @@ STATUS_FUNCTION(status_test)
 	return retval;
 }
 
+/*
+ * This returns something if this window is currently in scrollback mode.
+ * Useful if you sometimes forget!
+ */
+STATUS_FUNCTION(status_swappable)
+{
+	char *stuff;
+
+	if (!window->swappable && 
+	    (stuff = get_string_var(STATUS_NOSWAP_VAR)))
+		return stuff;
+	else
+		return empty_string;
+}
+
+
