@@ -1,8 +1,34 @@
+/* $EPIC: who.c,v 1.28 2004/01/29 01:32:06 jnelson Exp $ */
 /*
  * who.c -- The WHO queue.  The ISON queue.  The USERHOST queue.
  *
- * Written by Jeremy Nelson
- * Copyright 1996, 1997 EPIC Software Labs
+ * Copyright  1996, 2003 EPIC Software Labs.
+ * All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notices, the above paragraph (the one permitting redistribution),
+ *    this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ * 3. The names of the author(s) may not be used to endorse or promote
+ *    products derived from this software without specific prior written
+ *    permission.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
+ * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
+ * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
+ * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
+ * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
+ * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+ * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
+ * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
+ * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
+ * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ * SUCH DAMAGE.
  */
 
 #define NEED_SERVER_LIST
@@ -10,6 +36,7 @@
 #include "commands.h"
 #include "ircaux.h"
 #include "who.h"
+#include "ssl.h"
 #include "server.h"
 #include "window.h"
 #include "vars.h"
@@ -19,7 +46,7 @@
 #include "parse.h"
 #include "if.h"
 #include "names.h"
-
+#include "words.h"
 
 /*
  *
@@ -54,11 +81,17 @@
  * flag has *no effect* in any other context than this.
  */
 #define WHO_INVISIBLE	0x2000
+#define WHO_OPERSPY	0x4000
 
 
-static WhoEntry *who_queue_top (void)
+static WhoEntry *who_queue_top (int refnum)
 {
-	return server_list[from_server].who_queue;
+	Server *s;
+
+	if (!(s = get_server(refnum)))
+		return NULL;
+
+	return s->who_queue;
 }
 
 /*
@@ -67,25 +100,35 @@ static WhoEntry *who_queue_top (void)
  * asking, and they want to know who is LAST (before them)
  * So it sucks.  Sue me.
  */
-static WhoEntry *who_previous_query (WhoEntry *me)
+static WhoEntry *who_previous_query (int refnum, WhoEntry *me)
 {
-	WhoEntry *what = who_queue_top();
+	WhoEntry *what;
+	Server *s;
 
+	if (!(s = get_server(refnum)))
+		return NULL;
+
+	what = who_queue_top(refnum);
 	while (what && what->next != me)
 		what = what->next;
 
 	return what;
 }
 
-static void who_queue_add (WhoEntry *item)
+static void who_queue_add (int refnum, WhoEntry *item)
 {
-	WhoEntry *bottom = who_queue_top();
+	WhoEntry *bottom;
+	Server *s;
 
+	if (!(s = get_server(refnum)))
+		return;
+
+	bottom = who_queue_top(refnum);
 	while (bottom && bottom->next)
 		bottom = bottom->next;
 
 	if (!bottom)
-		server_list[from_server].who_queue = item;
+		s->who_queue = item;
 	else
 		bottom->next = item;
 
@@ -105,18 +148,22 @@ static void delete_who_item (WhoEntry *save)
 	new_free((char **)&save);
 }
 
-static void who_queue_pop (void)
+static void who_queue_pop (int refnum)
 {
 	WhoEntry *save;
 	int	piggyback;
+	Server *s;
+
+	if (!(s = get_server(refnum)))
+		return;
 
 	do
 	{
-		if (!(save = server_list[from_server].who_queue))
+		if (!(save = s->who_queue))
 			break;
 
 		piggyback = save->piggyback;
-		server_list[from_server].who_queue = save->next;
+		s->who_queue = save->next;
 		delete_who_item(save);
 	}
 	while (piggyback);
@@ -147,13 +194,16 @@ static WhoEntry *get_new_who_entry (void)
 	return new_w;
 }
 
-static void who_queue_list (void)
+static void who_queue_list (int refnum)
 {
-	WhoEntry *item = server_list[from_server].who_queue;
+	WhoEntry *item;
 	int count = 0;
+	Server *s;
 
-	item = server_list[from_server].who_queue;
-	while (item)
+	if (!(s = get_server(refnum)))
+		return;
+
+	for (item = s->who_queue; item; item = item->next)
 	{
 		yell("[%d] [%d] [%s] [%s] [%s]", count,
 			item->who_mask,
@@ -161,18 +211,20 @@ static void who_queue_list (void)
 			item->who_stuff ? item->who_stuff : empty_string,
 			item->who_end ? item->who_end : empty_string);
 		count++;
-		item = item->next;
 	}
 }
 
-static void who_queue_flush (void)
+static void who_queue_flush (int refnum)
 {
-	WhoEntry *item;
+	Server *s;
 
-	while ((item = server_list[from_server].who_queue))
-		who_queue_pop();
+	if (!(s = get_server(refnum)))
+		return;
 
-	yell("Who queue for server [%d] purged", from_server);
+	while (s->who_queue)
+		who_queue_pop(refnum);
+
+	yell("Who queue for server [%d] purged", refnum);
 }
 
 
@@ -183,23 +235,23 @@ static void who_queue_flush (void)
  */
 BUILT_IN_COMMAND(whocmd)
 {
-	whobase(args, NULL, NULL);
+	whobase(from_server, args, NULL, NULL);
 }
 
 
 /*
  * whobase: What does all the work.
  */
-void 	whobase(char *args, void (*line) (char *, char **), void (*end) (char *, char **))
+void 	whobase (int refnum, char *args, void (*line) (int, const char *, const char *, const char **), void (*end) (int, const char *, const char *, const char **))
 {
-	char	*arg,
-		*channel = NULL;
-	int	no_args = 1,
-		len;
-	WhoEntry *new_w, *old;
+	char *		arg;
+	const char *	channel = NULL;
+	int		no_args = 1,
+			len;
+	WhoEntry *	new_w, *old;
 
 	/* Maybe should output a warning? */
-	if (!is_server_connected(from_server))
+	if (!is_server_registered(refnum))
 		return;
 
 	new_w = get_new_who_entry();
@@ -220,35 +272,41 @@ void 	whobase(char *args, void (*line) (char *, char **), void (*end) (char *, c
 			return;
 		}
 
-		if (!strncmp(arg, "line", 4))		/* LINE */
+		else if (!strncmp(arg, "away", MAX(len, 1)))
+			new_w->who_mask |= WHO_AWAY;
+		else if (!strncmp(arg, "chops", MAX(len, 2)))
+			new_w->who_mask |= WHO_CHOPS;
+		else if (!strncmp(arg, "delete", MAX(len, 1)))
 		{
-			char *line;
-
-			if ((line = next_expr(&args, '{')))
-				malloc_strcpy(&new_w->who_stuff, line);
-			else
-				say("Need {...} argument for -LINE argument.");
+			who_queue_list(refnum);
+			delete_who_item(new_w);
+			return;
 		}
-		else if (!strncmp(arg, "end", 3))	/* END */
+		else if (!strncmp(arg, "dx", MAX(len, 2)))
 		{
-			char *line;
+			new_w->dalnet_extended = 1;
+			new_w->dalnet_extended_args = new_next_arg(args, &args);
+			channel = args;		/* Grab the rest of args */
+			args = NULL;
+		}
+		else if (!strncmp(arg, "end", MAX(len, 3)))
+		{
+			char *stuff;
 
-			if ((line = next_expr(&args, '{')))
-				malloc_strcpy(&new_w->who_end, line);
+			if ((stuff = next_expr(&args, '{')))
+				malloc_strcpy(&new_w->who_end, stuff);
 			else
 				say("Need {...} argument for -END argument.");
 		}
-		else if (!strncmp(arg, "o", 1))		/* OPS */
-			new_w->who_mask |= WHO_OPS;
-		else if (!strncmp(arg, "lu", 2))	/* LUSERS */
-			new_w->who_mask |= WHO_LUSERS;
-		else if (!strncmp(arg, "ch", 2))	/* CHOPS */
-			new_w->who_mask |= WHO_CHOPS;
-		else if (!strncmp(arg, "no", 2))	/* NOCHOPS */
-			new_w->who_mask |= WHO_NOCHOPS;
-		else if (!strncmp(arg, "u-i", 3))	/* INVIS */
-			new_w->who_mask |= WHO_INVISIBLE;
-		else if (!strncmp(arg, "ho", 2))	/* HOSTS */
+		else if (!strncmp(arg, "flush", MAX(len, 1)))
+		{
+			who_queue_flush(refnum);
+			delete_who_item(new_w);
+			return;
+		}
+	 	else if (!strncmp(arg, "here", MAX(len, 2)))
+			new_w->who_mask |= WHO_HERE;
+		else if (!strncmp(arg, "hosts", MAX(len, 2)))
 		{
 			if ((arg = next_arg(args, &args)) == NULL)
 			{
@@ -260,23 +318,32 @@ void 	whobase(char *args, void (*line) (char *, char **), void (*end) (char *, c
 			malloc_strcpy(&new_w->who_host, arg);
 			channel = new_w->who_host;
 		}
-	 	else if (!strncmp(arg, "he", 2))	/* here */
-			new_w->who_mask |= WHO_HERE;
-		else if (!strncmp(arg, "a", 1))		/* away */
-			new_w->who_mask |= WHO_AWAY;
-		else if (!strncmp(arg, "s", 1)) 	/* servers */
-		{
-			if ((arg = next_arg(args, &args)) == NULL)
-			{
-				say("WHO -SERVER: missing arguement");
-				return;
-			}
 
-			new_w->who_mask |= WHO_SERVER;
-			malloc_strcpy(&new_w->who_server, arg);
-			channel = new_w->who_server;
+		else if (!strncmp(arg, "line", MAX(len, 4)))
+		{
+			char *stuff;
+
+			if ((stuff = next_expr(&args, '{')))
+				malloc_strcpy(&new_w->who_stuff, stuff);
+			else
+				say("Need {...} argument for -LINE argument.");
 		}
-		else if (!strncmp(arg, "na", 2))
+		else if (!strncmp(arg, "literal", MAX(len, 3)))
+		{
+			/* Hope for the best */
+			new_w->who_mask = 0;	/* For safety reasons */
+			new_w->dalnet_extended = 0;
+			new_w->undernet_extended = 0;
+			new_free(&new_w->who_stuff);
+			new_free(&new_w->who_end);
+			who_queue_add(refnum, new_w);
+
+			send_to_aserver(refnum, "WHO %s", args);
+			return;
+		}
+		else if (!strncmp(arg, "lusers", MAX(len, 2)))
+			new_w->who_mask |= WHO_LUSERS;
+		else if (!strncmp(arg, "name", MAX(len, 2)))
 		{
 			if ((arg = next_arg(args, &args)) == NULL)
 			{
@@ -288,19 +355,7 @@ void 	whobase(char *args, void (*line) (char *, char **), void (*end) (char *, c
 			malloc_strcpy(&new_w->who_name, arg);
 			channel = new_w->who_name;
 		}
-		else if (!strncmp(arg, "r", 1))
-		{
-			if ((arg = next_arg(args, &args)) == NULL)
-			{
-				say("WHO -REALNAME: missing arguement");
-				return;
-			}
-
-			new_w->who_mask |= WHO_REAL;
-			malloc_strcpy(&new_w->who_real, arg);
-			channel = new_w->who_real;
-		}
-		else if (!strncmp(arg, "ni", 2))
+		else if (!strncmp(arg, "nick", MAX(len, 2)))
 		{
 			if ((arg = next_arg(args, &args)) == NULL)
 			{
@@ -312,43 +367,44 @@ void 	whobase(char *args, void (*line) (char *, char **), void (*end) (char *, c
 			malloc_strcpy(&new_w->who_nick, arg);
 			channel = new_w->who_nick;
 		}
-		else if (!strncmp(arg, "f", 1))
+		else if (!strncmp(arg, "nochops", MAX(len, 2)))
+			new_w->who_mask |= WHO_NOCHOPS;
+		else if (!strncmp(arg, "oper", MAX(len, 1)))
+			new_w->who_mask |= WHO_OPS;
+		else if (!strncmp(arg, "operspy", MAX(len, 5)))
+			new_w->who_mask |= WHO_OPERSPY;
+		else if (!strncmp(arg, "realname", MAX(len, 1)))
 		{
-			who_queue_flush();
-			delete_who_item(new_w);
-			return;
+			if ((arg = next_arg(args, &args)) == NULL)
+			{
+				say("WHO -REALNAME: missing arguement");
+				return;
+			}
+
+			new_w->who_mask |= WHO_REAL;
+			malloc_strcpy(&new_w->who_real, arg);
+			channel = new_w->who_real;
 		}
-		else if (!strncmp(arg, "ux", 2))
+		else if (!strncmp(arg, "servers", MAX(len, 1)))
+		{
+			if ((arg = next_arg(args, &args)) == NULL)
+			{
+				say("WHO -SERVER: missing arguement");
+				return;
+			}
+
+			new_w->who_mask |= WHO_SERVER;
+			malloc_strcpy(&new_w->who_server, arg);
+			channel = new_w->who_server;
+		}
+		else if (!strncmp(arg, "u-i", MAX(len, 3)))
+			new_w->who_mask |= WHO_INVISIBLE;
+		else if (!strncmp(arg, "ux", MAX(len, 2)))
 		{
 			new_w->undernet_extended = 1;
 			new_w->undernet_extended_args = args;
 			args = NULL;
 		}
-		else if (!strncmp(arg, "dx", 2))
-		{
-			new_w->dalnet_extended = 1;
-			new_w->dalnet_extended_args = new_next_arg(args, &args);
-			channel = args;		/* Grab the rest of args */
-			args = NULL;
-		}
-		else if (!strncmp(arg, "d", 1))
-		{
-			who_queue_list();
-			delete_who_item(new_w);
-			return;
-		}
-		else if (!strncmp(arg, "lit", 3))	/* Hope for the best */
-		{
-			new_w->who_mask = 0;	/* For safety reasons */
-			new_w->dalnet_extended = 0;
-			new_w->undernet_extended = 0;
-			new_free(&new_w->who_stuff);
-			new_free(&new_w->who_end);
-			who_queue_add(new_w);
-
-			send_to_server("WHO %s", args);
-			return;
-		}
 		else
 		{
 			say("Unknown or missing flag");
@@ -358,7 +414,7 @@ void 	whobase(char *args, void (*line) (char *, char **), void (*end) (char *, c
 	    }
 	    else if (strcmp(arg, "*") == 0)
 	    {
-		channel = get_channel_by_refnum(0);
+		channel = get_echannel_by_refnum(0);
 		if (!channel || !*channel)
 		{
 			say("You are not on a channel.  "
@@ -380,14 +436,14 @@ void 	whobase(char *args, void (*line) (char *, char **), void (*end) (char *, c
 
 	if (!channel && (new_w->who_mask & WHO_OPS))
 		channel = "*.*";
-	new_w->who_target = m_strdup(channel);
+	new_w->who_target = malloc_strdup(channel);
 
-	who_queue_add(new_w);
+	who_queue_add(refnum, new_w);
 
 	/*
 	 * Check to see if we can piggyback
 	 */
-	old = who_previous_query(new_w);
+	old = who_previous_query(refnum, new_w);
 	if (old && !old->dirty && old->who_target && channel && 
 		!strcmp(old->who_target, channel))
 	{
@@ -397,7 +453,8 @@ void 	whobase(char *args, void (*line) (char *, char **), void (*end) (char *, c
 	}
 	else if (new_w->undernet_extended)
 	{
-		send_to_server("WHO %s %s%s%s", new_w->who_target,
+		send_to_aserver(refnum, "WHO %s %s%s%s", 
+			new_w->who_target,
 			(new_w->who_mask & WHO_OPS) ?  "o" : "",
 			(new_w->who_mask & WHO_INVISIBLE) ? "x" : "",
 			new_w->undernet_extended_args ? 
@@ -405,11 +462,18 @@ void 	whobase(char *args, void (*line) (char *, char **), void (*end) (char *, c
 	}
 	else if (new_w->dalnet_extended)
 	{
-		send_to_server("WHO %s %s", new_w->dalnet_extended_args,
+		send_to_aserver(refnum, "WHO %s %s", 
+			new_w->dalnet_extended_args,
 			new_w->who_target);
 	}
+	else if (new_w->who_mask & WHO_OPERSPY)
+		send_to_aserver(refnum, "OPERSPY WHO %s %s%s", 
+			new_w->who_target,
+			(new_w->who_mask & WHO_OPS) ?  "o" : "",
+			(new_w->who_mask & WHO_INVISIBLE) ? "x" : "");
 	else
-		send_to_server("WHO %s %s%s", new_w->who_target,
+		send_to_aserver(refnum, "WHO %s %s%s", 
+			new_w->who_target,
 			(new_w->who_mask & WHO_OPS) ?  "o" : "",
 			(new_w->who_mask & WHO_INVISIBLE) ? "x" : "");
 
@@ -417,36 +481,37 @@ void 	whobase(char *args, void (*line) (char *, char **), void (*end) (char *, c
 
 static int who_whine = 0;
 
-void	whoreply (char *from, char **ArgList)
+void	whoreply (int refnum, const char *from, const char *comm, const char **ArgList)
 {
 static	char	format[40];
 static	int	last_width = -1;
 	int	ok = 1;
-	char	*channel,
-		*user,
-		*host,
-		*server,
-		*nick,
-		*stat,
-		*name;
-	WhoEntry *new_w = who_queue_top();
-
-	if (!ArgList[5])
-		return;		/* Fake! */
+	const char	*channel,
+			*user,
+			*host,
+			*server,
+			*nick,
+			*status,
+			*ircname;
+	char 	*name;
+	WhoEntry *new_w = who_queue_top(refnum);
 
 	if (!new_w)
 	{
                 new_w = get_new_who_entry();
                 new_w->line = NULL;
                 new_w->end = NULL;
-                new_w->who_target = m_strdup(star);
-                who_queue_add(new_w);
+                new_w->who_target = malloc_strdup(star);
+                who_queue_add(refnum, new_w);
 	}
 
 	if (new_w->undernet_extended)
 		yell("### You asked for an extended undernet request but "
 			"didn't get one back. ###");
 
+	/* Who replies always go to the current window. */
+	message_from(new_w->who_target, LOG_CRAP);
+
 do
 {
 	/*
@@ -462,62 +527,72 @@ do
 	 */
 	if (new_w->line)
 	{
-		new_w->line(from, ArgList);
+		new_w->line(refnum, from, comm, ArgList);
 		continue;
 	}
 
 	if (last_width != get_int_var(CHANNEL_NAME_WIDTH_VAR))
 	{
 		if ((last_width = get_int_var(CHANNEL_NAME_WIDTH_VAR)) != 0)
-		    sprintf(format, "%%-%u.%us %%-9s %%-3s %%s@%%s (%%s)",
+		    snprintf(format, sizeof format, 
+				"%%-%u.%us %%-9s %%-3s %%s@%%s (%%s)",
 					(unsigned) last_width,
 					(unsigned) last_width);
 		else
-		    strcpy(format, "%s\t%-9s %-3s %s@%s (%s)");
+		    strlcpy(format, "%s\t%-9s %-3s %s@%s (%s)", sizeof format);
 	}
-	channel = ArgList[0];
-	user    = ArgList[1];
-	host    = ArgList[2];
-	server  = ArgList[3];
-	nick    = ArgList[4];
-	stat    = ArgList[5];
-	PasteArgs(ArgList, 6);
-	name    = ArgList[6];
 
-	if (*stat == 'S')	/* this only true for the header WHOREPLY */
+	if (!(channel = ArgList[0]))
+		{ rfc1459_odd(from, comm, ArgList); break; }
+	if (!(user    = ArgList[1]))
+		{ rfc1459_odd(from, comm, ArgList); break; }
+	if (!(host    = ArgList[2]))
+		{ rfc1459_odd(from, comm, ArgList); break; }
+	if (!(server  = ArgList[3]))
+		{ rfc1459_odd(from, comm, ArgList); break; }
+	if (!(nick    = ArgList[4]))
+		{ rfc1459_odd(from, comm, ArgList); break; }
+	if (!(status  = ArgList[5]))
+		{ rfc1459_odd(from, comm, ArgList); break; }
+	PasteArgs(ArgList, 6);
+	if (!(ircname  = ArgList[6]))
+		{ rfc1459_odd(from, comm, ArgList); break; }
+	name = LOCAL_COPY(ircname);
+
+	if (*status == 'S')	/* this only true for the header WHOREPLY */
 	{
 		char buffer[1024];
 
 		channel = "Channel";
 		snprintf(buffer, 1024, "%s %s %s %s %s %s %s", channel,
-				nick, stat, user, host, server, name);
+				nick, status, user, host, server, name);
 
 		if (new_w->who_stuff)
 			;			/* munch it */
-
 		else if (do_hook(WHO_LIST, "%s", buffer))
-			put_it(format, channel, nick, stat, user, host, name);
+			put_it(format, channel, nick, status, user, host, name);
+
 		return;
 	}
 
 	if (new_w && new_w->who_mask)
 	{
 		if (new_w->who_mask & WHO_HERE)
-			ok = ok && (*stat == 'H');
+			ok = ok && (*status == 'H');
 		if (new_w->who_mask & WHO_AWAY)
-			ok = ok && (*stat == 'G');
+			ok = ok && (*status == 'G');
 		if (new_w->who_mask & WHO_OPS)
-			ok = ok && (*(stat + 1) == '*');
+			ok = ok && (*(status + 1) == '*');
 		if (new_w->who_mask & WHO_LUSERS)
-			ok = ok && (*(stat + 1) != '*');
+			ok = ok && (*(status + 1) != '*');
 		if (new_w->who_mask & WHO_CHOPS)
-			ok = ok && ((*(stat + 1) == '@') ||
-				    (*(stat + 2) == '@') ||
-				    (*(stat + 3) == '@'));
+			ok = ok && ((*(status + 1) == '@') ||
+				    (*(status + 2) == '@') ||
+				    (*(status + 3) == '@'));
 		if (new_w->who_mask & WHO_NOCHOPS)
-			ok = ok && ((*(stat + 1) != '@') &&
-				    (*(stat + 2) != '@') &&
-				    (*(stat + 3) != '@'));
+			ok = ok && ((*(status + 1) != '@') &&
+				    (*(status + 2) != '@') &&
+				    (*(status + 3) != '@'));
 		if (new_w->who_mask & WHO_NAME)
 			ok = ok && wild_match(new_w->who_name, user);
 		if (new_w->who_mask & WHO_NICK)
@@ -535,7 +610,7 @@ do
 		char buffer[1024];
 
 		snprintf(buffer, 1023, "%s %s %s %s %s %s %s", channel,
-				nick, stat, user, host, server, name);
+				nick, status, user, host, server, name);
 
 		if (new_w->who_stuff)
 			parse_line(NULL, new_w->who_stuff, buffer, 0, 0);
@@ -547,7 +622,7 @@ do
 				if (!get_int_var(SHOW_WHO_HOPCOUNT_VAR))
 					next_arg(name, &name);
 
-				put_it(format, channel, nick, stat, user, host, name);
+				put_it(format, channel, nick, status, user, host, name);
 			}
 		}
 	}
@@ -556,45 +631,51 @@ while (new_w->piggyback && (new_w = new_w->next));
 }
 
 /* Undernet's 354 numeric reply. */
-void	xwhoreply (char *from, char **ArgList)
+void	xwhoreply (int refnum, const char *from, const char *comm, const char **ArgList)
 {
-	WhoEntry *new_w = who_queue_top();
+	WhoEntry *new_w = who_queue_top(refnum);
 
 	if (!new_w)
 	{
 		new_w = get_new_who_entry();
 		new_w->line = NULL;
 		new_w->end = NULL;
-		new_w->who_target = m_strdup(star);
+		new_w->who_target = malloc_strdup(star);
 		new_w->undernet_extended = 1;
-		who_queue_add(new_w);
+		who_queue_add(refnum, new_w);
 	}
 
 	if (!new_w->undernet_extended)
 		yell("### You got an extended undernet request back "
 			"even though you didn't ask for one. ###");
 
+	/* Who replies always go to the current window */
+	message_from(new_w->who_target, LOG_CRAP);
+
 	PasteArgs(ArgList, 0);
 	if (do_hook(current_numeric, "%s", ArgList[0]))
-		put_it("%s %s", numeric_banner(), ArgList[0]);
+		put_it("%s %s", banner(), ArgList[0]);
 }
 
 
-void	who_end (char *from, char **ArgList)
+void	who_end (int refnum, const char *from, const char *comm, const char **ArgList)
 {
-	WhoEntry 	*new_w = who_queue_top();
+	WhoEntry 	*new_w = who_queue_top(refnum);
 	char 		buffer[1025];
 
+	PasteArgs(ArgList, 0);
+
 	if (who_whine)
 		who_whine = 0;
 	if (!new_w)
 		return;	
 
+	message_from(new_w->who_target, LOG_CRAP);
 	do
 	{
 		/* Defer to another function, if neccesary.  */
 		if (new_w->end)
-			new_w->end(from, ArgList);
+			new_w->end(refnum, from, comm, ArgList);
 		else
 		{
 			snprintf(buffer, 1024, "%s %s", from, ArgList[0]);
@@ -603,12 +684,12 @@ void	who_end (char *from, char **ArgList)
 
 			else if (get_int_var(SHOW_END_OF_MSGS_VAR))
 			    if (do_hook(current_numeric, "%s", buffer))
-				put_it("%s %s", numeric_banner(), buffer);
+				put_it("%s %s", banner(), buffer);
 		}
 	} 
 	while (new_w->piggyback && (new_w = new_w->next));
 
-	who_queue_pop();
+	who_queue_pop(refnum);
 }
 
 /*
@@ -624,9 +705,9 @@ void	who_end (char *from, char **ArgList)
  * to correctly match up error codes to requests.  Gee whiz, you wouldn't
  * think it would be that hard to get this right.
  */
-int	fake_who_end (char *from, char *who_target)
+int	fake_who_end (int refnum, const char *from, const char *comm, const char *who_target)
 {
-	WhoEntry 	*new_w = who_queue_top();
+	WhoEntry 	*new_w = who_queue_top(refnum);
 
 	if (who_whine)
 		who_whine = 0;
@@ -639,29 +720,35 @@ int	fake_who_end (char *from, char *who_target)
 
 	if (who_target != NULL)
 	{
-		while (last_char(who_target) == ' ')
-			chop(who_target, 1);
+		char *target;
+
+		target = LOCAL_COPY(who_target);
+		while (last_char(target) == ' ')
+			chop(target, 1);
 
 		/*
 		 * So 'who_target' isn't NULL here.  Make sure it's a 
 		 * legitimate match to our current top of who request.
 		 */
-		if (strncmp(new_w->who_target, who_target, strlen(who_target)))
+		if (strncmp(new_w->who_target, target, strlen(target)))
 			return 0;
+
+		who_target = target;
 	}
 
+	message_from(new_w->who_target, LOG_CRAP);
 	do
 	{
 		/* Defer to another function, if neccesary.  */
 		if (new_w->end)
 		{
-			char *fake_ArgList[3];
+			const char *fake_ArgList[3];
 
 			/* Fabricate a fake argument list */
 			fake_ArgList[0] = new_w->who_target;
 			fake_ArgList[1] = "fake_who_end";
 			fake_ArgList[2] = NULL;
-			new_w->end(from, fake_ArgList);
+			new_w->end(refnum, from, comm, fake_ArgList);
 		}
 		else if (new_w->who_end)
 		{
@@ -674,7 +761,7 @@ int	fake_who_end (char *from, char *who_target)
 	} 
 	while (new_w->piggyback && (new_w = new_w->next));
 
-	who_queue_pop();
+	who_queue_pop(refnum);
 	return 1;
 }
 
@@ -688,27 +775,37 @@ int	fake_who_end (char *from, char *who_target)
  *
  *
  */
-static void ison_queue_add (IsonEntry *item)
+static void ison_queue_add (int refnum, IsonEntry *item)
 {
-	IsonEntry *bottom = server_list[from_server].ison_queue;
+	Server *s;
+	IsonEntry *bottom;
 
+	if (!(s = get_server(refnum)))
+		return;
+
+	bottom = s->ison_queue;
 	while (bottom && bottom->next)
 		bottom = bottom->next;
 
 	if (!bottom)
-		server_list[from_server].ison_queue = item;
+		s->ison_queue = item;
 	else
 		bottom->next = item;
 
 	return;
 }
 
-static void ison_queue_pop (void)
+static void ison_queue_pop (int refnum)
 {
-	IsonEntry *save = server_list[from_server].ison_queue;
-	if (save)
+	Server *s;
+	IsonEntry *save;
+
+	if (!(s = get_server(refnum)))
+		return;
+
+	if ((save = s->ison_queue))
 	{
-		server_list[from_server].ison_queue = save->next;
+		s->ison_queue = save->next;
 		new_free(&save->ison_asked);
 		new_free(&save->ison_got);
 		new_free((char **)&save);
@@ -716,33 +813,46 @@ static void ison_queue_pop (void)
 	return;
 }
 
-static IsonEntry *ison_queue_top (void)
+static IsonEntry *ison_queue_top (int refnum)
 {
-	return server_list[from_server].ison_queue;
+	Server *s;
+
+	if (!(s = get_server(refnum)))
+		return NULL;
+
+	return s->ison_queue;
 }
 
-static IsonEntry *get_new_ison_entry (void)
+static IsonEntry *get_new_ison_entry (int refnum)
 {
-	IsonEntry *new_w = (IsonEntry *)new_malloc(sizeof(IsonEntry));
+	Server *s;
+	IsonEntry *new_w;
+
+	if (!(s = get_server(refnum)))
+		return NULL;
+
+	new_w = (IsonEntry *)new_malloc(sizeof(IsonEntry));
 	new_w->ison_asked = NULL;
 	new_w->ison_got = NULL;
 	new_w->next = NULL;
 	new_w->line = NULL;
-	ison_queue_add(new_w);
+	ison_queue_add(refnum, new_w);
 	return new_w;
 }
 
-static void ison_queue_list (void)
+static void ison_queue_list (int refnum)
 {
-	IsonEntry *item = server_list[from_server].ison_queue;
+	Server *s;
+	IsonEntry *item;
 	int count = 0;
 
-	while (item)
+	if (!(s = get_server(refnum)))
+		return;
+
+	for (item = s->ison_queue; item; item = item->next, count++)
 	{
 		yell("[%d] [%s] [%#x]", count, item->ison_asked, 
 				(unsigned)item->line);
-		count++;
-		item = item->next;
 	}
 }
 
@@ -751,35 +861,33 @@ BUILT_IN_COMMAND(isoncmd)
 	if (!args || !*args)
 		args = LOCAL_COPY(get_server_nickname(from_server));
 
-#if 1
 	if (!my_stricmp(args, "-d"))
 	{
-		ison_queue_list();
+		ison_queue_list(from_server);
 		return;
 	}
 	if (!my_stricmp(args, "-f"))
 	{
-		while (ison_queue_top())
-			ison_queue_pop();
+		while (ison_queue_top(from_server))
+			ison_queue_pop(from_server);
 		return;
 	}
-#endif
 
-	isonbase(args, NULL);
+	isonbase(from_server, args, NULL);
 }
 
-void isonbase (char *args, void (*line) (char *, char *))
+void isonbase (int refnum, char *args, void (*line) (int, char *, char *))
 {
 	IsonEntry 	*new_i;
 	char 		*next = args;
 
 	/* Maybe should output a warning? */
-	if (!is_server_connected(from_server))
+	if (!is_server_registered(refnum))
 		return;
 
 	while ((args = next))
 	{
-		new_i = get_new_ison_entry();
+		new_i = get_new_ison_entry(refnum);
 		new_i->line = line;
 		if (strlen(args) > 500)
 		{
@@ -792,7 +900,7 @@ void isonbase (char *args, void (*line) (char *, char *))
 			next = NULL;
 
 		malloc_strcpy(&new_i->ison_asked, args);
-		send_to_server("ISON %s", new_i->ison_asked);
+		send_to_aserver(refnum, "ISON %s", new_i->ison_asked);
 	}
 }
 
@@ -802,9 +910,9 @@ void isonbase (char *args, void (*line) (char *, char *))
  * Although we will check first that the top element expected is
  * actually an ISON.
  */
-void	ison_returned (char *from, char **ArgList)
+void	ison_returned (int refnum, const char *from, const char *comm, const char **ArgList)
 {
-	IsonEntry *new_i = ison_queue_top();
+	IsonEntry *new_i = ison_queue_top(refnum);
 
 	if (!new_i)
 	{
@@ -813,15 +921,18 @@ void	ison_returned (char *from, char **ArgList)
 	}
 
 	PasteArgs(ArgList, 0);
-	if (new_i->line)
-		new_i->line(new_i->ison_asked, ArgList[0]);
+	if (new_i->line) 
+	{
+		char *ison_ret = LOCAL_COPY(ArgList[0]);
+		new_i->line(refnum, new_i->ison_asked, ison_ret);
+	}
 	else
 	{
 		if (do_hook(current_numeric, "%s", ArgList[0]))
-			put_it("%s Currently online: %s", numeric_banner(), ArgList[0]);
+			put_it("%s Currently online: %s", banner(), ArgList[0]);
 	}
 
-	ison_queue_pop();
+	ison_queue_pop(refnum);
 	return;
 }
 
@@ -837,45 +948,60 @@ void	ison_returned (char *from, char **ArgList)
  *
  *
  */
-static void userhost_queue_add (UserhostEntry *item)
+static UserhostEntry *userhost_queue_top (int refnum)
 {
-	UserhostEntry *bottom = server_list[from_server].userhost_queue;
+	Server *s;
 
+	if (!(s = get_server(refnum)))
+		return NULL;
+
+	return s->userhost_queue;
+}
+
+static void userhost_queue_add (int refnum, UserhostEntry *item)
+{
+	UserhostEntry *bottom;
+	Server *s;
+
+	if (!(s = get_server(refnum)))
+		return;
+
+	bottom = userhost_queue_top(refnum);
 	while (bottom && bottom->next)
 		bottom = bottom->next;
 
 	if (!bottom)
-		server_list[from_server].userhost_queue = item;
+		s->userhost_queue = item;
 	else
 		bottom->next = item;
 
 	return;
 }
 
-static void userhost_queue_pop (void)
+static void userhost_queue_pop (int refnum)
 {
-	UserhostEntry *save = server_list[from_server].userhost_queue;
-	server_list[from_server].userhost_queue = save->next;
+	UserhostEntry *save;
+	Server *s;
 
+	if (!(s = get_server(refnum)))
+		return;
+
+	save = s->userhost_queue;
+	s->userhost_queue = save->next;
 	new_free(&save->userhost_asked);
 	new_free(&save->text);
 	new_free((char **)&save);
 	return;
 }
 
-static UserhostEntry *userhost_queue_top (void)
-{
-	return server_list[from_server].userhost_queue;
-}
-
-static UserhostEntry *get_new_userhost_entry (void)
+static UserhostEntry *get_new_userhost_entry (int refnum)
 {
 	UserhostEntry *new_u = (UserhostEntry *)new_malloc(sizeof(UserhostEntry));
 	new_u->userhost_asked = NULL;
 	new_u->text = NULL;
 	new_u->next = NULL;
 	new_u->func = NULL;
-	userhost_queue_add(new_u);
+	userhost_queue_add(refnum, new_u);
 	return new_u;
 }
 
@@ -885,20 +1011,20 @@ static UserhostEntry *get_new_userhost_entry (void)
  */
 BUILT_IN_COMMAND(userhostcmd)
 {
-	userhostbase(args, NULL, 1);
+	userhostbase(from_server, args, NULL, 1);
 }
 
 BUILT_IN_COMMAND(useripcmd)
 {
-	userhostbase(args, NULL, 0);
+	userhostbase(from_server, args, NULL, 0);
 }
 
 BUILT_IN_COMMAND(usripcmd)
 {
-	userhostbase(args, NULL, 2);
+	userhostbase(from_server, args, NULL, 2);
 }
 
-void userhostbase(char *args, void (*line) (UserhostItem *, char *, char *), int userhost)
+void userhostbase (int refnum, char *args, void (*line) (int, UserhostItem *, const char *, const char *), int do_userhost)
 {
 	int	total = 0,
 		userhost_cmd = 0;
@@ -910,7 +1036,7 @@ void userhostbase(char *args, void (*line) (UserhostItem *, char *, char *), int
 		*body = NULL;
 
 	/* Maybe should output a warning? */
-	if (!is_server_connected(from_server))
+	if (!is_server_registered(refnum))
 		return;
 
 	*buffer = 0;
@@ -919,21 +1045,21 @@ void userhostbase(char *args, void (*line) (UserhostItem *, char *, char *), int
 		if (check_nickname(nick, 1))
 		{
 			total++;
-			if (!fetch_userhost(from_server, nick))
+			if (!fetch_userhost(refnum, nick))
 				server_query_reqd++;
 
 			if (*buffer)
-				strmcat(buffer, " ", BIG_BUFFER_SIZE);
-			strmcat(buffer, nick, BIG_BUFFER_SIZE);
+				strlcat(buffer, " ", sizeof buffer);
+			strlcat(buffer, nick, sizeof buffer);
 		}
 
 		else if (!my_strnicmp(nick, "-cmd", 2))
 		{
 			if (!total)
 			{
-				if (userhost == 1)
+				if (do_userhost == 1)
 					say("USERHOST -cmd with no nicks specified");
-				else if (userhost == 0)
+				else if (do_userhost == 0)
 					say("USERIP -cmd with no nicks specified");
 				else
 					say("USRIP -cmd with no nicks specified");
@@ -957,7 +1083,7 @@ void userhostbase(char *args, void (*line) (UserhostItem *, char *, char *), int
 	if (!userhost_cmd && !total)
 	{
 		server_query_reqd++;
-		strlcpy(buffer, get_server_nickname(from_server), BIG_BUFFER_SIZE);
+		strlcpy(buffer, get_server_nickname(refnum), sizeof buffer);
 	}
 
 	ptr = buffer;
@@ -967,23 +1093,23 @@ void userhostbase(char *args, void (*line) (UserhostItem *, char *, char *), int
 		ptr = buffer;
 		while (ptr && *ptr)
 		{
-			UserhostEntry *new_u = get_new_userhost_entry();
+			UserhostEntry *new_u = get_new_userhost_entry(refnum);
 
 			move_to_abs_word(ptr, (const char **)&next_ptr, 5);
 
 			if (next_ptr && *next_ptr && next_ptr > ptr)
 				next_ptr[-1] = 0;
 
-			new_u->userhost_asked = m_strdup(ptr);
-			if (userhost == 1)
-				send_to_server("USERHOST %s", new_u->userhost_asked);
-			else if (userhost == 0)
-				send_to_server("USERIP %s", new_u->userhost_asked);
+			new_u->userhost_asked = malloc_strdup(ptr);
+			if (do_userhost == 1)
+				send_to_aserver(refnum, "USERHOST %s", new_u->userhost_asked);
+			else if (do_userhost == 0)
+				send_to_aserver(refnum, "USERIP %s", new_u->userhost_asked);
 			else
-				send_to_server("USRIP %s", new_u->userhost_asked);
+				send_to_aserver(refnum, "USRIP %s", new_u->userhost_asked);
 
 			if (userhost_cmd)
-				new_u->text = m_strdup(body);
+				new_u->text = malloc_strdup(body);
 
 			if (line)
 				new_u->func = line;
@@ -999,27 +1125,28 @@ void userhostbase(char *args, void (*line) (UserhostItem *, char *, char *), int
 	{
 		while (ptr && *ptr)
 		{
-			char *nick = next_arg(ptr, &ptr);
-			const char *ouh = fetch_userhost(from_server, nick);
-			char *uh;
+			char *my_nick = next_arg(ptr, &ptr);
+			const char *ouh = fetch_userhost(refnum, my_nick);
+			char *uh, *host;
 			UserhostItem item = {NULL, 0, 0, 0, NULL, NULL};
 
 			uh = LOCAL_COPY(ouh);
-			item.nick = nick;
+			item.nick = my_nick;
 			item.oper = 0;
 			item.connected = 1;
 			item.away = 0;
 			item.user = uh;
-			item.host = strchr(uh, '@');
-			if (item.host)
-				*item.host++ = 0;
-			else
+			host = strchr(uh, '@');
+			if (host) {
+				*host++ = 0;
+				item.host = host;
+			} else
 				item.host = "<UNKNOWN>";
 
 			if (line)
-				line(&item, nick, body);
+				line(refnum, &item, my_nick, body);
 			else if (userhost_cmd)
-				userhost_cmd_returned(&item, nick, body);
+				userhost_cmd_returned(refnum, &item, my_nick, body);
 			else
 				yell("Yowza!  I dont know what to do here!");
 		}
@@ -1033,10 +1160,11 @@ void userhostbase(char *args, void (*line) (UserhostItem *, char *, char *), int
  * through this queue will cause it to be corrupted and the client will
  * go higgledy-piggledy.
  */
-void	userhost_returned (char *from, char **ArgList)
+void	userhost_returned (int refnum, const char *from, const char *comm, const char **ArgList)
 {
-	UserhostEntry *top = userhost_queue_top();
+	UserhostEntry *top = userhost_queue_top(refnum);
 	char *ptr;
+	char *results;
 
 	if (!top)
 	{
@@ -1044,6 +1172,8 @@ void	userhost_returned (char *from, char **ArgList)
 		return;
 	}
 
+	PasteArgs(ArgList, 0);
+	results = LOCAL_COPY(ArgList[0]);
 	ptr = top->userhost_asked;
 
 	/*
@@ -1063,55 +1193,59 @@ void	userhost_returned (char *from, char **ArgList)
 		 * part of ArgList, and the following char will
 		 * either be a * or an = (eg, nick*= or nick=)
 		 */
-		if (ArgList && *ArgList && (!my_strnicmp(cnick, *ArgList, len)
-	            && ((*ArgList)[len] == '*' || (*ArgList)[len] == '=')))
+		if (results && (!my_strnicmp(cnick, results, len)
+	            && (results[len] == '*' || results[len] == '=')))
 		{
 			UserhostItem item;
+			char *nick, *user, *host;
 
 			/* Extract all the interesting info */
 			item.connected = 1;
-			item.nick = next_arg(*ArgList, ArgList);
-			item.user = strchr(item.nick, '=');
-			if (!item.user)
+			nick = next_arg(results, &results);
+			user = strchr(nick, '=');
+			if (!user)
 			{
-				yell("Can't parse broken USERHOST reply [%s]", 
-						*ArgList);
-				userhost_queue_pop();
+				yell("Can't parse useless USERHOST reply [%s]", 
+						ArgList[0]);
+				userhost_queue_pop(refnum);
 			}
 
-			if (item.user[-1] == '*')
+			if (user[-1] == '*')
 			{
-				item.user[-1] = 0;
+				user[-1] = 0;
 				item.oper = 1;
 			}
 			else
 				item.oper = 0;
 
-			if (item.user[1] == '+')
+			if (user[1] == '+')
 				item.away = 0;
 			else
 				item.away = 1;
 
-			*item.user++ = 0;
-			item.user++;
+			*user++ = 0;
+			user++;
 
-			item.host = strchr(item.user, '@');
-			if (!item.host)
+			host = strchr(user, '@');
+			if (!host)
 			{
-				yell("Can't parse broken USERHOST reply [%s]", 
-						*ArgList);
-				userhost_queue_pop();
+				yell("Can't parse useless USERHOST reply [%s]", 
+						ArgList[0]);
+				userhost_queue_pop(refnum);
 				return;
 			}
-			*item.host++ = 0;
+			*host++ = 0;
 
+			item.nick = nick;
+			item.user = user;
+			item.host = host;
 
 			/*
 			 * If the user wanted a callback, then
 			 * feed the callback with the info.
 			 */
 			if (top->func)
-				top->func(&item, cnick, top->text);
+				top->func(refnum, &item, cnick, top->text);
 
 			/*
 			 * Otherwise, the user just did /userhost,
@@ -1123,7 +1257,7 @@ void	userhost_returned (char *from, char **ArgList)
 						item.oper ? "+" : "-", 
 						item.away ? "-" : "+", 
 						item.user, item.host))
-				put_it("%s %s is %s@%s%s%s", numeric_banner(),
+				put_it("%s %s is %s@%s%s%s", banner(),
 						item.nick, item.user, item.host, 
 						item.oper ?  " (Is an IRC operator)" : empty_string,
 						item.away ? " (away)" : empty_string);
@@ -1155,47 +1289,41 @@ void	userhost_returned (char *from, char **ArgList)
 				item.oper = item.away = 0;
 				item.connected = 1;
 
-				top->func(&item, cnick, top->text);
+				top->func(refnum, &item, cnick, top->text);
 			}
 		}
 	}
 
-	userhost_queue_pop();
+	userhost_queue_pop(refnum);
 }
 
-void	userhost_cmd_returned (UserhostItem *stuff, char *nick, char *text)
+void	userhost_cmd_returned (int refnum, UserhostItem *stuff, const char *nick, const char *text)
 {
-	char	args[BIG_BUFFER_SIZE + 1];
+	char	*args = NULL;
+	size_t	clue = 0;
 
 	/* This should be safe, though its playing it fast and loose */
-	strcpy(args, stuff->nick ? stuff->nick : empty_string);
-	strcat(args, stuff->oper ? " + " : " - ");
-	strcat(args, stuff->away ? "+ " : "- ");
-	strcat(args, stuff->user ? stuff->user : empty_string);
-	strcat(args, space);
-	strcat(args, stuff->host ? stuff->host : empty_string);
+	malloc_strcat_c(&args, stuff->nick ? stuff->nick : empty_string, &clue);
+	malloc_strcat_c(&args, stuff->oper ? " + " : " - ", &clue);
+	malloc_strcat_c(&args, stuff->away ? "+ " : "- ", &clue);
+	malloc_strcat_c(&args, stuff->user ? stuff->user : empty_string, &clue);
+	malloc_strcat_c(&args, space, &clue);
+	malloc_strcat_c(&args, stuff->host ? stuff->host : empty_string, &clue);
 	parse_line(NULL, text, args, 0, 0);
+
+	new_free(&args);
 }
 
-void clean_server_queues (int i)
+void	clean_server_queues (int i)
 {
-	int old_from_server = from_server;
+	while (who_queue_top(i))
+		who_queue_pop(i);
 
-	if (i == -1 || !server_list /* || !server_list[i].connected */)
-		return;		/* Whatever */
+	while (ison_queue_top(i))
+		ison_queue_pop(i);
 
-	from_server = i;
-
-	while (server_list[i].who_queue)
-		who_queue_pop();
-
-	while (server_list[i].ison_queue)
-		ison_queue_pop();
-
-	while (server_list[i].userhost_queue)
-		userhost_queue_pop();
-
-	from_server = old_from_server;
+	while (userhost_queue_top(i))
+		userhost_queue_pop(i);
 }
 
 
